//************************************************************
//  Interpolation, differentiation, and integration functions
//************************************************************
// Reference:   SOFTWARE SOLUTIONS
// Workspace  : Chapter12\Test Un32_10
// Entry point: Test Un32_10
// Other files:
//              Un32_9.h
//              Un32_9.obj      
// Development: Visual C++ 6.0 Console Application
// Date:        March 1/2001
//
// Description:
//              Header file for primtive functions used in
//    interpolation, differentiation, and integration.
//************************************************************

// Library functions
extern "C" LOG_E();         // base e logarithm
extern "C" ALOG_E();        // base e antilogarithm
extern "C" X_TO_Y_MIXED();  // x**y by logs and binary powering

//**************************************************************
//                       Prototypes
//**************************************************************

//**************************************************************
//                   support functions
//**************************************************************
double LogE(double x)
{
// Logarithm base E
// On entry:
//     parameter x number
// On exit:
//     returns Log E of x
double logx;
_asm
{
	FLD     x
	CALL    LOG_E
	FSTP    logx
}
  return logx;  
}

double AlogE(double x)
{
// Antilogarithm base E
// On entry:
//     parameter x is logarithm base E
// On exit:
//     returns number whose log is x
double numx;
_asm
{
	FLD     x
	CALL    ALOG_E
	FSTP    numx
}
  return numx;  
}

double XToYMixed(double x, double y)
{
// Approximate calculation of x**y by logarithms
// and binary powering
// On entry:
//     parameter x is base
//     parameter y is the exponent
// On exit:
//     returns x**y

double xtoy = 0;
_asm
{
	FLD     x
    FLD     y
	CALL    X_TO_Y_MIXED
	FSTP    xtoy
}
  return xtoy;  
}

//**************************************************************
//                   C++ functions 
//**************************************************************

double InterpLinPky(double pxi,
				    double pyi,
				    double  pxj,
				    double pyj,
				    double pxk)
{
// Find the y coordinate of point Pk, which is assumed to be
// along a straight line connecting points Pi and Pj, as per
// the following equation:
//
//  Pyk = ((Pyj-Pyi)(Pxk-Pxi)/(Pxj-Pxi))+Pyi
//
// where:
//      Pyj is the y coordinate of point Pj
//      Pxj is the x coordinate of point Pj
//      Pyi is the y coordinate of point Pi
//      Pxi is the x coordinate of point Pi
//      Pxk is the x coordinate of the desired point
//      Pyk is the y value found by interpolation
// On entry:
//      parameters pxj, pyj, pxi, pyi, pxk are
//      the coordinates of the corresponding points
// Returns:
//      The y coordinate of point Pk found by linear
//    	interpolation, as a value of type double.
     
return (((pyj - pyi)*(pxk - pxi))/(pxj - pxi)) + pyi;

}

double InterpLinFromTable(double *datX,  // x array
						  double *datY,  // y array
						  double pxk,    // x coordinate
						  int size)      // array size 
{
// Find the y coordinate of point Pk, which is assumed to be
// along a straight line connecting two points, Pi and Pj,
// from the coordinate data contained in two arrays of
// points on the curve. Uses the InterpLinPky() function.
// On entry:
//      array datX[] contains x coordinates of points
//      array datY[] contains y coordinates of points
//      parameter pkx is the x coordinate of point Pk
//      size is the number of elements in both arrays
//      calculated by the caller, as follows:
//             size = sizeof datX / sizeof datY[0];
// Assumptions:
//      Coordinate points in arrays datX[] and datY[]
//      are sorted in increasing order.
// Processing:
//      Routine searches the array of x coordinates
//      until one is found that is larger that the value
//      Pkx. This point is labled Pj. If there is a
//      previous x to point Pj it is labled Pi. The
//      coordinates of Pi and Pj, and the value Pkx
//      are then passed to the InterpLinPky()
//      function.
// Returns:
//      The y coordinate of point Pk found by linear
//    	interpolation, as a value of type double.
//      Returns 0 if there are no adjacent points
//      in the data set.

// Find point Pkj in array datX[]
double pyj, pyi, pxj, pxi;
double pky = 0;

int found = 0;

for(int w = 0; w < size; w++)	
{
	if(datX[w] > pxk || w != 0)
	{
		pxj = datX[w];
		pxi = datX[w-1];
		pyj = datY[w];
		pyi = datY[w-1];
		found = 1;
		break;
	}
}

if(found == 0)
   return 0;

// Calculate Pky using InterpLinPky() function
pky = InterpLinPky(pxi, pyi, pxj, pyj, pxk);
return pky;
}


double InterpLagrange(double *datX,  // x array
					  double *datY,  // y array
					  double pxk,    // x coordinate
					  int size)      // array size 
{
// Find the y coordinate of point Pk using Lagrangian
// interpolation
// On entry:
//      array datX[] contains x coordinates of points
//      array datY[] contains y coordinates of points
//      parameter pkx is the x coordinate of point Pk
//      size is the number of elements in both arrays
//      calculated by the caller, as follows:
//             size = sizeof datX / sizeof datY[0];
// Assumptions:
//      Coordinate points in arrays datX[] and datY[]
//      are sorted in increasing order.
// Returns:
//      The y coordinate of point Pk found by linear
//    	interpolation, as a value of type double.
//      Returns 0 if there are no adjacent points
//      in the data set.

double lTerm = 0;   // Each Lagrange term
double pxy = 0;     // Value of the dependent variable

for(int i = 0; i < size; i++)
{
	lTerm =  datY[i];

	for(int j = 0; j < size; j++)
	{
		if(j != i)
			lTerm = lTerm * (pxk - datX[j])/(datX[i]-datX[j]);
	}
	pxy = pxy + lTerm;
}
return pxy;
}


template <class A>
double LinearLSInterp(A *datX,      // Array of x coordinates
					  A *datY,      // Array of y coordinates
					  A pxk,        // x coordinate of point Pk
					  double *eSum, // Caller's variable for 
					                // error sum 
					  int size)     // Array size
{
// Function to find the y coordinate of point Pk using
// the least-squares methods
// On entry:
//      array datX[] contains x coordinates of points
//      array datY[] contains y coordinates of points
//      parameter pkx is the x coordinate of point Pk
//      size is the number of elements in both arrays
//      calculated by the caller, as follows:
//             size = sizeof datX / sizeof datY[0];
//
// STEP 1:
// Calculate slope (m) in the formula:
//
//      y' = mx + b
//
//            n(SIGxy) - (SIGx)(SIGy)
//      m = -----------------------------------
//               n(SIGx^2) - (SIGx)^2
// where:
//       x are the data elements of the array datX
//       y are the data elements of the array dayY
//       n is the number of elements
//       m is the slope of y' = mx + b
//
// STEP 2:
// Calculate y-intercept (b) in the formula:
//
//      y' = mx + b
//
//            SIGy - m(SIGx)
//      b = -----------------
//                 n
// where:
//       x are the data elements of the array datX
//       y are the data elements of the array dayY
//       n is the number of elements
//       m is the slope of y' = mx + b
//
// STEP 3:
//       Solve for point y applying
//           pyk = mx + b
//

double m = 0;          // y-intercept
double nume;           // Storage for numerator
double d1, d2;         // Storage for denominator
A sumXY = 0;      // SIGxy
A sumX = 0;       // SIGx
A sumY = 0;       // SIGy
A sumXSq = 0;     // SIGx^2
// Variables for calculating slope 
double b = 0;          // slope  
double pyk;            // Y coordinate of point Pk
// Variables cor calculating error sum
double res = 0;        // Residual accumulator
double newY = 0;
double err = 0;        // Accumulator for error sum

// Calcuate sums required in formula
for(int i = 0; i < size; i++)
{
    sumXY = sumXY + (datX[i] * datY[i]);
    sumX = sumX + datX[i];
    sumY = sumY + datY[i];
    sumXSq = sumXSq + (datX[i] * datX[i]);
}

// Calculate slope (m)
//
//            n(SIGxy) - (SIGx)(SIGy)
//      m = -----------------------------------
//               n(SIGx^2) - (SIGx)^2
//

// Start with numerator: nSIGxy - (SIGx)(SIGy)
nume = (double) size * sumXY - (sumX * sumY);

// Calculate denominator left and right elements
// d1 = n(SIGx^2)
// d2 = (SIGx)^2
d1 = size * sumXSq;
d2 = sumX * sumX;
// Now obtain y-intercept 
m = nume / (d1 - d2);

// Calculate slope (b)
//
//            SIGy - m(SIGx)
//      b = -----------------
//                 n
//

// Start with numerator: SIGy - m(SIGx)
nume = (double) sumY - (m * sumX);
// Now obtain b
b = nume / size;

// At this point:
//    variable b holds slope
//    variable m holds y-intercept
//    parameter pxk holds value of x
// Calculate y-coordinate using
//    y' = mx + b 
pyk = (m * pxk) + b; 

// Calculate error sum using equation and residuals
for(int k = 0; k < size; k++)
{
   newY = (m * datX[k]) + b;
   res = datY[k] - newY;
   err = err + (res * res);
}
// Store error sum in caller's variable
*eSum = err;
// Return y coordinate for point Pk
return pyk;
}


double NonLinLSInterp(double *datX,   // Array of x coordinates
					  double *datY,   // Array of y coordinates
					  double  pxk,    // x coordinate of point Pk
					  double *eSum,   // Caller's variable for 
					                  // error sum 
					  int size)       // Array size
{
// Function to find the y coordinate of point Pk using
// the least-squares methods, assuming that the function
// is of the form:
//
//         f(x) = bx^m
//
// by performing the following transformation:
//
//        ln f(x) = m * ln x + ln b
//
// where:
//        y  = f(x)
//        b' = ln b
//        y' = ln y
//        x' = ln x
//        
// On entry:
//      array datX[] contains x coordinates of points
//      array datY[] contains y coordinates of points
//      parameter pkx is the x coordinate of point Pk
//      size is the number of elements in both arrays
//      calculated by the caller, as follows:
//             size = sizeof datX / sizeof datY[0];
//
// PROCESSING:
// STEP 1:
//        Convert the values in arrays datX[] and dat[Y]
//        natural logs. 
//
// STEP 2:
//        Use these values to calculate y-intercept (m)
//
// STEP 3:
//        Calculate slope b' and convert to b.
//
// STEP 4:
//        Solve for point pxy by applying the original
//        equation:
//
//        y = bx^m
//
// STEP 5:
//        Calculate and store error sum in caller's
//        variable

double m = 0;          // y-intercept
double nume;           // Storage for numerator
double d1, d2;         // Storage for denominator
double sumXY = 0;      // SIGxy
double sumX = 0;       // SIGx
double sumY = 0;       // SIGy
double sumXSq = 0;     // SIGx^2
// Variables for calculating slope 
double b = 0;          // slope  
double lnb = 0;        // ln of slope
double pyk;            // Y coordinate of point Pk
// Variables cor calculating error sum
double res = 0;        // Residual accumulator
double err = 0;
double newY = 0;
// Arrays for natural logs of x and y
double *lnX = new double[size];
double *lnY = new double[size]; 

// Convert data to ln
for(int l = 0; l < size; l++)
{
	lnX[l] = LogE(datX[l]);
    lnY[l] = LogE(datY[l]);
}


// Calcuate sums required in formula
for(int i = 0; i < size; i++)
{
    sumXY = sumXY + (lnX[i] * lnY[i]);
    sumX = sumX + lnX[i];
    sumY = sumY + lnY[i];
    sumXSq = sumXSq + (lnX[i] * lnX[i]);
}

// Calculate y-intercept (m)
//
//            n(SIGxy) - (SIGx)(SIGy)
//      m = -----------------------------------
//               n(SIGx^2) - (SIGx)^2
//

// Start with numerator: nSIGxy - (SIGx)(SIGy)
nume = (double) size * sumXY - (sumX * sumY);

// Calculate denominator left and right elements
// d1 = n(SIGx^2)
// d2 = (SIGx)^2
d1 = size * sumXSq;
d2 = sumX * sumX;
// Now obtain y-intercept 
m = nume / (d1 - d2);

// Calculate ln of slope (b')
//
//             SIGy - m(SIGx)
//      b' = -----------------
//                   n
//
// Start with numerator: SIGy - m(SIGx)
nume = (double) sumY - (m * sumX);
// Now obtain b
lnb = nume / size;
// Solve for b
b = AlogE(lnb);

// Calculate y-coordinate using
//    y = bx^m 
pyk = b * XToYMixed(pxk, m);

// Calculate error sum using equation and residuals
for(int k = 0; k < size; k++)
{
   newY = b * (XToYMixed(datX[k], m));
   res = datY[k] - newY;
   err = err + (res * res);
}
// Store error sum in caller's variable
*eSum = err;

return pyk;
}

void FirstDeriv(double *datY,   // Array of y coordinates
			    double *drv1,   // Array for first derivatives
  			    int size)       // Size of data arrays
{
// Function to find the first derivative of a set of coordinate
// points using the forward difference, central difference,
// and backward difference formulas:
//
//    Py'k = Pyk - Pyi        <= forward difference
//
//    Py'k = Pyj - Pyi / 2h   <= central difference
//
//    Py'k = Pyj - Pyk        <= backward difference
//
// where:
//    Py'k is the derivative at point Pk
//    Pyj is the coordinate of a point Pj, adjacent to Pk
//    Pyi is the coordinate of a point Pi, adjacent to Pk
//    h is assumed to be 1.
// On entry:
//      array datY[] contains y coordinates of points
//      array drv1[] is used to return the first derivatives
//      size is the number of elements in array datY[]
//      calculated by the caller, as follows:
//             size = sizeof datY / sizeof datY[0];

// Calculate derivative of first point using forward difference
drv1[0] = datY[1] - datY[0];

// Calculate central points using central difference
for(int x = 1; x < (size - 1); x++)
	drv1[x] = (datY[x + 1] - datY[x - 1]) / 2.0;

// Calculate the derivative of the last point using backward
// difference
drv1[size - 1] = datY[size - 1] - datY[size - 2];

return;
}

void SecondDeriv(double *datY,  // Array of y coordinates
			    double *drv2,   // Array for second derivatives
  			    int size)       // Size of data arrays
{
// Function to find the second derivative of a set of coordinate
// points using the central difference formula:
//
//    Py"k = Pyj - 2Pyk + Pyi / h^2
//
// where:
//    Py"k is the second derivative at point Pk
//    Pyj is the coordinate of a point Pj, adjacent to Pk
//    Pyi is the coordinate of a point Pi, adjacent to Pk
//    h is assumed to be 1.
// On entry:
//      array datY[] contains y coordinates of points
//      array drv1[] is used to return the first derivatives
//      size is the number of elements in array datY[]
//      calculated by the caller, as follows:
//             size = sizeof datY / sizeof datY[0];
// Note:
//      Since the central difference formula cannot be applied
//      at end points the second derivative at offset [0] and
//      at offset [size - 1] are undefined
//

// Calculate central points using central difference
for(int x = 1; x < (size - 1); x++)
	drv2[x] = datY[x + 1] - (2 * datY[x]) + datY[x - 1];

return;
}

//*********************************************
//                integration
//*********************************************
double TrapRule(double *datX,  // Array of x values
    	        double *datY,  // Array of f(x) values
		        int size)      // Size of datY[] array 
{
// Calculates integral by estimating the area under the curve.
// Integration is performed by adding the areas of individual
// trapezoids, calculated according to the formula:
//
//          (Xk - Xk+1)(Yk + Yk+1)
//    Tk = -----------------------
//                   2
//
// where:
//       Tk is the area of the trapezoid
//       Yk, Yk+1 ... Ysize are values of f(x) stored in array
//                    datY[]
//       Xk, Xk+1 ... Xsize are values of x stored in array
//                    datX[]
//       size is the number of elements in arrays datX[] and
//       datY[]
//
// The number of trapezoids is one less than the number of
// data points in the arrays datX[] and datY[]
//    
// Local variables
double integral = 0;
double base = 0;
double *tAreas = new double[size - 1];           // Array for trapezoid areas

// Calculate the area of size-1 trapezoids
for(int i = 0; i < (size - 1); i++)
{
	// Calculate area for this trapezoid
	base = datX[i+1] - datX[i];    // Base for this trapezoid
    tAreas[i] = (base * (datY[i] + datY[i+1]))/2;
}

// At this point there are size - 1 entries in the array of
// trapezoid areas. Add values to calculate integral
for(int z = 0; z < (size -1) ; z++)
  integral = integral + tAreas[z];  

return integral;
}

double SimpRule(double *datX,  // Array of x values
    	        double *datY,  // Array of f(x) values
		        int size)      // Size of datY[] array 
{
// Calculates integral by estimating the area under the curve.
// Integration is performed by means of Simpson's rule,
// as follows:
//           
//    Np =  step/3 * (y1 + 4y2 + 2y3 + 4y4 + 2y5 + 4y6 .. yN+1) 
//          ------    ----------------------------------------
//             |                        |
//             |                        |___ sum term
//             |____________________________ first term
//
// where:
//       N is the number of area elements used in the
//       computation. Must be even.
//       y1, y2 .. yN+1 are values of f(x)
//       step = difference between x coordinates
//       Np is the integral
//
// The sum term consists of 4 operations:
//    1. y1
//    2. 4 * (sums of all yj, j even from 2 to N)
//    3. 2 * (sums of all yk, k odd from 3 to N - 1)
//    4. yN+1	
//  
// Local variables

double area = 0;
int N = size - 1;       // Number of areas
double *tAreas = new double[N];
double base = 0;

// Calculate base size using the formula:
// b = (b - a)/N
// where a and b are the limit values of the independent
// variable
base = (datX[size - 1] - datX[0]) /N;

// Perform sum term operation 1 and store in local array
tAreas[0] = datY[0];

// Perform sum term operation 4 and store in local array
tAreas[1] = datY[N];

// Perform sum term operation 2 and store sums in local array
// Operation: 4 * (sums of all yj, j even, from 2 to N)
for(int j = 2; j <= N; j += 2)
    tAreas[j] = 4 * datY[j-1];

// Perform sum term operation 3 and store sums in local array
// Operation: 2 * (sums of all yk, k odd, from 3 to N - 1)
for(int k = 3; k <= (N - 1); k += 2)
    tAreas[k] = 2 * datY[k-1];

// At this point there are N number of entries in the
// local array holding sum term. Add values.
for(int z = 0; z < size ; z++)
  area = area + tAreas[z];

// Apply first term of Simpson's equation
area = (base * area) /3;  

return area; 
}

