//************************************************************
//                    Equation Parsing
//************************************************************
// Reference:   SOFTWARE SOLUTIONS
// Workspace  : Chapter14\Test Un32_13
// Entry point: Test Un32_13
// Other files:
// Development: Visual C++ 6.0 Console Application
// Date:        March 27/2001
//
// Description:
//              Header file for primtive functions used in
//              solving parsing equations.
//************************************************************

// MATH32 library functions
extern "C" void EVALUATE(); 
extern "C" void CALCULATE_Y();
//**************************************************************
//                      Interface Functions
//**************************************************************
double Evaluate(char *equation,
			    double varX)
{
// Function to evaluate an equation, or part of an equation,
// that contains no parenthetical components, according to
// the following rules of equation grammar:
//
//          SIN (sine)         01H
//          COS (cosine)       02H
//          TAN (tangent)      03H
//          ASIN (arcsine)     04H
//          ACOS (arcosine)    05H
//          ATAN (arctangent)  06H
//
// Recognized symbols:
//                   ^ = exponentiation
//                   * = multiplication
//                   / = division
//                   + = addition (or unary +)
//                   - = subtraction (or unary -)
//
// Other elements of equation grammar:
//            1. ASCII constants, for example:
//                   1, 2.33, or -1.2E+233
//            2. The independent variable, represented by the
//               letter x or X.
//            3. Binary constants, represented by a single digit
//               entry in excess 80H format
//
// Rules of precedence:
//           1. Equation is evaluated left-to-right
//           2. Trigonometric functions are evaluated first
//           3. then exponentiation operations
//           4. then multiplication and division
//           5. addition and subtraction are performed last
//
// On entry:
//          equation[] is an equation string terminated in 0H
//          Equal sign and elements to its left must be removed
//          by the calling routine
//          Variable varX is the independent variable
// Uses:
//          The _EVALUATE procedure in the Un32_14 module of
//          the MATH32 library
// On exit:
//          returns the solution of equation or mathematical
//          expression passed by the caller in a type double
//

double result;
_asm
{
		 MOV     ESI,equation         // Address to ESI
		 FLD     varX
         CALL    EVALUATE
		 FSTP    result
}
return result;
}


double CalculateY(char *equation,
			      double varX)
{
// Function to calculate the dependent variable in a mathematical
// expression that may include parenthetical terms, possibly nested.
// This function uses the _EVALUATE procedure to solve the
// expressions enclosed in parenthesis, or to resolve an expression
// without parenthesis.
// On entry:
//         equation[] is an array of char containing the equation
//         text in the format:
//                 y = w
//                 where w is an expression that may include
//                 parenthetical terms, the independent
//                 variable, numeric constants, and the keywords
//                 and operators listed in the header of the
//                 _EVALUATE procedure
//                 The equation must terminate in 0H or 0DH
//         ST(0) = value of the independent variable (x)
//
// Uses:
//          The _EVALUATE procedure in the Un32_14 module of
//          the MATH32 library
// On exit:
//          returns the solution of equation or mathematical
//          expression passed by the caller, in a type double
//

double result;
_asm
{
		 MOV     ESI,equation         // Address to ESI
		 FLD     varX
         CALL    CALCULATE_Y
		 FSTP    result
}
return result;
}