// ***********************************************************************
// *******************************************
// Languge: C++
// Machine: Pentium PC under Windows 98/2000
// Environment: Visual C++ Version 6.0
// Program Type: Console application
// Filename: Backprop.cpp
// *******************************************
//
//*************************************************************************
//                          program description
//*************************************************************************
// 1. The NeuronS() function simulates a neuron that returns a sigmoid
//    function, in this case the function used is the logistic.
// 2. The BackProp323() function implements backpropagation on a network
//    with 3 input layers, 2 hidden layers, and 3 output layers
// 3. The function Network323() is used to iterate a data set through the
//    323 network and capture the values in the output layer.
// 4. The function BPTrain323() is used to train the 323 network using
//    the function Network323() and the function BackProp323(). Training
//    ends when a predefined numer of epochs are iterated, or when the
//    weighted error sum becomes smaller than a predefined limit.
// 5. The training data is as follows:
//    Data set 1: input [1 2 3] expected response [0.1 0.3 0.7]
//    data set 2: input [3 4 5] expected response [0.2 0.3 0.5]
//    The backpropagation routine reaches the predefined error limit with
//    both data sets.
//    Note: the multiple-node design of this network makes it difficult to
//          use with the test cases used in previous exercises. 
// 
//*************************************************************************
//                        core program functions
//*************************************************************************
//*****************************
//        NeuronS() 
//*****************************
// Neuron-like unit that returns the logistic sigmoid function
// 1. NeuronS()receives the following paramenters:
//        xIn[] is and array of values
//        wts is an array of weights
//        size is the number of elements in data arrays
//     NeuronS() computes:
//     f(x) = the sum of products of xIn[] by wts[]
//     Returns the logistic function:
//     f(x) = 1/1+e^-f(x)
//
//*******************************
//      BackProp323()
//*******************************
// Function to perform backpropagation on a network with 3 nodes
// in the input layer, 2 nodes in the hidden layer, and 3 nodes
// in the output layer
// Function takes the following parameters:
//		double inD[],     // Input array of 3 elements
//		double outD[],    // 3 elements with desired output 
//		double wHL1[],    // 2, 3-element hidden layer weigths
//		double wHL2[],
//		double wOL1[],    // 3, 2-element output layer weights
//		double wOL2[],
//		double wOL3[],
//		double eDif[])    // Storage for 3 error differences 
//
// Processing consists of the following steps
//
// Step 1:
// Presents the input data, stored in the global array to the hidden
// layer neurons using weight tables passed as arguments
//
// Step 2:
// Outputs from the hidden layer are presented to 3 output layers
// using the weight tables passed as arguments
//
// Step 3:
// Function calculates e = t - t^ for each output cell
// Then calculates the derivative of the output unit responses:
//    Logistic function:     f(x) = 1/1+e^-f(x)
//    Derivative:            fDx = f(x)[1-f(x)]
// Finally, it calculates the error signal:
// deltaE = f'(x) * e
//
// Step 4:
// Calculate error signal of the output units:
// Multiply error signal of each output unit by the output unit
// weights 
//
// Step 5:
// Calculate error signal of hidden units using the derivatives
// of the hidden unit responses:
//
// Step 6:
// Calculate the error value for correcting weights, as follows:
//        deltaZ = eta * inData[] * eSignal[]
//
// Step 7:
// Update weights in hidden layer unit
//
// Step 8:
// Update weights in output unit

//***********************************
//        Network323()
//***********************************
// Function to execute a network with 3 neurons in the input layer,
// 2 in the hidden layer, and 3 in the output layer
// Input parameters:
//		double inD[],     // Input array of 3 elements
//		double outD[],    // Array of 3 elements for results 
//		double wHL1[],    // 2, 3-element hidden layer weigths sets
//		double wHL2[],
//		double wOL1[],    // 3, 2-element output layer weights sets
//		double wOL2[],
//		double wOL3[])
//
// Step 1:
// Present the input data, stored in the array named inD[]
// to the hidden layer neurons using weights tables wHL1[] and wHL2[].
//
// Step 2:
// Outputs from the hidden layer (in outHL1[]) are presented 
// to the 3 output layers using the weights in wOL1[], wOL2[],
// and wO3[] 
//
//**********************************
//        BPTrain323()
//**********************************
// This function uses BackProp323() and Network323() to train the
// neural network until a LIMIT number of training epochs is reached
// or when a predefined error difference (ELIMIT) is achieved
// Input parameters:
//		double inD[],     // Input array of 3 elements//
//	    double outD[],    // Array of 3 elements for results 
//	    double wHL1[],    // 2, 3-element hidden layer weigths sets
//	    double wHL2[],
//	    double wOL1[],    // 3, 2-element output layer weights sets
//	    double wOL2[],
//	    double wOL3[],
//		double *eTotal)   // Error total returned to caller
//*************************************************************************
//*************************************************************************

#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>


// Prototypes
double NeuronS(double[], double[], int);
void BackProp232(double [], double [],double [],
			 	 double [], double [],double [],
				 double [], double[]);
void Network232(double [], double [],double [],
				double [], double [],double [],
				double []);
int BPTrain323(double [], double [],double [],
			   double [], double [],double [],
			   double [], double*);

//******************************
//      Global variables
//******************************
// Weights for hidden layer
double wH1[] = {0.5, 0.3, 0.1};
double wH2[] = {0.3, 0.2, 0.1};

// Weights for output layer
double wO1[] = {0.1, 0.2};
double wO2[] = {0.3, 0.4};
double wO3[] = {0.5, 0.6};

// Input data
double inData[] = {1.0, 2.0, 3.0};
// Output data
double outData[] = {0.1, 0.3, 0.7}; 

// Storage for original weights
double h1[] = {0.5, 0.3, 0.1};
double h2[] = {0.3, 0.2, 0.1};

// Weights for output layer
double o1[] = {0.1, 0.2};
double o2[] = {0.3, 0.4};
double o3[] = {0.5, 0.6};

// Array for results
double results[3];    // Output values
double errors[3];     // Error differences

// Alternate data set
// Input data
double inData1[] = {1.0, 2.0, 3.0};
// Output data
double outData1[] = {0.1, 0.3, 0.7};

// Input data
double inData2[] = {3.0, 4.0, 5.0};
// Output data
double outData2[] = {0.2, 0.3, 0.5};

//*********************************************
//                 main()
//*********************************************
main()
{

  // Local data
  int option = 1;
  int j;
  double error = 0;
  
  while(option > 0)
  {
    // Menu for neuron testing
	cout << "\nTESTING PROGRAM OPERATIONS\n"
	     << "   1 Train network using backpropagation\n"
	     << "   2 Backpropagation epoch-by-epoch\n"
	     << "   3 Reset original weights\n"
	     << "   4 Reset original data\n"
	     << "   5 Set alternate data\n"	
	     << "   0 to end\n"
	     << "Enter option number: ";
	cin >> option;
	cout << "\n\n";

      switch(option)
      {
	  case 1:
	  cout << "Training network with backpropagation\n"
           << "Stimulus: [" << inData[0] << " " << inData[1]
	       << " " << inData[2] << "]  "
	       << " expected response: [" << outData[0] << " " << outData[1]
		   << " " << outData[2] << "]\n";
	  
      // Call training function
      if (BPTrain323(inData, outData, wH1, wH2, wO1, wO2, wO3, &error))
	  {
		// At this point termination is due to limit elements
		// Iterate network with final weights
		Network232(inData, results, wH1, wH2, wO1, wO2, wO3);
		cout << "Response after backpropagation training\n"
			 << "[" << results[0] << " " << results[1]
			 << " " << results[2] << "]\n"
	         << "error limit reached: " << error << "\n";
	  }
	  else
		  // Termination is due to number of iterations
		  cout << "epoch limit was reached\n";
	  break;

	  case 2:
      // Testing backpropagation step-by-step  
	  // Change weights using backpropagation
      BackProp232(inData, outData, wH1, wH2, wO1, wO2, wO3, errors);
	  // Iterate network with new weights
      Network232(inData, results, wH1, wH2, wO1, wO2, wO3);
	  cout << "Output data after backpropagation";
	  cout << "   1: " << results[0] << "   2: " << results[1];
	  cout << "   3: " << results[2] << "\n";
	  cout << " Errors: " << errors[0] << "   ";
	  cout << errors[1] << "   " << errors[2] << "\n"; 
	  break;

	  case 3:
      // Reset original weights
      for(j = 0; j < 2; j++)
	  {
		  wH2[j] = h1[j];
		  wH2[j] = h2[j];
	  }
	  for(j = 0; j < 3; j++)
	  {
		  wO1[j] = o1[j];
		  wO2[j] = o2[j];
		  wO3[j] = o3[j];
	  }
	  cout << "Original weights are now reset\n";
	  break;

	  case 4:
      // Reset original data
      for(j = 0; j < 3; j++)
	  {
	  	  inData[j] = inData1[j];
		  outData[j] = outData1[j];
	  }
	        // Reset original weights
      for(j = 0; j < 2; j++)
	  {
		  wH2[j] = h1[j];
		  wH2[j] = h2[j];
	  }
	  for(j = 0; j < 3; j++)
	  {
		  wO1[j] = o1[j];
		  wO2[j] = o2[j];
		  wO3[j] = o3[j];
	  }
	  cout << "Original data restored\n";
	  break;

	  case 5:
      // Set alternate data
      for(j = 0; j < 3; j++)
	  {
	  	  inData[j] = inData2[j];
		  outData[j] = outData2[j];
	  }
	  // Reset original weights
      for(j = 0; j < 2; j++)
	  {
		  wH2[j] = h1[j];
		  wH2[j] = h2[j];
	  }
	  for(j = 0; j < 3; j++)
	  {
		  wO1[j] = o1[j];
		  wO2[j] = o2[j];
		  wO3[j] = o3[j];
	  }
	  cout << "Alternate data set installed\n";
	  break;

	  default:
		  break;

	  } 

  }
  return 0;
}
	
//**************************************************************
//                Sigmoid function neuron
//**************************************************************
double NeuronS(double xIn[], double wts[], int size)
{
// 1. NeuronS()receives the following parameters:
//        xIn[] is and array of values
//        wts is an array of weights
//        size is the number of elements in data arrays
// 
//     NeuronS() computes:
//     f(x) = the sum of products of xIn[] by wts[]
//     Returns the logistic function:
//     f(x) = 1/1+e^-f(x)
//
double sum = 0;         // f(x)
double result = 0;      // 1/1-e^-f(x) 
int i;

// Routine logic
for(i = 0; i < size; i++)
   sum += wts[i] * xIn[i];

result = 1/(1+exp(-sum));

return result;
}

//*****************************************************************
//                         Backpropagation
//*****************************************************************
void BackProp232(double inD[],     // Input array of 3 elements
				 double outD[],    // 3 elements with desired output 
				 double wHL1[],    // 2, 3-element hidden layer weigths
				 double wHL2[],
				 double wOL1[],    // 3, 2-element output layer weights
				 double wOL2[],
				 double wOL3[],
				 double eDif[])    // Storage for 3 error differences 
{
double eta = 1.0;      // Learning rate
double hidN[2];        // Storage for hidden neuron 
double outN[3];        // Storage for output layer results 
double e[3];           // Difference between actual and expected respose
double fDx[3];         // Derivative of output unit responses
double deltaO[3];      // error signals
double outError[2];    // hidden unit responsability for output error  
double fDhu[2];        // derivatives of hidden unit responses
double eSignal[2];     // error signal to use in backpropagation
double deltaZ1[3];     // factors for output unit weight corrections
double deltaZ2[3];
int j;

// Step 1:
// Presents the input data, stored in the global array named inD[]
// to the hidden layer neurons using weights tables wHL1[] and wHL2[].
hidN[0] = NeuronS(inD, wHL1, 3);
hidN[1] = NeuronS(inD, wHL2, 3);

// Step 2:
// Outputs from the hidden layer (in hidN[]) are presented to
// the 3 output layers using the weights tables wOL1[], wOL2[], and
// wOL3[] 
outN[0] = NeuronS(hidN, wOL1, 2); 
outN[1] = NeuronS(hidN, wOL2, 2);
outN[2] = NeuronS(hidN, wOL3, 2);

// Step 3:
// First calculate e = t - t^ for each output cell
// Then calculate the derivative of the output unit responses:
//    Logistic function:     f(x) = 1/1+e^-f(x)
//    Derivative:            fDx = f(x)[1-f(x)]
// Finally, calculate the error signal:
// deltaE = f'(x) * e
for(j = 0; j < 3; j++)
{
	e[j] = outD[j]-outN[j];
	fDx[j] = outN[j] * (1-outN[j]);
    deltaO[j] = fDx[j] * e[j];
}

// Step 4:
// Calculate error signal of output units:
// Multiply error signal of each output unit (in deltaO[]) by the
// output unit weights (in wO1[], wO2[], and wO3[]) 
outError[0] = wOL1[0] * deltaO[0] +\
              wOL2[0] * deltaO[1] +\
			  wOL3[0] * deltaO[2];
outError[1] = wOL1[1] * deltaO[0] +\
              wOL2[1] * deltaO[1] +\
			  wOL3[1] * deltaO[2];

// Step 5:
// Calculate error signal of hidden units
// Start with the derivative of the hidden unit responses:
fDhu[0] = hidN[0] * (1-hidN[0]);
fDhu[1] = hidN[1] * (1-hidN[1]);
eSignal[0] = fDhu[0] * outError[0];
eSignal[1] = fDhu[1] * outError[1];

// Step 6:
// Correct the output weights (in wO1[], wO2[], and wO3[])
// As follows:
//        Z = Z + deltaZ
//        deltaZ = eta * inD[] * eSignal[]
deltaZ1[0] = eta * (inD[0] * eSignal[0]);
deltaZ1[1] = eta * (inD[1] * eSignal[0]);
deltaZ1[2] = eta * (inD[2] * eSignal[0]);
//
deltaZ2[0] = eta * (inD[0] * eSignal[1]);
deltaZ2[1] = eta * (inD[1] * eSignal[1]);
deltaZ2[2] = eta * (inD[2] * eSignal[1]);

// Update weights in hidden layer unit
wHL1[0] = wHL1[0] + deltaZ1[0];
wHL1[1] = wHL1[1] + deltaZ1[1];
wHL1[2] = wHL1[2] + deltaZ1[2];

wHL2[0] = wHL2[0] + deltaZ2[0];
wHL2[1] = wHL2[1] + deltaZ2[1];
wHL2[2] = wHL2[2] + deltaZ2[2];

// Step 7:
// Update weights in output unit
wOL1[0] = wOL1[0] + (eta * (hidN[0] * deltaO[0]));
wOL1[1] = wOL1[1] + (eta * (hidN[1] * deltaO[0]));

wOL2[0] = wOL2[0] + (eta * (hidN[0] * deltaO[1]));
wOL2[1] = wOL2[1] + (eta * (hidN[1] * deltaO[1]));

wOL3[0] = wOL3[0] + (eta * (hidN[0] * deltaO[2]));
wOL3[1] = wOL3[1] + (eta * (hidN[1] * deltaO[2]));

// Return error differences
for(j = 0; j < 3; j++)
   eDif[j] = e[j];

return;
}

//****************************************************
//      execution function for a network with
//              3 input layers
//              2 hidden layers
//              3 output layers
//****************************************************
void Network232(double inD[],     // Input array of 3 elements
				double outD[],    // Array of 3 elements for results 
				double wHL1[],    // 2, 3-element hidden layer weigths sets
				double wHL2[],
				double wOL1[],    // 3, 2-element output layer weights sets
				double wOL2[],
				double wOL3[])
{
// Local variables
double outHL1[2];
	
// Function to execute a network with 3 neurons in the input layer,
// 2 in the hidden layer, and 3 in the output layer
// Step 1:
// Presents the input data, stored in the array named inD[]
// to the hidden layer neurons using weights tables wHL1[] and wHL2[].
outHL1[0] = NeuronS(inD, wHL1, 3);
outHL1[1] = NeuronS(inD, wHL2, 3);

// Step 2:
// Outputs from the hidden layer (in outHL1[]) are presented 
// to the 3 output layers using the weights in wOL1[], wOL2[],
// and wO3[] 
outD[0] = NeuronS(outHL1, wOL1, 2); 
outD[1] = NeuronS(outHL1, wOL2, 2);
outD[2] = NeuronS(outHL1, wOL3, 2);

// Done

return;
}

//**********************************************************************
//                      backpropagation trainer
//**********************************************************************
int BPTrain323(double inD[],     // Input array of 3 elements
	 		   double outD[],    // Array of 3 elements for results 
			   double wHL1[],    // 2, 3-element hidden layer weigths sets
			   double wHL2[],
			   double wOL1[],    // 3, 2-element output layer weights sets
			   double wOL2[],
			   double wOL3[],
			   double *eTotal)   // Error total returned to caller
{
// This function uses BackProp323() and Network323() to train the
// neural network until a LIMIT number of training epochs is reached
// or when a predefined error difference (ELIMIT) is achieved

int LIMIT = 5000;
int count = 0;
double ELIMIT = 0.0001;

int k;
double eData[3];      // Storage for error data returned by BackProp323()
double eSum;          // Error sum

while(LIMIT > count)
{
      BackProp232(inD, outD, wHL1, wHL2, wOL1, wOL2, wOL3, eData);
	  // Init error sum
      eSum = 0;
	  // Check for error limit
	  // First add error data
      for(k = 0; k < 3; k++)
		  eSum += (eData[k] * eData[k])/2;  
      // Test for limit
	  if(eSum < ELIMIT)
	  {
		  *eTotal = eSum;
		  return 1;
	  }
	  // Bump epoch counter
	  count++;
}
// Epoch limit was reached
*eTotal = eSum;
return 0;
}


