// ************************************************************************
// *******************************************
// Languge: C++
// Machine: Pentium PC under Windows 98/2000
// Environment: Visual C++ Version 6.0
// Program Type: Console application
// Filename: perceptron.cpp
// *******************************************
//
//*************************************************************************
//                          program description
//*************************************************************************
// 1. The function Percetron() was developed net in order to establish
//    a reference point for the LMS algorithm.
// 2. The function LSTrainer() uses a linear least squares filter (contained
//    in the getErrorSum() function) in order to train the Perceptron().
// 3. The function named Adaline() is a classifier net that uses the LMS
//    algorithm.
// 4. All three processing routines can be executed with a default and an
//    alternate data set. 
// 5. Two primitive neuron functions, named MPNeuron() and MPNeuron(),
//    provide basic processing for the net data 
//
//**************************************************************************
//                        core program functions
//**************************************************************************
//*************************
//       MPNeuron()
//************************* 
// The function MPNeuron() models the McCulloch-Pitts neuron.
// The function receives the following parameters:
//        x is the value of the independent variable
//        y is the value of the dependent variable
//        w[] is an array of three weights
//        bias is a constant defined as 1.0
// MPNeuron() computes:
//     f(x) = f((w[0] * x) + (w[1] * y) + (w[2] * bias))
//     Returns a discrete result, as follows:
//     if f(x) is positive, lpNeuron() returns 1
//     if f(x) is negative, lpNeuron() returns -1
//
//**************************
//       MPNeuron()
//**************************
// The function MPNeuron() produces a continuous output
// This operation is necessary for the Adeline()
// training algorithm
//    The function receives the following parameters:
//        x is the value of the independent variable
//        y is the value of the dependent variable
//        w[] is an array of three weights
//        bias is a constant defined as 1.0
// MPNeuron() computes:
//     f(x) = f((w[0] * x) + (w[1] * y) + (w[2] x bias))
// Returns a continuous result in a variable of type double
//
//****************************
//     Perceptron()
//****************************
// The function Perceptron() is a single layer classifier that
// uses the MPNeuron() function. Perceptron() is declared as
// follows:
///int Perceptron(double x[],      // array of x values
//		  double y[],      // array of y values
//		  int expected[],   // array of expected classification
//		  int dataSize,     // number of elements in arrays
//		  double weights[]) // array of 3 weights
// Terms:
// Array data is defined for dataSize entries of x[] and y[]
// array expected[] contains 1 or -1 according to the mapping
// of the corresponding values of x and y. 
// There is a limit of 500 tests in each training session
// Perceptron() returns 1 if succesful, 0 otherwise
// The training constant c is 0.2, therefore the learning
// increment is either +2c or -2c (2c = 0.4)

// Terminology:
//    the variables are labled x[k], y[k], expected[k] for each
//    kth iteration. 
//
// Learning algorithm:
// The perceptron is trained according to the following logic:
// 1. x[k], y[k], and w[] are fed into the MPNeuron() for each
//    iterations in the data set, from k = 0 to k = dataSize. 
//
// 2. At each iteration k, the result returned by MPNeuron() is
//    compared with the expected results (as stored in the expected[k]
//    array).
//    If the values match, execution proceeds at the next iteration
//    using the same weights. If not, then the following actions
//    take place:
//    a. If the difference between the expected and the actual
//       output of MPNeuron() is negative, then the learning
//       increment (li) = -0.4, otherwise li = 0.4
//    b. A new set of weights is created by adding or subtracting,
//       from the current weight, the product of the input[k]
//       times the learning increment, as follows:
//       if li < 0 then
//         new w[0] = old w[0] -(li * x[k])
//         new w[1] = old w[1] -(li * y[k])
//         new w[2] = old w[2] -(li * expected[k]
//       if li >= 0
//         new w[0] = old w[0] +(li * x[k])
//         new w[1] = old w[1] +(li * y[k])
//         new w[2] = old w[2] +(li * expected[k]
// 3. Execution continues until a valid function is found, that is,
//    one that correctly maps all the values in the data set, or until
//    number of iterations defined in the constant PLIMIT is reached.

//**********************************
// Adaline (Adaptive Linear Neuron)
//**********************************
// The function Adaline() is a single layer classifier that
// uses the MPNeuron() function. Adeline() net uses the LMS
// training algorithm. The function is declared as follows:
// int Adaline(double x[],   // array of x values
//		    double y[],      // array of y values
//		    int expected[],   // array of expected classification
//			int dataSize,     // number of elements in arrays
//			double wt[])      // array of 3 weights
// Terms:
// Array data is defined for dataSize entries of x[] and y[].
// Array expected[] contains 1 or -1 according to the mapping
// of the corresponding values of x and y. 
// There is a limit of 500 epochs in each training session
// Adaline() returns 1 if succesful, 0 otherwise
// Terminology:
//    the variables are labled x[k], y[k], expected[k] for each
//    kth epoch. 
//
// Training algorithm for Adaline():
// The Adaline is trained according to the following logic:
// 1. x[k], y[k], and w[] are fed into the MPNeuron() for each
//    iterations in the data set, from k = 0 to k = dataSize. 
//
// 2. At each jth iteration the result returned by MPNeuron() is
//    stored in the local int variable netVal according to the
//    following rule:
//      if netj is positive, netVal = 1
//      if netj is negative, netVal = -1
//    netVal is ten compared with the expected results (as stored in
//    the expected[k] array).
//    If the values match, execution proceeds at the next iteration
//    using the same weights.
//    If not, then the new weight is calculated according to the
//    formula:
//    wn = wo + Deltaw
//    where wn is the new weight
//          wo is the old weight,
//          Detalw is an increment applied to the old weight:
//       Deltaw = n(dj - netj)ij
//       where n is the learning rate (0.1)
//       dj is the desired output value (1 or -1) from expected[j]
//       netj is the value returned by the MPNeuron() for test case j
//       ij is the input vector
//
// 3. Execution continues until a valid function is found, that is,
//    one that correctly maps all the values in the data set, or until
//    number of iterations defined in the constant ALIMIT is reached.
//
//***************************
//      LSTrainer()
//***************************
// The function LSTrainer() randomly trains the Perceptron() by
// minimizing the error sum. The function is declated as follows:
// void LSTrainer(double x[],   // array of x values
//  			  double y[],      // array of y values
//	              int expected[],   // array of expected classification
//			      int dataSize)     // number of elements in arrays
// Function to train Perceptron() by minimizing the error sum
// Function logic:
// 1. Routines calculates an initial random set of weights in
//    the range 20 to -20.
//    Perceptron() is called with these weights. If Perceptron()
//    returns a valid equation, the error sum for this
//    equation is remembered.
// 2. A new set of random weights is calculated and Perceptron()
//    is again called with these new weights. If Perceptron
//    returns a valid equation, the error sum for this
//    new equation is compared to the error sum of the previous
//    one. The weights for the better error sum are preserved and
//    the other ones are discarded.
// 3. This is repeated 400 iterations after which the weights
//	  with the smallest error sum are stored in the global
//    array named bestWeights[]. The best error sum is calculated
//    and returned in a global variable
//
//****************************
//    GetErrorSum()
//****************************
// This function returns the error sum given the x and y values
// and the weights. The function is declated as follows:
// double GetErrorSum(double wt[],  // Array of weights
//                    double x[],  // Array of x values
//                    double y[],  // Array of y values
//                    int dataSize) // Number of elements in data set
// In the equation form ax + by + c = 0:
// w[0] = a
// w[1] = b
// w[2] = c
// Then, the slope-intercept form (y = mx + b) is determined from
// the weights, as follows:
//       y = -a/bx - a/c
// which makes:
//       m = -a/b
//       b = -a/c 
// Given the equation for the line, the error sum is determined by
// calculating each value of the dependent variable and subtracting this
// value from the corresponding one in the data arrays. These differences,
// called the residuals, are added and then squared in order to eliminate
// sign compensations. One-half of this result is returned as the error
// sum.
//            
// ************************************************************************
// ************************************************************************

#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#define PLIMIT 500                // Maximum number of training
                                  // sessions

// Prototypes
int MPNeuron(double, double, double[]);
double CNeuron(double, double, double []);
int Perceptron(double[], double[], int[], int, double[]);
int Adaline(double[], double[], int[], int, double[]);
void BFRTrainer(double[], double[], int[], int);
double GetErrorSum(double[], double[], double[], int);
int Random();
void InitRandom();

//******************************
//      Global variables
//******************************
// Fixed bias
double bias = 1.0;
double biasA = -1.0;
         
// Test data
double xData[] = {1.0, 9.4, 2.5, 8.0, 0.5, 7.9, 7.0, 2.8, 1.2, 7.8};
double yData[] = {1.0, 6.4, 2.1, 7.7, 2.2, 8.4, 7.0, 0.8, 3.0, 6.1};
int x3Data[] = {1, -1, 1, -1, 1, -1, -1, 1, 1, -1};
int dataCount = 10;

// Test data set 1
double xaData[] = {1.0, 9.4, 2.5, 8.0, 0.5, 7.9, 7.0, 2.8, 1.2, 7.8};
double yaData[] = {1.0, 6.4, 2.1, 7.7, 2.2, 8.4, 7.0, 0.8, 3.0, 6.1};
int x3aData[] = {1, -1, 1, -1, 1, -1, -1, 1, 1, -1};
int dataCounta = 10;

// Test data set 2
double xbData[] = {1.2, 6.2, 2.1, 2.2, 3.2, 8.0, 6.2, 2.0, 8.1};
double ybData[] = {0.5, 6.4, 1.0, 2.3, 3.0, 9.1, 8.0, 2.0, 7.9};
int x3bData[] = {1, -1, 1, 1, 1, -1, -1, 1, -1};
int dataCountb = 9;

// Arrays for weights
double w[] = {0.75, 0.5, -0.6};       // Initial random weights      
double bestWeights[] = {0.0, 0.0, 0.0};
double w1[] = {1.0, 1.0, -1.0};

// Error sums
double thisESum = 0;
double oldESum = 0;
double bestESum = 0;

// Iteration and test counters
int changeCnt = 0;       // Iterations
int testCnt = 0;         // Tests performed

//*********************************************
//                 main()
//*********************************************
main()
{


  int option = 1;
  int j;

  while(option > 0)
  {

    // Menu for neuron testing
	cout << "\nTESTING PROGRAM OPERATIONS\n";
	cout << "   1 Find weights with Perceptron\n";
	cout << "   2 Find weights using brute-force-random method\n";
	cout << "     with linear least square filter\n";      
	cout << "   3 Find weigths using Adaline (LMS algorithm)\n";
	cout << "   4 Change to alterante data set\n";
	cout << "   5 Change to default data set\n";
	cout << "   0 to end\n";
    cout << "Enter option number: ";
	cin >> option;
	cout << "\n\n";

      switch(option)
      {
	  case 1:
		cout << "Finding weights using perceptron\n";
    	// Initalize random data and variables
		w[0] = 0.75;
		w[1] = 0.5;
		w[2] = -0.6;

        if(Perceptron(xData, yData, x3Data, dataCount, w) == 0)
			cout << "No Valid function found";
		else
		{
			cout << "Valid function found";
			cout << "\nNumber of weight changes: " << changeCnt;
			cout << "\nNumber of test iterations: " << testCnt;
			cout << "\nWeights:";
			cout << "   W1: " << w[0] << "   W2: " << w[1];
			cout << "   W3: " << w[2] << "\n";
			// Calculate and display error sum for these weights
			thisESum = GetErrorSum(w, xData, yData, dataCount);
			cout << "Error sum: " << thisESum << "\n\n";
		}
		break;
   
	  case 2:
		// Testing linear least square trainer
		cout << "Finding weights by least squares\n";
		cout << "\nWeights found:";
		BFRTrainer(xData, yData, x3Data, dataCount);
		cout << "\nW1: " << bestWeights[0];
		cout << "   W2: " << bestWeights[1];
		cout << "   W3: " << bestWeights[2];
		cout << "\nError sum: " << bestESum; 
		cout << "\n\n";
		break;

	  case 3:
		cout << "Finding weights using Adaline\n";
    	// Initalize random data and variables
		w[0] = 0.75;
		w[1] = 0.5;
		w[2] = -0.6;

        if(Adaline(xData, yData, x3Data, dataCount, w) == 0)
			cout << "No Valid function found";
		else
		{
			cout << "Valid function found";
			cout << "\nNumber of weight changes: " << changeCnt;
			cout << "\nNumber of test iterations: " << testCnt;
			cout << "\nWeights:";
			cout << "   W1: " << w[0] << "   W2: " << w[1];
			cout << "   W3: " << w[2] << "\n";
			// Calculate and display error sum for these weights
			thisESum = GetErrorSum(w, xData, yData, dataCount);
			cout << "Error sum: " << thisESum << "\n\n";
		}
		break;


	  case 4:
	    // Change to alternate data set
       for(j = 0; j < dataCountb; j++)
	   {
           xData[j] = xbData[j];
		   yData[j] = ybData[j];
           x3Data[j] = x3bData[j];
	   }
       dataCount = dataCountb;
	   testCnt = 0;
	   changeCnt = 0;
	   cout << "\nAlternate data set enabled\n";
	   break;

	  case 5:
	    // Change to default data set
	  for(j = 0; j < dataCounta; j++)
	   {
           xData[j] = xaData[j];
		   yData[j] = yaData[j];
           x3Data[j] = x3aData[j];
	   }
       dataCount = dataCounta;
	   testCnt = 0;
	   changeCnt = 0;
	   cout << "\nDefault data set enabled\n";
	   break;
	   
	  }
  }
  return 0;
}
	
//**************************************************************
//                 McCulloch-Pitts model neuron
//**************************************************************
int MPNeuron(double x, double y, double wn[])
{
// 1. The function MPNeuron() models the McCulloch-Pitts neuron.
//    The function receives the following parameters:
//        x is the value of the independent variable
//        y is the value of the dependent variable
//        w[] is an array of three weights
//        bias is a constant defined as 1.0
//     MPNeuron() computes:
//     f(x) = f((w[0] * x) + (w[1] * y) + (w[2] x bias))
//     Returns:
//     if f(x) is positive, lpNeuron() returns 1
//     if f(x) is negative, lpNeuron() returns -1
//
double result = 0;

// Routine logic, per McCulloch-Pitts model
result = (wn[0] * x) + (wn[1] * y) + (wn[2] * bias);
if(result < 0)
	return -1;
else
	return 1;
}

//**************************************************************
//                Continuous output neuron
//**************************************************************
double CNeuron(double x, double y, double wn[])
{
// 1. CNeuron()receives the following paramenters:
//        x is the value of the independent variable
//        y is the value of the dependent variable
//        w[] is an array of three weights
//        bias is a constant defined as 1.0
//     CNeuron() computes:
//     f(x) = f((w[0] * x) + (w[1] * y) + (w[2] x bias))
//     Returns:
//     f(x) in a variable of type double
//
double result = 0;

// Routine logic
result = (wn[0] * x) + (wn[1] * y) + (wn[2] * bias);
return result;
}

//**************************************************************
//        Perceptron (uses a hard limiting bipolar
//               linear threshold function)
//**************************************************************
int Perceptron(double x[],      // array of x values
			   double y[],      // array of y values
			   int expected[],   // array of expected classification
			   int dataSize,     // number of elements in arrays
			   double weights[]) // array of 3 weights
{
// The function Perceptron() is a single layer classifier that
// uses the MPNeuron() function. 
// Terms:
// Array data is defined for dataSize entries of x[] and y[].
// Array expected[] contains 1 or -1 according to the mapping
// of the corresponding values of x and y. 
// There is a limit of 500 tests in each training session
// Perceptron() returns 1 if succesful, 0 otherwise
// The training constant c is 0.2, therefore the learning
// increment is either +2c or -2c (2c = 0.4)
// Terminology:
//    the variables are labled x[k], y[k], expected[k] for each
//    kth iteration. 
//
// Training algorithm for Perceptron():
// The perceptron is trained according to the following logic:
// 1. x[k], y[k], and w[] are fed into the MPNeuron() for each
//    iterations in the data set, from k = 0 to k = dataSize. 
//
// 2. At each iteration k, the result returned by MPNeuron() is
//    compared with the expected results (as stored in the expected[k]
//    array).
//    If the values match, execution proceeds at the next iteration
//    using the same weights. If not, then the following actions
//    take place:
//    a. If the difference between the expected and the actual
//       output of MPNeuron() is negative, then the learning
//       increment (li) = -0.4, otherwise li = 0.4
//    b. A new set of weights is created by adding or subtracting,
//       from the current weight, the product of the input[k]
//       times the learning increment, as follows:
//       if li < 0 then
//         new w[0] = old w[0] -(li * x[k])
//         new w[1] = old w[1] -(li * y[k])
//         new w[2] = old w[2] -(li * expected[k]
//       if li >= 0
//         new w[0] = old w[0] +(li * x[k])
//         new w[1] = old w[1] +(li * y[k])
//         new w[2] = old w[2] +(li * expected[k]
// 3. Execution continues until a valid function is found, that is,
//    one that correctly maps all the values in the data set, or until
//    number of iterations defined in the constant PLIMIT is reached.

	
// Local variables
double li = 0.4;         // Learning increment is +/-0.4
int k = 0;               // Iteration counter
int thisTest = 0;        // Test counter
int thisResult = 0;      // Value returned by lbNeuron()
int thisDif = 0;         // expected[k] - thisResult
int wc = 0;              // Weight change flag

// Clear global variables
changeCnt = 0;
testCnt = 0;

while(thisTest < PLIMIT)
 {
 
 
  for (k = 0; k < dataSize; k++)
  {
	  if(k == 0)
		  wc = 0;
	  thisResult = MPNeuron(x[k], y[k], weights);
	  thisDif = expected[k] - thisResult;
	  // Test for expected result, if not change weights
	  if(thisResult != expected[k])
	  {
		   // ASSERT: weights need changing
		   if(thisDif < 0)
           {
           weights[0] = weights[0] - (li * x[k]);
           weights[1] = weights[1] - (li * y[k]);		   
           weights[2] = weights[2] - (li * bias);
		   }
		   else
		   {
           weights[0] = weights[0] + (li * x[k]);
           weights[1] = weights[1] + (li * y[k]);		   
           weights[2] = weights[2] + (li * bias);
		   }
           // Set weight change flag
		   wc = 1;
		   // Count weight changes
		   changeCnt++;
	  }

	} // End of for loop
      // At this point one full iteration with has taken place
      // Test for now weight change during this iteration. This
      // indicates a successful function.
  	  if(wc == 0)
	    return 1;        // Return success!
	  // if not
	  testCnt++;            // Increment global test counter
 
 } // end of while loop
   // At this point the limit in the number of iterations has
   // been reached without finding a valid equation
   return 0; 
}

//**************************************************************
//             Adaline (Adaptive Linear Neuron)
//**************************************************************
int Adaline(double x[],      // array of x values
		    double y[],      // array of y values
		    int expected[],   // array of expected classification
			int dataSize,     // number of elements in arrays
			double wt[])      // array of 3 weights
{
// The function Adaline() is a single layer classifier that
// uses the MPNeuron() function. Adeline() net uses the LMS
// training algorithm.
// Terms:
// Array data is defined for dataSize entries of x[] and y[].
// Array expected[] contains 1 or -1 according to the mapping
// of the corresponding values of x and y. 
// There is a limit of 500 epochs in each training session
// Adaline() returns 1 if succesful, 0 otherwise
// Terminology:
//    the variables are labled x[k], y[k], expected[k] for each
//    kth epoch. 
//
// Training algorithm for Adaline():
// The Adaline is trained according to the following logic:
// 1. x[k], y[k], and w[] are fed into the MPNeuron() for each
//    iterations in the data set, from k = 0 to k = dataSize. 
//
// 2. At each jth iteration the result returned by MPNeuron() is
//    stored in the local int variable netVal according to the
//    following rule:
//      if netj is positive, netVal = 1
//      if netj is negative, netVal = -1
//    netVal is ten compared with the expected results (as stored in
//    the expected[k] array).
//    If the values match, execution proceeds at the next iteration
//    using the same weights.
//    If not, then the new weight is calculated according to the
//    formula:
//    wn = wo + Deltaw
//    where wn is the new weight
//          wo is the old weight,
//          Detalw is an increment applied to the old weight:
//       Deltaw = n(dj - netj)ij
//       where n is the learning rate (0.1)
//       dj is the desired output value (1 or -1) from expected[j]
//       netj is the value returned by the MPNeuron() for test case j
//       ij is the input vector
//
// 3. Execution continues until a valid function is found, that is,
//    one that correctly maps all the values in the data set, or until
//    number of iterations defined in the constant ALIMIT is reached.
	
// Local variables

int ALIMIT = 500;
double n = 0.1;          // learning increment
int k = 0;               // Iteration counter
int thisTest = 0;        // Test counter
double thisNet = 0.0;    // Value returned by MPNeuron()
int netVal = 1;          // Binary result of net 
int wc = 0;              // Weight change flag
double deltaW = 0.0;     // Increment for weight change

// Clear global variables
changeCnt = 0;
testCnt = 0;

while(thisTest < ALIMIT)
 {

  // Start of epoch	
  for (k = 0; k < dataSize; k++)
  {
	  if(k == 0)
		  wc = 0;

	  thisNet = CNeuron(x[k], y[k], wt);
	  // Convert to binary result 
	  if(thisNet < 0)
		netVal = -1;
      else
	    netVal = 1;

	  // Test for expected result, if not change weights
	  if(netVal != expected[k])
	  {
		  // ASSERT: weights need changing
          // Calculate deltaw = n(dj - netj)ij
		  // For the first input vector
          deltaW = n * x[k] * (expected[k] - thisNet);
          wt[0] = wt[0] + deltaW;
		  // For the second input vector
          deltaW = n * y[k] * (expected[k] - thisNet);
          wt[1] = wt[1] + deltaW;
		  // For the bias vector
          deltaW = n * (expected[k] - thisNet);
          wt[2] = wt[2] + deltaW;
	      // Set weight change flag
          wc = 1;
	      // Count weight changes
	      changeCnt++;
	   }
	
	} // End of epoch
      // At this point one epoch has taken place.
      // Test for no weight change during epoch. This indicates a
      // valid function.
 	  if(wc == 0)
	    return 1;        // Return success!
	  // if not
	  testCnt++;            // Increment global test counter
      thisTest++;           // Iteration counter
} // end of while loop
   // At this point the limit in the number of iterations has
   // been reached without finding a valid equation
   return 0; 
}


//**************************************************************
//            brute-force random weights trainer
//**************************************************************
void BFRTrainer(double x[],    // array of x values
			  double y[],      // array of y values
			  int expected[],   // array of expected classification
			  int dataSize)     // number of elements in arrays

{
// The function BFRTrainer() trains the Perceptron() by means of a
// brute-force algorithm in which the weights are selected at
// random during each iteration. For each iteration that produces
// a valid equation the error sum is calculated and stored.
// Calculations are repeated 400 times and the weights with the
// smallest error sum are returned.
// The function is declared as follows:
// void BFRTrainer(double x[],   // array of x values
//  	 		  double y[],      // array of y values
//	              int expected[],   // array of expected classification
//			      int dataSize)     // number of elements in arrays
// Function logic:
// 1. Routines calculates an initial random set of weights in
//    the range 20 to -20.
//    Perceptron() is called with these weights. If Perceptron()
//    returns a valid equation, the error sum for this
//    equation is remembered.
// 2. A new set of random weights is calculated and Perceptron()
//    is again called with these new weights. If Perceptron
//    returns a valid equation, the error sum for this
//    new equation is compared to the error sum of the previous
//    one. The weights for the better error sum are preserved and
//    the other ones are discarded.
// 3. This is repeated 400 iterations after which the weights
//	  with the smallest error sum are stored in the global
//    array named bestWeights[]. The best error sum is calculated
//    and returned in a global variable

int limit = 400;             // Number of iterations
int counter = 0 ;

// Set initial weights
double patWeights[] = {0.0, 0.0, 0.0};
double weights[] = {0.0, 0.0, 0.0};

oldESum = 100000;           // Start with large error sum

// initialize random number'
InitRandom();

// Set initial pattern weights at random
patWeights[0] = Random();
patWeights[1] = Random();
patWeights[2] = Random();

while(counter < limit)
{
	// Set weights to local weight pattern
	weights[0] = patWeights[0];
	weights[1] = patWeights[1];
	weights[2] = patWeights[2];
	if(Perceptron(x, y, expected, 10, weights))
	{
		thisESum = GetErrorSum(weights, x, y, dataSize);
		// Store best weights
		if(thisESum < oldESum)
		{
			bestWeights[0] = weights[0];
			bestWeights[1] = weights[1];
			bestWeights[2] = weights[2];
			oldESum = thisESum;
		}
	}
    // Reset pattern weights using random numbers
	patWeights[0] = Random();
    patWeights[1] = Random();
	patWeights[2] = Random();
	
    counter++;
}
  // Calculate error sum and store in global variable
  bestESum = GetErrorSum(bestWeights, xData, yData, dataSize);
  return;
}

//*********************************************
//           error sum calculation
//*********************************************
double GetErrorSum(double wt[],    // Array of weights
				   double x[],     // Array of x values
				   double y[],     // Array of y values
				   int dataSize)   // Number of data elements
{
// This function returns the error sum given the x and y values
// and the weights.
// In the equation form ax + by + c = 0:
// w[0] = a
// w[1] = b
// w[2] = c
// Then, the slope-intercept form (y = mx + b) is determined from
// the weights, as follows:
//       y = -a/bx - a/c
// which makes:
//       m = -a/b
//       b = -a/c 
// Given the equation for the line, the error sum is determined by
// calculating each value of the dependent variable and subtracting this
// value from the corresponding one in the data arrays. These differences,
// called the residuals, are added and then squared to eliminate sign
// compensations. One-half of this result is returned as the error sum.

double m = 0;        // Slope
double b = 0;        // Y-intercept
int j = 0;
double newY = 0;     // Temporary storage for y
double res = 0;      // Accumulator for residuals
double errSum = 0;   // Error sum

// First calculate m and b
m = -wt[0]/wt[1];
b = -(wt[2]/wt[1]);
// Use m and b to calculate error sum
	for(j = 0; j < dataSize; j++)
    {
		newY = (m * x[j]) + b;
        res = y[j] - newY;
		errSum = errSum + (res * res);
	}
	return errSum/2;
}

//*************************************************************
//             random number generator functions
//*************************************************************
void InitRandom()
// Initialize random number for generator
{
	srand((unsigned)time(NULL));
}

int Random()
{
int num;
// Generate a random numb in the range 20 to -20
num = rand()%40;
	if(num > 20)
		num = (num - 20)*-1;
return num;
}
