/**********************************************************************
  PROJECT: Hello Windows
  Source: WinHello.cpp
  Chapter reference: 18
***********************************************************************
  Description: 
    A Hello Windows program
  Topics:
  	1. Program icon 
    2. Displays a greeting message on the client area
***********************************************************************/

#include <windows.h>
#include "resource.h"		// Load resource file for icon

// Predeclaration of the window procedure
LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM) ;

//****************************
//         WinMain
//****************************
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    PSTR szCmdLine, int iCmdShow)
{
     static char szAppName[] = "Demo" ; // Class name
     HWND        hwnd ;
     MSG         msg ;

     // Defining a structure of type WNDCLASSEX
	 //   The program icon is loaded in the hIcon and hIconSm
	 //   structure members
     WNDCLASSEX  wndclass ;
     wndclass.cbSize        = sizeof (wndclass) ;
     wndclass.style         = CS_HREDRAW | CS_VREDRAW ;
     wndclass.lpfnWndProc   = WndProc ;
     wndclass.cbClsExtra    = 0 ;
     wndclass.cbWndExtra    = 0 ;
     wndclass.hInstance     = hInstance ;
     wndclass.hIcon         = LoadImage(hInstance,
		                                MAKEINTRESOURCE(IDI_ICON1),
										IMAGE_ICON,
										32, 32,
										LR_DEFAULTCOLOR) ;
     wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
     wndclass.hbrBackground = (HBRUSH) GetStockObject
                              (WHITE_BRUSH) ;
     wndclass.lpszMenuName  = NULL ;
     wndclass.lpszClassName = szAppName ; 
     wndclass.hIconSm       = LoadImage(hInstance,
		                                MAKEINTRESOURCE(IDI_ICON1),
										IMAGE_ICON,
										16, 16,
										LR_DEFAULTCOLOR) ;
     // Registering the structure wmdclass
     RegisterClassEx (&wndclass) ;

     // CreateWindow()
     hwnd = CreateWindowEx (
		    WS_EX_LEFT,              // Left aligned (default)
		    szAppName,		         // pointer to class name
           "Hello Windows",          // window caption
            WS_OVERLAPPEDWINDOW,     // window style
            CW_USEDEFAULT,           // initial x position
            CW_USEDEFAULT,           // initial y position
            CW_USEDEFAULT,           // initial x size
            CW_USEDEFAULT,           // initial y size
            NULL,                    // parent window handle
            NULL,                    // window menu handle
            hInstance,               // program instance handle
            NULL) ;                  // creation parameters

     ShowWindow (hwnd, iCmdShow) ;
     UpdateWindow (hwnd) ;

     // Message loop
     while (GetMessage (&msg, NULL, 0, 0))
          {
          TranslateMessage (&msg) ;
          DispatchMessage (&msg) ;
          }
     return msg.wParam ;
}
//****************************
//    Windows Procedure
//****************************
LRESULT CALLBACK WndProc (HWND hwnd, UINT iMsg, WPARAM wParam,
LPARAM lParam)
{

     PAINTSTRUCT ps ;
     RECT        rect ;
	 HDC         hdc;

     switch (iMsg)
         {
     // Windows message processing
	      case WM_CREATE:
			   return 0;

          case WM_PAINT :
               hdc = BeginPaint (hwnd, &ps) ;
               GetClientRect (hwnd, &rect) ;
               
			   // Display message in the client area
			   DrawText (hdc,
				        "Hello World from Windows",
				         -1,
						 &rect,
   					     DT_SINGLELINE | DT_CENTER | DT_VCENTER);
			   

               EndPaint (hwnd, &ps) ;
               return 0 ;
  
          // End of program execution
          case WM_DESTROY :
               PostQuitMessage (0) ;
               return 0 ;
          }
     return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
}
