/**********************************************************************
  PROJECT: Mouse Demo
  Source: mou_demo.cpp
  Chapter reference: 5
***********************************************************************
  Description:
    Mouse and cursor handling 
  Topics:
    1. Using the mouse buttons to change the cursor
	2. Displaying the mouse cursor position
	3. Creating and using a custom cursor
***********************************************************************/


#include <windows.h>   // standard Windows header
#include <stdlib.h>
#include "resource.h"

// Global handle to instance
HINSTANCE  pInstance;

// Prototype for WndProc
LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM) ;

//************************
//       WinMain()
//************************
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    PSTR szCmdLine, int iCmdShow) {


     static char szAppName[] = "Demos" ;
     HWND        hwnd ;
     MSG         msg ;
     WNDCLASSEX  wndclass ;

     wndclass.cbSize        = sizeof (wndclass) ;
     wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC ;
     wndclass.lpfnWndProc   = WndProc ;
     wndclass.cbClsExtra    = 0 ;
     wndclass.cbWndExtra    = 0 ;
     wndclass.hInstance     = hInstance ;
     wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION) ;
     wndclass.hCursor       = NULL;
     wndclass.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH) ;
     wndclass.lpszMenuName  = NULL ;
     wndclass.lpszClassName = szAppName ;
     wndclass.hIconSm       = LoadIcon (NULL, IDI_APPLICATION) ;

	 // Store instance in global variable
	 pInstance = hInstance;

	// Register the wmdclass structure
     RegisterClassEx (&wndclass) ;

	 hwnd = CreateWindow (szAppName,
		  "Mouse and Cursor Demo",          // Window caption
           WS_OVERLAPPEDWINDOW,				// style
		   CW_USEDEFAULT,					// initial x
		   CW_USEDEFAULT,					// initial y
		   CW_USEDEFAULT,					// initial x size
		   CW_USEDEFAULT,					// initial y size
		   NULL,							// parent window handle
		   NULL,							// window menu handle
		   hInstance,						// program handle
		   NULL) ;							// creation parameters

     ShowWindow (hwnd, iCmdShow) ;
     UpdateWindow (hwnd) ;

//************************
//    Message loop
//************************
     while (GetMessage (&msg, NULL, 0, 0))
          {
          TranslateMessage (&msg) ;
          DispatchMessage (&msg) ;
          }
     return msg.wParam ;
     }


//************************
//    Windows Procedure
//************************
LRESULT CALLBACK WndProc (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	char GreetStr[] = {"Click left mouse button to change cursor\n"
		"Click right button for custom cursor"};
	char CurXStr[] = {"x =           "};
	char CurXBlk[] = {"              "};
	char CurYStr[] = {"y =           "};


 	 static int  cxChar, cyChar ;        // Character dimensions
	 static int  cxClient, cyClient;     // Client area parameters
	 static int  cursorX, cursorY;       // Cursor position
     static HDC         hdc ;            // handle to private DC

	 // Structures and handles
     PAINTSTRUCT ps;
	 TEXTMETRIC  tm;
	 HFONT       hFont;                 
	 RECT		 textRect;                // RECT-type

	 // Cursor handles and controls
	 HCURSOR     aCursor;
	 static int  curNum = 0;


	 switch (iMsg) {

		 case WM_CREATE :
               hdc = GetDC (hwnd) ;

			   // Create a logical font
			   hFont = CreateFont (
					   30 * GetDeviceCaps (hdc, LOGPIXELSY) / 72, // height
					   0,                  // width
					   0,                  // escapement angle
					   0,                  // orientation angle
					   FW_NORMAL,          // weight
					   0,                  // not italics
					   0,                  // not underlined
					   0,                  // not strikedout
					   DEFAULT_CHARSET,    // character set
					   OUT_DEFAULT_PRECIS, // precision
					   CLIP_DEFAULT_PRECIS,// clipping precision
					   PROOF_QUALITY,      // quality
					   DEFAULT_PITCH | FF_DONTCARE, // pitch and family
					   "Courier");         // typeface name

			   SelectObject( hdc, hFont);  // Select font into DC

               // Text metrics are obtained after font is selected 
               GetTextMetrics (hdc, &tm) ;
			   
               // Calculate and store character dimensions
               cxChar = tm.tmAveCharWidth ;
               cyChar = tm.tmHeight + tm.tmExternalLeading ;
			   // Store size of client area
			   cxClient = LOWORD(lParam);
			   cyClient = HIWORD(lParam);

	           // Select and display a cursor
			   aCursor = LoadCursor(NULL, IDC_UPARROW);
			   SetCursor(aCursor);
		   
			   return 0 ;

		 case WM_SIZE:
			   // Determine and store size of client area
			   cxClient = LOWORD(lParam);
			   cyClient = HIWORD(lParam);
			   return 0;
         
		  case WM_PAINT :
               BeginPaint (hwnd, &ps) ;
	          // Initialize rectangle structure
			  SetRect (&textRect,                // address of structure
			           2 * cxChar,               // x for start
					   cyChar,                   // y for start
					   cxClient -(2 * cxChar),   // x for end
					   cyClient);                // y for end

			  // Display greeting message
			  DrawText( hdc, GreetStr, -1, &textRect,
			              DT_LEFT | DT_WORDBREAK);
              EndPaint (hwnd, &ps);

              return 0 ;

           case WM_LBUTTONDOWN:
            
              curNum++;			// bump to next cursor

			  switch (curNum) {
				case 1:
					aCursor = LoadCursor(NULL, IDC_WAIT);
					SetCursor(aCursor);
					break;
                case 2:
					aCursor = LoadCursor(NULL, IDC_APPSTARTING);
					SetCursor(aCursor);
					break;
				case 3:
					aCursor = LoadCursor(NULL, IDC_CROSS);
					SetCursor(aCursor);
					break;
                case 4:
					aCursor = LoadCursor(NULL, IDC_HELP);
					SetCursor(aCursor);
					break;
				case 5:
					aCursor = LoadCursor(NULL, IDC_IBEAM);
					SetCursor(aCursor);
					break;
				case 6:
					aCursor = LoadCursor(NULL, IDC_NO);
					SetCursor(aCursor);
					break;
                case 7:
					aCursor = LoadCursor(NULL, IDC_SIZEALL);
					SetCursor(aCursor);
					break;
				case 8:
					aCursor = LoadCursor(NULL, IDC_SIZENESW);
					SetCursor(aCursor);
					break;
				case 9:
					aCursor = LoadCursor(NULL, IDC_SIZENS);
					SetCursor(aCursor);
					break;
                case 10:
					aCursor = LoadCursor(NULL, IDC_SIZENWSE);
					SetCursor(aCursor);
					break;
				case 11:
					aCursor = LoadCursor(NULL, IDC_SIZEWE);
					SetCursor(aCursor);
					break;
				case 12:
					aCursor = LoadCursor(NULL, IDC_UPARROW);
					SetCursor(aCursor);
					curNum = 0;
					break;
			  }
               		  
			  return 0;
  
		  case WM_RBUTTONDOWN:
					aCursor = LoadCursor(pInstance, 
					   MAKEINTRESOURCE(IDC_CURSOR1));
					SetCursor(aCursor);

               return 0; 
		
		  case WM_MOUSEMOVE:
			  cursorX = LOWORD(lParam);
			  cursorY = HIWORD(lParam);

              // Convert integer to ASCII string
			  _itoa(cursorX, CurXStr + 4, 10);
			  _itoa(cursorY, CurYStr + 4, 10);

              // Display x cordinate of mouse cursor
	          // Initialize rectangle structure
			  SetRect (&textRect,                // address of structure
			           2 * cxChar,              // x for start
					   3 * cyChar,               // y for start
					   cxClient -(2 * cxChar),   // x for end
					   cyClient);                // y for end

			  // Display x coordinate
			  // Erase old value
			  DrawText( hdc, CurXBlk, -1, &textRect,
			              DT_LEFT | DT_WORDBREAK);
			  // Display new string
              DrawText( hdc, CurXStr, -1, &textRect,
			              DT_LEFT | DT_WORDBREAK);

              // Display y cordinate of mouse cursor
	          // Initialize rectangle structure
			  SetRect (&textRect,                // address of structure
			           2 * cxChar,               // x for start
					   4 * cyChar,               // y for start
					   cxClient -(2 * cxChar),   // x for end
					   cyClient);                // y for end

			  // Erase old value
			  DrawText( hdc, CurXBlk, -1, &textRect,
			              DT_LEFT | DT_WORDBREAK);

			  // Display new string
              DrawText( hdc, CurYStr, -1, &textRect,
			              DT_LEFT | DT_WORDBREAK);

			  return 0;

		  case WM_SETCURSOR:
			  return 1;
 
           
          case WM_DESTROY :
               PostQuitMessage (0) ;
               return 0 ;
          }

     return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
     }
