/**********************************************************************
  PROJECT: Child Window Demo
  Source: chi_demo.cpp
  Chapter reference: 22
***********************************************************************
  Description:
    Creates and displays a child window with independent DC
	and message processing when left mouse button is pressed. 
 
  Topics:
    1. Child and parent windows
	2. Window-specific message processing
***********************************************************************/

#include <windows.h>


// Prototypes
LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM) ;
LRESULT CALLBACK ChildWndProc (HWND, UINT, WPARAM, LPARAM) ;

//**********************************
//          global data
// *********************************
HWND         hwnd, hChild ;
HINSTANCE    pInstance;
HDC          hdc, childDc;

//***********************************
//            WinMain()
//***********************************
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    PSTR szCmdLine, int iCmdShow)
     {
     static char szAppName[] = "Demo" ;

     MSG         msg ;
     WNDCLASSEX  wndclass ;

     wndclass.cbSize        = sizeof (wndclass) ;
     wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
     wndclass.lpfnWndProc   = WndProc ;
     wndclass.cbClsExtra    = 0 ;
     wndclass.cbWndExtra    = sizeof (WORD);
     wndclass.hInstance     = hInstance ;
     wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION) ;
     wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
     wndclass.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH) ;
     wndclass.lpszMenuName  = NULL;
     wndclass.lpszClassName = szAppName ;
     wndclass.hIconSm       = LoadIcon (NULL, IDI_APPLICATION) ;

     RegisterClassEx (&wndclass) ;

     pInstance = hInstance;


     hwnd = CreateWindow (szAppName,
                "Child Window Demo",
                WS_OVERLAPPEDWINDOW,
                CW_USEDEFAULT, CW_USEDEFAULT,
                CW_USEDEFAULT, CW_USEDEFAULT,
                NULL, NULL, hInstance, NULL) ;

     ShowWindow (hwnd, iCmdShow) ;
     UpdateWindow (hwnd) ;

     while (GetMessage (&msg, NULL, 0, 0))
          {
          TranslateMessage (&msg) ;
          DispatchMessage (&msg) ;
          }
     return msg.wParam ;
     }



//*********************************
//        Windows Procedure
// ********************************
LRESULT CALLBACK WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
     {
   
     PAINTSTRUCT ps ;
     WNDCLASSEX  chiclass ;

   switch (iMsg)
          {
         case WM_CREATE:
           hdc = GetDC (hwnd) ;
           // The system monospaced font is selected
           SelectObject (hdc, GetStockObject (SYSTEM_FIXED_FONT)) ;
		   // Create a child window
				chiclass.cbSize        = sizeof (chiclass) ;
				chiclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
				chiclass.lpfnWndProc   = ChildWndProc ;
				chiclass.cbClsExtra    = 0 ;
				chiclass.cbWndExtra    = 0 ;
				chiclass.hInstance     = pInstance ;
				chiclass.hIcon         = NULL;
				chiclass.hCursor       = LoadCursor (NULL, IDC_CROSS) ;
				chiclass.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH) ;
				chiclass.lpszMenuName  = NULL;
				chiclass.lpszClassName = "ChildWindow" ;
				chiclass.hIconSm       = NULL;

			RegisterClassEx (&chiclass) ;
            hChild = CreateWindow ("ChildWindow",
				  "A Child Window",  // caption
                  WS_CHILD | WS_OVERLAPPEDWINDOW ,
                  //WS_CHILD | WS_VISIBLE | WS_OVERLAPPEDWINDOW ,
                  40, 40,          // x and y of window location
                  400, 100,        // x and y of window size
                  hwnd,            // handle to the parent window 
                  (HMENU) 1001,    // child window designation
                  pInstance,       // program instance
                  NULL) ;

			 return 0 ;

        case WM_PAINT:
           BeginPaint (hwnd, &ps) ;

		   // Display child window
		   UpdateWindow(hChild);
		   ShowWindow(hChild, SW_SHOWNORMAL);

           EndPaint (hwnd, &ps);
           return 0 ;

           case	WM_LBUTTONDOWN:
			   // Display text in parent and erase text in child
			   TextOut(hdc, 10, 10, "Mouse action in PARENT", 22);
			   TextOut(childDc, 10, 10, "                      ", 22);			   
			   return 0;

         case WM_DESTROY:
          PostQuitMessage (0) ;
           return 0 ;
          }

     return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
}


//**********************************
//    child windows procedure
//**********************************
LRESULT CALLBACK ChildWndProc (HWND hChild, UINT iMsg, WPARAM wParam,
							   LPARAM lParam) {
	switch (iMsg) {

	   case WM_CREATE:
          childDc = GetDC(hChild);
          SelectObject (childDc, GetStockObject (SYSTEM_FIXED_FONT)) ;
          return 0;
			 
       case	WM_LBUTTONDOWN:
          // Display message in child and erase text in parent
	      TextOut(childDc, 10, 10, "Mouse action in child ", 22);
	      TextOut(hdc, 10, 10, "                      ", 22);			   
	      return 0;

	   case WM_DESTROY:
          return 0;
	 }
	    return DefWindowProc (hChild, iMsg, wParam, lParam) ;
}

