/**********************************************************************
  PROJECT: Dialog Box Demo
  Source: dia_demo.cpp
  Chapter reference: 22
***********************************************************************
  Description:
    Dialog box demonstration

  Topics:
    1. Using the MessageBox() function
	2. Displaying a bitmap in a dialog box
	3. Buttons in a dialog box
	4. Dialog box with listbox
    5. Modeless dialog box
	6. Common dialog boxes for font and color selection
***********************************************************************/

#include <windows.h>
#include "resource.h"
#include <commdlg.h>

// Prototypes for WndProc and Dialog procedures
LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM) ;
BOOL    CALLBACK AboutDlgProc(HWND, UINT, WPARAM, LPARAM);
BOOL    CALLBACK ButtsDlgProc(HWND, UINT, WPARAM, LPARAM);
BOOL    CALLBACK ListbDlgProc(HWND, UINT, WPARAM, LPARAM);
BOOL    CALLBACK TestDlgProc (HWND, UINT, WPARAM, LPARAM);

// Global data
HINSTANCE   pInstance;            // Handle to instance
HWND		hDlgModeless;		  // Handle to modeless DB

// Variables accessed by several functions
int modeless = 0;
int  iIndex;                  
char szBuffer[] = "                                                 ";
char clearStr[] = "                                                 ";


/*********************************
            WinMain()
 ********************************/
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    PSTR szCmdLine, int iCmdShow)
     {
     static char szAppName[] = "Demo" ;
     HWND        hwnd ;
     MSG         msg ;
     WNDCLASSEX  wndclass ;

	 wndclass.cbSize        = sizeof (wndclass) ;
     wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC ;
     wndclass.lpfnWndProc   = WndProc ;
     wndclass.cbClsExtra    = 0 ;
     wndclass.cbWndExtra    = 0 ;
     wndclass.hInstance     = hInstance ;
     wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION) ;
     wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
     wndclass.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH) ;
     wndclass.lpszMenuName  = MAKEINTRESOURCE(IDR_MENU1);
     wndclass.lpszClassName = szAppName ;
	 wndclass.hIconSm       = LoadIcon (NULL, IDI_APPLICATION) ;

     RegisterClassEx (&wndclass) ;
	 pInstance = hInstance; 

     hwnd = CreateWindow (szAppName,
		                  "Dialog Box Demo",
                          WS_OVERLAPPEDWINDOW,
                          CW_USEDEFAULT, CW_USEDEFAULT,
                          CW_USEDEFAULT, CW_USEDEFAULT,
                          NULL, NULL, hInstance, NULL) ;

     ShowWindow (hwnd, iCmdShow) ;
     UpdateWindow (hwnd) ;

     // The message loop must be modified to handle the modless
	 // dialog box
     while (GetMessage (&msg, NULL, 0, 0))
          {
		  if (hDlgModeless == 0 || !IsDialogMessage (hDlgModeless, &msg))
		      {
                TranslateMessage (&msg) ;
                DispatchMessage (&msg) ;
              }
	      }
     return msg.wParam ;
     }

/********************************
       Windows procedure
*********************************/
LRESULT CALLBACK WndProc (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
     {


     static       HDC          hdc ;
     TEXTMETRIC   tm ;
	 HBRUSH       hBrush;      // Handle to brush
	 HFONT        hfont;       // Handle to a font

     static int  cxChar, cyChar, cxClient, cyClient;

     // Variables for color and font common dialogs
     static CHOOSECOLOR cc ;               // Structure for ChooseColor()
     static COLORREF    custColors[16] ;   // Array for custom colors
	 static CHOOSEFONT  cf ;               // Font structure
	 static LOGFONT     logfont;           // Font attributes

     int    i;                  // counter for custom color display


     switch (iMsg)
          {

		  case WM_CREATE:
               hdc = GetDC (hwnd) ;
               GetTextMetrics (hdc, &tm) ;
               cxChar = tm.tmAveCharWidth ;
               cyChar = tm.tmHeight ;
               return 0 ;

          case WM_SIZE:
               // obtain window size in pixels
               cxClient = LOWORD (lParam) ;
               cyClient = HIWORD (lParam) ;
               return 0 ;
         
          /****************************
                Menu input processing
           ****************************/ 
	      case WM_COMMAND:

			  switch (LOWORD (wParam)) {

                /***************************
				   File menu processing
                 **************************/
				case MY_FILE_OPEN:
                   MessageBeep(0);
                   MessageBox (hwnd,
					           "File Open Requested",
							    "Menu Action",
								MB_ICONEXCLAMATION | MB_OK);
				   return 0;
				   
                 case MY_FILE_CLOSE:
                   MessageBeep(0);
                   MessageBox (hwnd,
					           "File Close Requested",
							    "Menu Action",
								MB_ICONEXCLAMATION | MB_OK);
				   return 0;

                case ID_FILE_COPY:
                   MessageBeep(0);
                   MessageBox (hwnd,
					           "File Copy Requested",
							    "Menu Action",
								MB_ICONEXCLAMATION | MB_OK);
				   return 0;

                /*******************************
				    Dialog menu processing
				 ******************************/
				 // Modeless dialog box command  
                  case ID_DIALOG_MODELESS:
				   if (modeless == 0) {	
					  modeless = 1; 
				      hDlgModeless = CreateDialog (pInstance,
					   	       MAKEINTRESOURCE (IDD_DIALOG4), hwnd,
							   (DLGPROC) TestDlgProc);
                   }   
                   return 0 ;

                  // About menu command
                  case ID_DIALOG_ABOUT:
				   MessageBeep(0);
				   DialogBox (pInstance, MAKEINTRESOURCE (IDD_DIALOG1),
					          hwnd, (DLGPROC) AboutDlgProc) ;
                   return 0 ;

				  // Buttons menu command
                  case ID_DIALOG_BUTTONS:
					 MessageBeep(0);
				     DialogBox (pInstance, MAKEINTRESOURCE (IDD_DIALOG2),
					            hwnd, (DLGPROC) ButtsDlgProc) ;
                  return 0 ;

                  // Listbox menu command
                  case ID_DIALOG_LISTBOX:
					 MessageBeep(0);
					 // Send execution to ListbDlgProc()
				     DialogBox (pInstance, MAKEINTRESOURCE (IDD_DIALOG3),
					            hwnd, (DLGPROC) ListbDlgProc) ;
					 // Execution returns when dialog ends
					 // Display text captured from listbox
					 // Clear the old text from screen
					 TextOut(hdc, 20, 200, clearStr, strlen(clearStr));
					 // Display the string captured from the Listbox
					 TextOut(hdc, 20, 200, szBuffer, strlen(szBuffer));
                  return 0 ;

                 // Color Selector common dialog
                  case ID_DIALOG_COLORSELECTOR:
						cc.lStructSize    = sizeof (CHOOSECOLOR) ;
						cc.hwndOwner      = hwnd ;
						cc.hInstance      = NULL ;
						cc.rgbResult      = RGB (0x80, 0x80, 0x80) ;
						cc.lpCustColors   = custColors ;
						cc.Flags          = CC_RGBINIT | CC_FULLOPEN ;
						cc.lCustData      = 0L ;
						cc.lpfnHook       = NULL ;
						cc.lpTemplateName = NULL ;

                        if (ChooseColor (&cc) == TRUE) {
						   // Start by clearing the client window
						   InvalidateRect (hwnd, NULL, TRUE) ;
                           UpdateWindow (hwnd) ;

                           // Create a solid brush and select it
                           hBrush = CreateSolidBrush(cc.rgbResult);
                           SelectObject (hdc, hBrush) ;
						   
						   // Draw a rectangle using the brush
                           Rectangle (hdc, 20, 20, 100, 100) ;

						   /* Display first eight custom colors using the
                              color triplets stored in the custColors array */
						   for (i = 0; i < 8; i++) {
						      hBrush = CreateSolidBrush(custColors[i]);
                              SelectObject (hdc, hBrush) ;
                              Rectangle (hdc, 20+(20 * i), 120,
								              40+(20 * i), 140) ;
						   }
                           // Clear and exit
						   DeleteObject (SelectObject (hdc, hBrush)) ;
                        }
                       return 0 ;

			      // Font Selector common dialog		   
                  case ID_DIALOG_FONTSELECTOR:
                        // Init CHOOSEFONT structure members
						cf.lStructSize      = sizeof (CHOOSEFONT) ;
						cf.hwndOwner        = hwnd ;
						cf.hDC              = NULL ;
						cf.lpLogFont        = &logfont ;
						cf.iPointSize       = 0 ;
						cf.Flags            = CF_INITTOLOGFONTSTRUCT | CF_SCREENFONTS
										                          | CF_EFFECTS ;
						cf.rgbColors        = 0L ;
						cf.lCustData        = 0L ;
						cf.lpfnHook         = NULL ;
						cf.lpTemplateName   = NULL ;
						cf.hInstance        = NULL ;
						cf.lpszStyle        = NULL ;
						cf.nFontType        = 0 ;               // Returned from ChooseFont
						cf.nSizeMin         = 0 ;
						cf.nSizeMax         = 0 ;

                        // Call ChooseFont function. If it returns TRUE then
						// display a message with the selected font
						if (ChooseFont(&cf) == TRUE){
                           // Start by clearing the client window
						   InvalidateRect (hwnd, NULL, TRUE) ;
                           UpdateWindow (hwnd) ;

                           // Get font selected by user and select it
                           hfont = CreateFontIndirect(cf.lpLogFont);
						   SelectObject(hdc, hfont);

						   // Display a line of text using font
						   TextOut(hdc, 20, 200, "This is a test", 14);

						   // Clear and exit
						   DeleteObject (SelectObject (hdc, hfont)) ;
						}

                       return 0 ;
				}

				break;
 
          /****************************
		       Exit processing
           ****************************/
          case WM_DESTROY :
               PostQuitMessage (0) ;
               return 0 ;
          }
     return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
     }

/*******************************************
     Modeless box dialog box procedure
 *******************************************/
BOOL CALLBACK TestDlgProc (HWND hDlg, UINT iMsg, WPARAM wParam,
						   LPARAM lParam) 
{
     switch (iMsg)
          {
          case WM_INITDIALOG :
			  BringWindowToTop(hDlg);
			  SetFocus(hDlg);
			  SendMessage(hDlg, DM_REPOSITION, 0, 0);
              return TRUE ;

          case WM_COMMAND :
			  switch (LOWORD (wParam)) {

				case ID_BOX_CLOSE: 
				case WM_CLOSE:
					MessageBeep(0);
					modeless = 0;      // Modeless box closed
					DestroyWindow (hDlg) ;

			    return TRUE ;
			  }
          break ;
	 }
     return FALSE ;
}

/*******************************************
         About dialog box procedure
 *******************************************/
BOOL CALLBACK AboutDlgProc (HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam)
     {

     switch (iMsg)
          {
          case WM_INITDIALOG :
               return TRUE ;

          case WM_COMMAND :
               switch (LOWORD (wParam))
                    {
		            case IDOK :
				    MessageBeep(0);
			        EndDialog (hDlg, 0) ;
					return TRUE ;
                    }
               break ;
          }
     return FALSE ;
     }

/*******************************************
         Buttons dialog box procedure
 *******************************************/
BOOL CALLBACK ButtsDlgProc (HWND hDlg, UINT iMsg,
							WPARAM wParam, LPARAM lParam)
     {

     switch (iMsg)
          {
          case WM_INITDIALOG :
               return TRUE ;

          case WM_COMMAND :
               switch (LOWORD (wParam))
                    {
		            case IDOK :
				       MessageBeep(0);
                       MessageBox (hDlg,      // Handle to the dialog window
					           "OK Button Pressed",
							    "Dialog Box Action",
						 MB_ICONEXCLAMATION | MB_OK);
					     EndDialog (hDlg, 0) ;
			        return TRUE ;

		            case IDCANCEL :
                       MessageBeep(0);
                       MessageBox (hDlg,      // Handle to the dialog window
					           "CANCEL Button Pressed",
							    "Dialog Box Action",
					   MB_ICONEXCLAMATION | MB_OK);
					   EndDialog (hDlg, 0) ;
			        return TRUE ;

                    case IDC_BUTTON1:
					   MessageBeep(0);
                       MessageBox (hDlg,      // Handle to the dialog window
					           "ABORT Button Pressed",
							    "Dialog Box Action",
					   MB_ICONEXCLAMATION | MB_OK);
					   EndDialog (hDlg, 0) ;
			        return TRUE ;
                         EndDialog (hDlg, 0) ;
			             return TRUE ;
                    }
               break ;
          }
     return FALSE ;
     }

/*******************************************
       Listbox  dialog box procedure
 *******************************************/
BOOL CALLBACK ListbDlgProc (HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam)
     {

     switch (iMsg)
          {
          case WM_INITDIALOG :
          // Insert message strings into listbox
          // First turn-off screen update
			 SendDlgItemMessage( hDlg, IDC_LIST1, WM_SETREDRAW,
				                 FALSE, 0L);
			 // Insert message strings
			 SendDlgItemMessage( hDlg, IDC_LIST1, LB_ADDSTRING, 0,
                    (DWORD)(LPSTR) "A Midsummers Night Dream");
			 SendDlgItemMessage( hDlg, IDC_LIST1, LB_ADDSTRING, 0,
                    (DWORD)(LPSTR) "Hamlet");
			 SendDlgItemMessage( hDlg, IDC_LIST1, LB_ADDSTRING, 0,
                    (DWORD)(LPSTR) "King Lear");
			 SendDlgItemMessage( hDlg, IDC_LIST1, LB_ADDSTRING, 0,
                    (DWORD)(LPSTR) "Macbeth");
			 SendDlgItemMessage( hDlg, IDC_LIST1, LB_ADDSTRING, 0,
                    (DWORD)(LPSTR) "Julius Cesar");
			 // Screen update back on
			 SendDlgItemMessage( hDlg, IDC_LIST1, WM_SETREDRAW,
				                 TRUE, 0L);				 
             return TRUE ;

          // Processing for user action on dialog box controls 
          case WM_COMMAND :
               switch (LOWORD (wParam))
                    {

				    // Is action on listbox?
			        case IDC_LIST1:
			          // Test for notification resulting from a change in the
                      // selected entry
					  if (HIWORD (wParam) == LBN_SELCHANGE) {
                         // Retreive index number and text of currently
                         // selected entry in the listbox and place in
                         // global variables
						  iIndex = SendDlgItemMessage (hDlg, IDC_LIST1,
							                           LB_GETCURSEL, 0, 0L);
                          SendDlgItemMessage(hDlg, IDC_LIST1, LB_GETTEXT,
							             iIndex, (LPARAM) szBuffer);
					  }
                       return TRUE;

		            case IDOK :
				       MessageBeep(0);
			           EndDialog (hDlg, 0) ;
					return TRUE ;
                    }
               break ;
          }
     return FALSE ;
     }



