//************************************************************
//         ASCII decimal to BCD12 conversions module
//************************************************************
// Reference:   NUMERICAL SOLUTIONS - Chapter 2
// Project:     BCD Conversions
// Entry point: BCD Test
// This file:   math2.cpp
// Header:      math2.h
// Development: Visual C++ 6.0 Console Application
// Date:        December 2000
//
// Description:
//              Function module with BCD to ASCII and ASCII
//              to BCD conversion functions. Includes
//              auxiliary funtions. 
// Note:
//              Functions use inline assembly
//***********************************************************
#include "stdafx.h"
#include <iostream.h>
#include "math2.h"

// Data
int OFFSET_ASC = 0;
int OFFSET_BCD = 0;
int reax;
int rebx;
int recx;
int redx;
int redi;
int resi;



// Prototypes for auxiliary functions in this module
void BIN_TO_BCD();
void ASCII_TO_DX();
void BCD_INTO_REG();
void BCD_TO_ASCII();
void RESTORE_RS();


void AsciiToBcd12( char asc[], char bcd[] )
{
_asm
{
// Conversion of an ASCII decimal number into BCD12 format
// On entry:
// Global variable ASC_STRING
//                 an ASCII decimal string, terminated in a NULL
//                 byte or control code
//             Note: the caller's buffer may contain the $ sign
//                   and commas as formatting symbols. The string
//                   terminator (00H) may be preceded by a CR
//                   (0DH) or LF (0AH) control codes. If the
//                   characters E or e are in the string, the
//                   procedure assumes that the string is in
//                   ASCII exponential form. The code assumes that
//                   exponential strings are normalized.
//
// Global variable BCD_VALUE
//                 12-byte memory area to holds the number encoded
//                 in BCD12 format
//
// Sample input strings:
//        123.223344
//       -0000.000002345
//        1.273E-22
//       -1.44556677 E230
//
//
// On exit:
//                 the caller's buffer with the input number
//                 reformatted in BCD12 format, as follows:
//
//          S s e e e e m . m m m m m m m m m m m m m m m m m m
//                        |
//                        |____ implicit decimal point
//
//    S = sign of number (1 BCD digit)
//    s = sign of exponent (1 BCD digit)
//    e = exponent (4 BCD digits)
//    m = normalized significand (18 BCD digits)
//    . = implicit decimal point between the first and second
//        significand digits
//
// CPU compatibility chart:
//                       | 8086/8088       | NO  |
//                       | 80286           | NO  |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |
//

; Set buffer pointers
		MOV		ESI,asc
		MOV		EDI,bcd
;**********************|
; save buffer addresses|
;**********************|
        MOV     OFFSET_ASC,ESI
        MOV     OFFSET_BCD,EDI
;**********************|
; save input registers |
;**********************|
; Pointer registers are saved. All others are destroyed
        PUSH    ESI
        PUSH    EDI
;**********************|
;   clear buffer for   |
;     BCD12 number     |
;**********************|
; DI --> 12-byte buffer to hold BCD12 number
        PUSH    EDI              ; Save output pointer
        MOV     CX,12           ; Bytes to clear
CLEAR_12:
        MOV     BYTE PTR [EDI],0           ;
        INC     EDI              ; Bump buffer pointer
        LOOP    CLEAR_12
        POP     EDI              ; Restore pointer to start
        MOV     ESI,OFFSET_ASC        ; Source pointer to start
                                        ; of ASCII
;**********************|
;  get sign of number  |
;**********************|
; Find first significant character in the caller's string
; Skip leading blanks, $ and # sign, and other invalid symbols
 GET_NB:
        MOV     AL,BYTE PTR [ESI]        ; Get character
        CMP     AL,'+'          ; Test for + sign
        JAE     FIRST_NB        ; Go if equal or larger
        INC     ESI
        JMP     GET_NB
; first non-blank is a symbol or number larger than the + sign
FIRST_NB:
        CMP     AL,'+'          ; Test for +
        JE      SIGN_IS_PLUS
        CMP     AL,'-'          ; Test for -
        JE      SIGN_IS_MINUS
; First non-blank character is not a sign. Assume a digit
; Positive number is the default setting in BCD buffer
        JMP     TEST_4_EXP      ; Test for exponential form
;**********************|
;  number is negative  |
;**********************|
SIGN_IS_MINUS:
        MOV     [EDI],00010000B ; Set first BCD to 1
;**********************|
;  number is positive  |
;**********************|
; Encoding for positive number (0000B) is already in buffer
SIGN_IS_PLUS:
        INC     ESI              ; Bump ASCII digits pointer
;**********************|
;   test for E or e    |
;**********************|
; The caller's string may be in ASCII exponential form
; Any value less than 20H terminates the ASCII string
TEST_4_EXP:
        MOV     OFFSET_ASC,ESI
                                ; Save pointer to first digit
TEST_AGAIN:
        CMP     [ESI],'E'
        JE      STRING_ISEX             ; String in FP format
        CMP     [ESI],'e'
        JE      STRING_ISEX
        CMP     [ESI],' '       ; Less than 20H is end
                                        ; of string
        JB      SKIP_ZERO               ; Not exponential form
        INC     ESI                      ; Bump digits pointer
        JMP     TEST_AGAIN
;**********************|
;   input is in ASCII  |
;   exponential form   |
;**********************|
STRING_ISEX:
; Sample exponential form ASCII strings:
;            1.2345 E12
;           -1.2345E-12
;           +1.2345 e+12
; SI --> letter e or E
; First test for negative exponent
        INC     ESI              ; Bump pointer past letter E
        PUSH    ESI              ; Save pointer to possible exponent
                                ; sign
; SI --> string of exponent digits
; Skip sign, if there is one
FIRST_0PLUS:
        CMP     [ESI],'0'       ; First valid ASCII number
        JAE     OK_SI_PTR       ; SI pointer is OK if 0 or >
        INC     ESI             ; Bump pointer if not
        JMP     FIRST_0PLUS     ; Continue
OK_SI_PTR:
		MOV		reax,EAX
		MOV		rebx,EBX
		MOV		recx,ECX
		MOV     redx,EDX
        CALL    ASCII_TO_DX     ; Convert to binary in DX
; DX now holds binary value of exponent
        POP     ESI             ; Restore point to possible sign
        CMP     [ESI],'-'       ; Negative exponent
        JNE     NON_NEG_EXP     ; Go if not negative
        JMP     NEG_EXPONENT    ; Routine for negative exponent
; Positive exponent code is already in place
NON_NEG_EXP:
        JMP     CONV_EXPONENT
;**********************|
; skip leading zeros   |
;**********************|
SKIP_ZERO:
; Leading zeros are not significant. skip them
; First reset pointer to start of string
        MOV     ESI,OFFSET_ASC        ; Pointer saved in memory
SKIP_1ZERO:
        CMP     BYTE PTR[ESI],'0'       ; Test for 0
        JNE     NOT_LEAD0
        INC     ESI              ; Bump to next digit
        JMP     SKIP_1ZERO
;**********************|
;  find decimal point  |
;    (if present)      |
;**********************|
;  DX is used to count the non-zero digits preceding the decimal
;  point. This value determines the number's exponent
NOT_LEAD0:
        MOV     DX,0            ; Clear integer digit counter
; Caller's string must end in a NULL byte (00H) or other control
; code < 20H
; Test for an integer zero input in the form 0000
        CMP     BYTE PTR[ESI],' '       ; Smaller than 20H is
                                ; end of input buffer
        JAE     GET_POINT       ; Continue if not end
        JMP     NUM_ISZERO      ; Number is zero
GET_POINT:
        MOV     AL,BYTE PTR [ESI]; Caller's ASCII digit
        CMP     AL,'.'          ; Test for decimal point
        JE      POINT_FOUND
        CMP     AL,','          ; Test for formatting comma
                                ; in the caller's string
        JE      SKIP_COMMA      ; Ignore comma
        CMP     AL,20H          ; String terminator
        JNB     LOOK_4_PT       ; Continue searching if not < 20H
        JMP     NO_POINT        ; No decimal point in string
LOOK_4_PT:
        INC     DX              ; Count one integer digit
SKIP_COMMA:
        INC     ESI              ; Bump integer digits counter
        JMP     GET_POINT
;**********************|
; caller's string is a |
;   decimal fractions  |
;**********************|
; If DX = 0 then the caller's string < 1 and the EXPONENT IS
;           NEGATIVE or the number is zero
; If DX = 1 then the caller's string is already normalized
;         and the EXPONENT = +0
; If DX > 1 then the caller's string is a number =< 10 and
;         the EXPONENT IS POSITIVE
POINT_FOUND:
        CMP     DX,1            ; Test for normalized input
        JNE     TEST_4_DX0
;**********************|
;       DX = 1         |
;**********************|
; ASCII input is in the form: d.dddd...ddd
; Number is between 1 and 10. Sign of exponent = 0000B
; 4 exponent digits = 0000H (these values are already in place)
        JMP     EXP_MOVED
TEST_4_DX0:
        CMP     DX,0            ; Test for negative exponent
        JE      DX_ISZERO       ; Go to DX = 0 processing
; If DX > 1 adjust and store
        DEC     DX
        JMP     CONV_EXPONENT
;**********************|
;       DX = 0         |
;**********************|
; Possible cases:
; non-zero -->   000.002345
; non-zero -->      .2345
;     zero -->     0.00
;     zero -->      .0
; Count decimal digits to normalized position, that is, following
; the first non-zero digit
; DX is used as counter
; At this point:
;                dddd.ddd...
;                    |____ SI points to decimal point
DX_ISZERO:
        INC     ESI              ; Skip decimal point
        INC     DX              ; Count one digit
;**********************|
;     normalize        |
;**********************|
NORMALIZE:
        CMP     BYTE PTR [ESI],'0'       ; Test for zero
        JE      ZERO_DIGIT              ; Go if zero digit
        JB      NUM_ISZERO              ; exit if less than zero
; If not equal or less than zero it must be larger
        JMP     NEG_EXPONENT    ; Normalized position found
ZERO_DIGIT:
        INC     ESI              ; Bump pointer
        INC     DX              ; and digits counter
        JMP     NORMALIZE
;**********************|
;   ASCII number = 0   |
;**********************|
NUM_ISZERO:
; Make sure zero encoding is in BCD buffer
ZERO_EXIT:
        MOV     CX,12           ; Bytes to clear
CLEAR_12_D:
        MOV     BYTE PTR [EDI],0 ;
        INC     EDI              ; Bump buffer pointer
        LOOP    CLEAR_12_D
        JMP     A2FP_EXIT       ; exit conversion
;**********************|
;  negative exponent   |
;**********************|
; DX holds binary value of negative exponent
; Convert to BCD and store
NEG_EXPONENT:
        MOV     EDI,OFFSET_BCD        ; Saved in CS data
; DI --> start of BCD buffer
        MOV     AL,BYTE PTR [EDI]         ; Get BCD sign digit
        MOV     AH,00000001B    ; Code for minus exponent
        OR      AL,AH           ; OR into sign digit
        MOV     BYTE PTR [EDI],AL        ; Store exponent sign
        JMP     CONV_EXPONENT
;**********************|
;  ASCII is an integer |
;**********************|
NO_POINT:
; Exponent of integer is positive. This encoding is already in place
; String has no decimal point. DX = exponent plus 1
        DEC     DX
;**********************|
;  convert and store   |
;      exponent        |
;**********************|
CONV_EXPONENT:
		MOV		reax,EAX
		MOV		rebx,EBX
		MOV		recx,ECX
		MOV     redx,EDX
        CALL    BIN_TO_BCD       ; Binary to BCD conversion
        MOV     EDI,OFFSET_BCD        ; Saved in variable
; DI --> start of BCD buffer
        INC     EDI              ; Bump pointer
; DI --> first BCD exponent digit
        MOV     BYTE PTR [EDI],DH  ; Store two high BCD digits
        INC     EDI              ; Bump pointer again
        MOV     BYTE PTR [EDI],DL; Store two low BCD digits
;**********************|
;  convert and store   |
;     significand      |
;**********************|
; Set pointer to first non-zero and non-sign digit in ASCII number
EXP_MOVED:
        MOV     ESI,OFFSET_ASC        ; Pointer to start of number
LOOK_4_1:
        MOV     AL, BYTE PTR [ESI]       ; Get ASCII digit
        CMP     AL,'1'          ; Test for non-zero and non-control
        JAE     FIRST_NON0      ; Go if 1 or greater
        INC     ESI              ; Bump pointer
        CMP     AL,20H          ; Test for control or NULL
        JAE     LOOK_4_1        ; Continue if not control or NULL
        JMP     ZERO_EXIT       ; Force zero and exit if no valid
                                ; significand digits
FIRST_NON0:
        MOV     CX,18           ; Counter for 18 digits maximum
        MOV     EDI,OFFSET_BCD        ; Set pointer
; DI --> start of BCD buffer
        ADD     EDI,3            ; Offset of significand area
; DI --> BCD significand area in buffer
; SI --> first non-zero digits in the ASCII string
; ASCII string must have at least one valid digit at this point
; Sample significand strings:
;                1.23456(0DH)
;                1.23456(00H)
;                -1.2345 E-12
;                123.45E345
; Operation:
;    The conversion routine takes the ASCII digits from the buffer,
;    converts them to BCD, and stores the digits in packed format.
;    The DH register is used to signal the high-order BCD digit
        MOV     DH,1            ; Code for next digit is high nibble
ASCII_MOVE:
        MOV     AL,BYTE PTR [ESI]        ; ASCII digit
; Test for invalid values
        CMP     AL,'0'          ; Lower limit
        JAE     VALID_1         ; Digit is 0 or larger
; If digit is less than 0, test for period, comma, and other symbols
; and skip. Exit routine if control code or NULL
        CMP     AL,20H          ; Control codes are < 20H
        JAE     NEXT_ASCII      ; Skip if space or symbol
        JMP     A2FP_EXIT       ; Exit with first non ASCII digit
VALID_1:
        CMP     AL,'9'          ; Upper limit of digit
        JBE     VALID_2         ; Go if OK
        JMP     A2FP_EXIT       ; Exit with first non ASCII digit
                                ; greater than 9
VALID_2:
; ASCII digit in AL is ready for storing in packed BCD format
        SUB     AL,30H          ; ASCII to BCD
        CMP     DH,1            ; Test high digit indicator
        JE      HIGH_BCD        ; Go if high
; Low order BCD digit
        MOV     AH,BYTE PTR [EDI]        ; Get previous high digit
        OR      AH,AL           ; OR low digit into AH
        MOV     BYTE PTR [EDI],AH        ; Store low and high
        INC     EDI             ; Bump destination pointer
        MOV     DH,1            ; Set switch for next digit is high
        JMP     NEXT_ASCII      ; Get next ASCII digit
; Routine to store high-order ASCII digit
; Binary value (in AL) must be shifted left 4 position
HIGH_BCD:
        PUSH    CX              ; Save digits counter
        MOV     CL,4            ; For a count of 4
        SHL     AL,CL           ; Shift AL digits
        POP     CX              ; Restore counter
        MOV     BYTE PTR [EDI],AL        ; Store digit in buffer
        MOV     DH,0            ; Clear high digit switch
                                ; Next digit is low
NEXT_ASCII:
        INC     ESI              ; Bump ASCII digits pointer
        LOOP    ASCII_MOVE      ; Continue moving digits
;**********************|
;  restore and exit    |
;**********************|
A2FP_EXIT:
        POP     EDI              ; Restore caller's data registers
        POP     ESI
}
}
// End of ASCII_TO_BCD12


// Local buffer for BIN_TO_BCD
char BCD_BUF[5];

void BIN_TO_BCD()
{
_asm
{
// BIN_TO_BCD
// Convert a 16-bit binary number to 5 BCD digits
// On entry:
//          DX holds binary number
// On exit:
//         5 packed BCD digits in the following registers:
//                CL = digit number 1 (high order digit)
//                DH = digits number 2 and 3
//                DL = digits number 4 and 5 (low order BCD digit)
//
// Operation:
// Binary in AX is first converted into 5 unpacked BCD digits and
// stored in a global buffer (BCD_BUF). Digits are then moved
// from this buffer and packed into DX and CL
//
//
// CPU compatibility chart:
//                       | 8086/8088       | YES |
//                       | 80286           | YES |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |
;

		MOV		AX,DX
;**********************|
;    clear buffer      |
;**********************|
        LEA     EDI,BCD_BUF     ; Set pointer to buffer
        PUSH    EDI             ; Save buffer start
        MOV     CX,5            ; Five digits to clear
CLEAR_5BCD:
        MOV     BYTE PTR [EDI],0         ; Clear digit
        INC     EDI             ; Bump pointer
        LOOP    CLEAR_5BCD      ; Repeat for 5 digits
        DEC     EDI             ; Adjust buffer pointer to last
                                ; digit
        MOV     CX,10           ; Decimal divisor to CX
GET_BIN:
        MOV     DX,0            ; Clear for word division
; Divide binary by 10
        DIV     CX              ; Perform division AX/CX
; Quotient is in AX and remainder in DL
        MOV     BYTE PTR [EDI],DL     ; Store digit in buffer
        DEC     EDI             ; Bump destination pointer
; Note: step 5 is performed automatically since the binary
; quotient is left in AX
        CMP     AX,0            ; Test for end of binary
        JNZ     GET_BIN         ; Continue if not end
; Last digit converted
;**********************|
;   BCD to registers   |
;      and pack        |
;**********************|
        POP     EDI             ; Restore pointer to start of
                                ; buffer
        ADD     EDI,4           ; Point to last digit
        MOV     AL,BYTE PTR [EDI]     ; BCD digit to AL
        MOV     BL,AL           ; To results register low nibble
        DEC     EDI             ; Point to fourth BCD digit
        MOV     AL,BYTE PTR [EDI]     ; BCD digit to AL
        MOV     CL,4            ; Shift counter
        SHL     AL,CL           ; Shift BCD to high nibble
        OR      BL,AL           ; OR into result register high
                                ; nibble
        DEC     EDI              ; Pointer to third BCD digit
        MOV     AL,BYTE PTR [EDI]     ; BCD digit to AL
        MOV     BH,AL           ; To results register low nibble
        DEC     EDI              ; Point to second BCD digit
        MOV     AL,BYTE PTR [EDI]     ; BCD digit to AL
        MOV     CL,4            ; Shift counter
        SHL     AL,CL           ; Shift BCD to high nibble
        OR      BH,AL           ; OR into result register high
                                ; nibble
; Move 4 low order BCD digits to DX
        MOV     DX,BX
; Move high order digit to CL
        DEC     EDI              ; Pointer to first BCD digit
        MOV     AL,BYTE PTR [EDI]     ; BCD digit to AL
        MOV     CL,AL           ; To results register low nibble
}
}


// Local data
int OFFSET_ASCBUF = 0;

void Bcd12ToAscStr( char asc[], char bcd[] )
{
_asm
{
// Convert a number in BCD12 format into an ASCII string
// On entry:
//         array bcd[] is a 12-byte BCD12 buffer holding number
//         array asc[] is a 31-byte ASCII buffer to hold string
// On exit:
//         array asc[] holds ASCII string
//
// Operation:
//         Numbers with exponent < 10 are converted to decimal
//         format. If exponent is > 10 the number is converted to
//         exponential format
//
// BCD12 format:
//          S s e e e e m . m m m m m m m m m m m m m m m m m m
//                        |
//                        |____ implicit decimal point
//
//    S = sign of number (1 BCD digit)
//    s = sign of exponent (1 BCD digit)
//    e = exponent (4 BCD digits)
//    m = normalized significand (18 BCD digits)
//        (in positive and negative numbers the first significand
//         digit must be non-zero)
//    . = implicit decimal point between the first and second
//        significand digits
//
// CPU compatibility chart:
//                       | 8086/8088       | NO  |
//                       | 80286           | NO  |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |
;

;**********************|
; save buffer addresses|
;**********************|
		MOV		EDI,asc
		MOV		ESI,bcd
        MOV     OFFSET_ASCBUF,EDI
; Save pointer registers
        PUSH    ESI
        PUSH    EDI
; Clear results buffer
        PUSH    EDI              ; Save pointer
        MOV     CX,31           ; Bytes to clear
CLEAR_23:
        MOV     BYTE PTR [EDI],0 ; Clear to zero
        INC     EDI              ; Bump pointer
        LOOP    CLEAR_23
        POP     EDI             ; Restore pointer to start of buffer
;**********************|
;  get sign of number  |
;**********************|
        MOV     BYTE PTR [EDI],' '      ; In case of positive
        MOV     AL,[ESI]         ; First byte of BCD12 number
        TEST    AL,00010000B    ; Bit 4 is negative number
        JZ      NNOT_NEG        ; Number is not negative
; Negative number. Set sign
        MOV     BYTE PTR [EDI],'-'       ; Sign to results buffer
NNOT_NEG:
        INC     EDI              ; Bump pointer
;**********************|
; get value of exponent|
;**********************|
        INC     ESI             ; Pointer to BCD12 exponent field
        INC     ESI             ; Point to low byte of BCD exponent
; Move BCD digits to AL-BH-BL-CH-CL for conversion routine
        MOV     AL,0            ; Highest digit is zero
        MOV     DL,BYTE PTR [ESI]       ; DL holds 2 packed digits
        MOV     DH,DL           ; Copy digits to DH
        AND     DL,00001111B    ; Erase high nibble in DL
        AND     DH,11110000B    ; and low in DH
        MOV     CX,4            ; For a count of 4
        SHR     DH,CL           ; Shift digit to low position
        DEC     ESI              ; Bump pointer to previous digit
        MOV     BL,BYTE PTR [ESI]        ; BL holds 2 packed digits
        MOV     BH,BL           ; Copy digits to BH
        AND     BL,00001111B    ; Erase high nibble in BL
        AND     BH,11110000B    ; and low in BH
        MOV     CX,4            ; For a count of 4
        SHR     BH,CL           ; Shift digit to low position
        MOV     CX,DX           ; Move digits to CX
        PUSH    ESI             ; Save source
        PUSH    EDI             ; and destination pointers
		MOV		reax,EAX
		MOV		rebx,EBX
		MOV		recx,ECX
		MOV     redx,EDX	
        CALL    BCD_INTO_REG    ; BCD to binary conversion
        POP     EDI             ; Restore pointers
        POP     ESI
; DX holds binary value of exponent
; Numbers with exponent = or < 10 are stored in conventional form
; Number with exponent > 10 are stored in exponential form
        CMP     DX,10           ; Test for limit
        JA      EXPO_FORM       ; Exponential form routine
        JMP     NORMAL_FORM     ; Conventional form if 10 or less
;**********************|
;   exponential form   |
;**********************|
; Buffer is formatted as follows for exponential form:
;       -1.2345 ...89 E-23
;       |||          ||_________ offset 21
;       |||          |__________ offset 20
;       |||_____________________ offset 2
;       ||______________________ offset 1
;       |_______________________ offset 0 (- sign or blank)
;**********************|
;  initialize buffer   |
;**********************|
EXPO_FORM:
        MOV     EDI,OFFSET_ASCBUF       ; Start of ASCII buffer
        MOV     BYTE PTR [EDI+2],'.'    ; Set period in buffer
        MOV     BYTE PTR [EDI+20],' '   ; Blank space separator
        MOV     BYTE PTR [EDI+21],'E'   ; Letter E
; Get and encode exponent sign
        ADD     EDI,22           ; Set pointer to sign position
        DEC     ESI              ; Set pointer to first BCD digit
        MOV     AL,BYTE PTR [ESI]        ; Get BCD sign digit
        TEST    AL,00000001B    ; Low bit is negative exponent
        JZ      POS_EXP_EXP     ; Exponent is positive
        MOV     BYTE PTR [EDI],'-'      ; Place negative sign
        INC     EDI
POS_EXP_EXP:
        INC     ESI              ; BCD pointer to exponent area
; DI --> position for first exponent digit
        MOV     EBX,OFFSET_ASCBUF     ; EBX --> start of buffer
        ADD     EBX,3           ; Buffer offset 3 to 6 is used
                                ; as a temporary storage
        MOV     AL,[ESI]        ; Get 2 packed BCD digits
        MOV     AH,AL           ; Copy digit in AH
        AND     AL,00001111B    ; Unpack low digit
        AND     AH,11110000B    ; and high digit
        MOV     CL,4            ; For a count of 4
        SHR     AH,CL           ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H          ; BCD + 30H = ASCII
        ADD     AH,30H
        MOV     BYTE PTR [EBX],AH       ; Store ASCII
        INC     EBX             ; Bump pointer to temporary area
        MOV     BYTE PTR [EBX],AL
        INC     EBX              ; Bump pointer again
        INC     ESI              ; BCD pointer to next digit pair
; Get next two BCD digits and store
        MOV     AL,[ESI]         ; Get 2 packed BCD digits
        MOV     AH,AL           ; Copy digit in AH
        AND     AL,00001111B    ; Unpack low digit
        AND     AH,11110000B    ; and high digit
        MOV     CL,4            ; For a count of 4
        SHR     AH,CL           ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H          ; BCD + 30H = ASCII
        ADD     AH,30H
        MOV     BYTE PTR [EBX],AH        ; Store ASCII
        INC     EBX              ; Bump pointer to temporary area
        MOV     BYTE PTR [EBX],AL
        SUB     EBX,3            ; BX --> first digit in temporary
                                ; area
; Move digits to final position skipping leading zeros
        MOV     CX,4            ; For a count of 4
SKIP_LEAD0:
        MOV     AL,BYTE PTR [EBX]        ; Digit in temporary area
        CMP     AL,'0'          ; Test for leading zero
        JNE     NOT_LEAD00      ; Go if not a leading zero
        INC     EBX              ; Bump pointer
        LOOP    SKIP_LEAD0      ; Test 4 digits
; Execution drops if all exponent digits are 0
        MOV     BYTE PTR [EDI],'0'       ; Place one zero in buffer
        INC     EDI             ; Bump buffer pointer
        JMP     EXPONENT_RDY
NOT_LEAD00:
        MOV     AL,BYTE PTR [EBX]        ; Get digit
        MOV     BYTE PTR [EDI],AL        ; and store it
        INC     EBX              ; Bump pointers
        INC     EDI
        LOOP    NOT_LEAD00      ; Continue for remaining digits
;
EXPONENT_RDY:
        MOV     BYTE PTR [EDI],0FFH      ; Special code for clean up
                                        ; routine
; Exponent is in buffer. Move significand using existing routine
        SUB     ESI,2            ; Set SI to start of BCD buffer
        MOV     EDI,OFFSET_ASCBUF     ; Start of ASCII buffer
        INC     EDI              ; Skip sign position
        MOV     DX,0            ; Normalized significand
        JMP     POS_EX_FORM     ; Handle by positive exponent
;**********************|
;     normal form      |
;**********************|
NORMAL_FORM:
; DX holds binary value of exponent, which must be 10 or less
; Possible storage forms:
; If sign of exponent is negative:
;       0.1234
;       0.00001234
; If sign of exponent is positive
;       12.3456
;       12345.67
;       1.2345 (exponent = 0)
;
        DEC     ESI             ; Set pointer to first BCD digit
        MOV     AL,BYTE PTR [ESI]        ; Get BCD sign digit
        TEST    AL,00000001B    ; Low bit is negative exponent
        JZ      POS_EX_FORM     ; Exponent is positive
        JMP     NEG_EX_FORM     ; Go to negative exponent routine
;**********************|
; exponent is positive |
;**********************|
; DX holds the number of digits to move to the right the decimal
; point from the normalized position
; DI --> ASCII buffer
POS_EX_FORM:
        PUSH    EDI             ; Save start of ASCII buffer
        ADD     DI,DX           ; Add exponent to start of buffer
        INC     EDI             ; Add one for normalized format
        MOV     BYTE PTR [EDI],'.'       ; Place point in buffer
        POP     EDI              ; Restore pointer to start position
MOVE_SIG:
        ADD     ESI,3            ; Set pointer to BCD significand
; SI --> first significand digit in packed BCD format
;**********************|
;   unpack and move    |
;     BCD digits       |
;**********************|
        MOV     DX,9            ; Counter for 9 packed BCD digits
UNPACK_2:
        MOV     AL,[ESI]         ; Get 2 packed BCD digits
        MOV     AH,AL           ; Copy digit in AH
        AND     AL,00001111B    ; Unpack low digit
        AND     AH,11110000B    ; and high digit
        MOV     CL,4            ; For a count of 4
        SHR     AH,CL           ; Move high digit to low nibble
; Convert BCD digits to ASCII
        ADD     AL,30H          ; BCD + 30H = ASCII
        ADD     AH,30H
; Place two digits in buffer
        CMP     BYTE PTR [EDI],'.'       ; Test for period in buffer
        JNE     NO_PERIOD1      ; OK if no period
        INC     EDI             ; Bump pointer if on period
NO_PERIOD1:
        MOV     BYTE PTR [EDI],AH       ; Store high digit
        INC     EDI             ; Bump pointer to next digit
        CMP     BYTE PTR [EDI],'.'      ; Test for period in buffer
        JNE     NO_PERIOD2      ; OK if no period
        INC     EDI             ; Bump pointer if on period
NO_PERIOD2:
        MOV     BYTE PTR [EDI],AL        ; Store low digit
        INC     EDI
        INC     ESI             ; Bump pointer to next BCD
        DEC     DX              ; Adjust digits counter
        JNZ     UNPACK_2
;**********************|
;       clean up       |
;**********************|
; Clear trailing zeros from buffer
; Example buffer digits:  1.23450000000000000(00)
;                         1234.00000000000000(00)
; Buffer after clean-up:  1.2345(00)
;                         1234(00)
;
        MOV     EDI,OFFSET_ASCBUF     ; Entry offset from storage
        ADD     EDI,30           ; Point to last digit
; Replace up to 27 trailing zeros with NULL bytes
        MOV     CX,29           ; Counter for maximum digits
REPLACE_27:
        CMP     BYTE PTR [EDI],0FFH     ; Special code for
                                        ; exponential form
        JNE     NT_EXP_FORM     ; Go if not
        MOV     BYTE PTR [EDI],00H       ; Clear special code
        JMP     EXIT_REP27      ; Do not fix trailing zeros
NT_EXP_FORM:
        CMP     BYTE PTR [EDI],'0'       ; Test for trailing 0
        JNE     TEST_27          ; Test for other values
        MOV     BYTE PTR [EDI],0 ; Replace zero with NULL
        JMP     NEXT_27
TEST_27:
        CMP     BYTE PTR [EDI],0; Test for NULL at end of buffer
        JNE     EXIT_REP27      ; Exit if not 0 or NULL
NEXT_27:
        DEC     EDI             ; Adjust pointer
        LOOP    REPLACE_27      ; Continue for 27 digits
EXIT_REP27:
; If the last character found is a period, then number is in the
; form 1234.0000..00.
        CMP     BYTE PTR [EDI],'.'       ; Test for period
        JNE     NO_PERIOD_FIX    ; Go if not
        MOV     BYTE PTR [EDI],0 ; Replace period with NULL
NO_PERIOD_FIX:
        JMP     EXIT_ASCSTR
;**********************|
; exponent is negative |
;**********************|
; Display forms for negative exponent
;       0.1234 (exponent = -1)
;       0.00001234 (exponent = -5)
; Operation:
;    The decimal point is first placed in the buffer
;    DX number of zeros are placed in the buffer, skipping the
;    decimal point
NEG_EX_FORM:
        INC     EDI              ; Leave blank for leading zero
        MOV     BYTE PTR [EDI],'.'       ; Place point in buffer
        DEC     EDI
PLACE_0S:
        CMP     BYTE PTR [EDI],'.'       ; Test for period in buffer
        JNE     NO_PERIOD3      ; No period, proceed
        INC     EDI              ; Bump buffer pointer to skip period
NO_PERIOD3:
        MOV     BYTE PTR [EDI],'0'       ; Place 0 digit in buffer
        INC     EDI              ; Bump pointer to next digit
        DEC     DX              ; Decrement counter for zeros
        JNZ     PLACE_0S        ; Continue if not zero
        JMP     MOVE_SIG        ; Use existing code to move
                                ; significand
;**********************|
;    end of routine    |
;**********************|
EXIT_ASCSTR:
        POP     EDI              ; Restore pointer registers
        POP     ESI
}
}


void BCD_INTO_REG()
{
char ASC_TEMP_BUF[5];
_asm
{
// Convert a 5-digit BCD number into a binary in DX
// On entry:
//         AL-BH-BL-CH-CL hold 5 BCD digits
//                        holding a value < 65536
// On exit:
//         DX = binary number (range 0 to 65535)
//
//
// CPU compatibility chart:
//                       | 8086/8088       | NO  |
//                       | 80286           | NO  |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |
//
//

		CALL	RESTORE_RS
        LEA     ESI,ASC_TEMP_BUF ; Address of buffer
		MOV		reax,EAX
		MOV		rebx,EBX
		MOV		recx,ECX
		MOV     redx,EDX
        CALL    BCD_TO_ASCII    ; Convert BCD digits to ASCII
                                ; string
		MOV		reax,EAX
		MOV		rebx,EBX
		MOV		recx,ECX
		MOV     redx,EDX
					
        CALL    ASCII_TO_DX     ; Conversion routine
}
}

void BCD_TO_ASCII()
{
_asm
{
// Convert 5 unpacked BCD digits in machine registers into to 5
// ASCII digits
// On entry:
//         AL-BH-BL-CH-CL hold 5 unpacked BCD digits
//                        holding a value < 65536
//         ESI --> 5 byte buffer to hold ASCII string
// On exit:
//         ESI --> buffer holding ASCII digits
//
// CPU compatibility chart:
//                       | 8086/8088       | NO  |
//                       | 80286           | NO  |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |
//

		CALL	RESTORE_RS
; Clear buffer for ASCII digits
        PUSH    ESI              ; Save start of buffer
        PUSH    ESI              ; Twice
        PUSH    CX              ; Save BCD high digit
        MOV     CX,5            ; Counter for 5 digits
CLEAR_BUF_5:
        MOV     BYTE PTR [ESI],' '    ; Set blank in buffer
        INC     ESI              ; Bump pointer
        LOOP    CLEAR_BUF_5
        POP     CX              ; Restore BCD digit
;**********************|
; convert to ASCII and |
;    move to buffer    |
;**********************|
; Convert BCD digits to ASCII
        ADD     AL,30H          ; Add 30H to each digit
        ADD     BH,30H
        ADD     BL,30H
        ADD     CH,30H
        ADD     CL,30H
        POP     ESI              ; Restore buffer start
        MOV     BYTE PTR [ESI],AL     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     BYTE PTR [ESI],BH     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     BYTE PTR [ESI],BL     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     BYTE PTR [ESI],CH     ; Digit to buffer
        INC     ESI              ; Bump buffer pointer
        MOV     BYTE PTR [ESI],CL     ; Digit to buffer
        POP     ESI              ; Restore start of buffer
}
}


void ASCII_TO_DX()
{
_asm
{
// Convert an ASCII number to binary and store in DX register
// On entry:
// ESI -->   start of 5-digit ASCII buffer holding decimal
//           string in the range 0 to 65,535
// On exit:
//           DX = binary number
//
// CPU compatibilty chart:
//                       | 8086/8088       | NO  |
//                       | 80286           | NO  |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |
//

		CALL	RESTORE_RS      
;***********************|
;  initialize registers |
;***********************|
        MOV     DX,0            ; Clear accumulator register
        MOV     CX,10           ; Load multiplier into CX
;***********************|
; test digit for range  |
;        0 to 9         |
;***********************|
DIGIT_TO_ACC16:
        MOV     AL,BYTE PTR [ESI]         ; Get ASCII digit
        CMP     AL,'0'          ; Test for lower limit
        JB      EXIT_ASC16      ; Exit if less than 0
        CMP     AL,'9'          ; Test for higher limit
        JA      EXIT_ASC16      ; Exit if larger than 9
;***********************|
; ASCII digit to binary |
;***********************|
        SUB     AL,30H          ; ASCII to decimal
;***********************|
;    accumulator x 10   |
;      + digit          |
;***********************|
        MOV     AH,0            ; Clear high byte in AX
        PUSH    AX              ; Save digit in stack
        MOV     AX,DX           ; Previous product to AX
        MUL     CX              ; Perform AX = AX * 10
        MOV     DX,AX           ; Move product to accumulator
        POP     AX              ; Restore decimal digit
        ADD     DX,AX           ; Add digit to accumulator
;***********************|
; pointer to next digit |
;***********************|
        INC     ESI              ; Bump pointer
        JMP     DIGIT_TO_ACC16  ; Continue
EXIT_ASC16:
}
}


void RESTORE_RS()
{
_asm
{
		MOV		EAX,reax
		MOV		EBX,rebx
		MOV		ECX,recx
		MOV		EDX,redx
}
}


void DecodeSingle(float single, int *sig, int *exp, int *signf )
{
_asm
{
// C++ function to separate the elements of a floating point number
// stored in memory in IEEE single format
// Pre:
//         Parameter single is an IEEE floating point single
// Post:
//         Parameter sig is the number's sign field:
//                    1 = negative sign
//                    0 = positive sign
//         Parameter exp is  8-bit exponent in bias 127 form
//         Parameter signf is 24-bit significand
//                    Implicit significand bit is set if exponent
//                    is not zero
//
// CPU compatibility chart:
//                       | 8086/8088       | NO  |
//                       | 80286           | NO  |
//                       | 80386           | YES |
//                       | 80486/PENTIUM   | YES |
//

         LEA     ESI,single
;***********************|
;       decode sign     |
;***********************|
        ADD     ESI,3            ; Set pointer to last byte
        MOV     AH,BYTE PTR[ESI]         ; Get byte into AH
        AND     AH,10000000B    ; Mask off exponent bits
        MOV     CL,7            ; Counter to shift 7 bits
        SHR     AH,CL           ; Shift sign bit to right-most bit
;***********************|
;     decode exponent   |
;***********************|
        MOV     AL,[ESI]         ; Get 7 high bits of exponent
        SHL     AL,1            ; Shift left one bit
        DEC     ESI              ; Pointer to byte at offset 2
        MOV     BH,[ESI]         ; Get byte into BH
        TEST    BH,10000000B    ; Test for exponent bit 8 set
        JZ      OK_EXPONENT     ; Exponent OK if bit not set
        OR      AL,00000001B    ; Set low exponent bit
;***********************|
;  decode significand   |
;***********************|
OK_EXPONENT:
; Test for a non-zero exponent so set implicit significand bit
        CMP     AL,0            ; Is exponent zero
        JE      IMPLICIT_0      ; Implicit bit is zero with a
                                ; zero exponent. Do not set
;***********************|
;     explicit 1 bit    |
;***********************|
        OR      BH,10000000B    ; Set BH high bit
IMPLICIT_0:
        DEC     ESI              ; Point to byte at offset 1
        MOV     BL,[ESI]         ; Load into BL
        DEC     ESI              ; Point to byte at offset 0
        MOV     CH,[ESI]         ; Load into CH
;***********************|
;  Data to variables    |
;***********************|
; Move IEEE format data from registers to variables
; Start with exponent, in AL
		MOV     EDX,EAX
        AND     EDX,0FFH          ; Preserve AL
        MOV     EDI,exp
		MOV     [EDI],EDX
; Store sign, in AH
		MOV     EDX,EAX
        AND     EDX,0FF00H        ; Preserve AH
		SHR     EDX,8
		MOV     EDI,sig
		MOV     [EDI],EDX
; Store significand (in BH, BL, and CH)
        MOV     EDX,EBX
		AND     EDX,0FFFFH        ; Clear high-order bits
        SHL     EDX,8             ; Make space for low order bits
		AND     ECX,0FF00H        ; Preserve CH
		SHR     ECX,8             ; Move to low order bits
		OR      EDX,ECX
        MOV     EDI, signf
		MOV     [EDI],EDX
}
}

