//************************************************************
//                   Calculator primitives
//************************************************************
// Reference:   SOFTWARE TOOLKIT for C++ Programmers
//              by Sanchez, Schilling, and Canton
// Workspace  : Chapter09\Calc Primitives
// Entry point: Calc Primitives.cpp
// Other files: Calc Primitives.h
//              Un32_5.obj
//              Un32_6.obj
//              Un32_7.obj
// Development: Visual C++ 6.0 Console Application
// Date:        January 12/2001
//              April 16/2001 
//
// Description:
//              Tests for calculator primitives
//************************************************************

// Library functions
extern "C" FACTORIAL();      // factorial
extern "C" SOLVE_FOR_A();    // Solve right trianglE
extern "C" ORDER();          // Order 2 value in math unit
extern "C" MAX();            // Return largest
extern "C" MIN();            // Returns smallest
extern "C" MODULUS();        // Modulo x/y
extern "C" PYTHAGORAS();     // Returns SQRT(a^2+b^2)
extern "C" INT_FRAC();       // Break integer and fraction
extern "C" QUAD_FORM_REAL(); // Real roots of quadratic equation
extern "C" QUAD_FORMULA();   // Real or imaginary roots
extern "C" ADD_CMPLX();      // Add complex numbers
extern "C" SUB_CMPLX();      // Subtract complex numbers
extern "C" MUL_CMPLX();      // Multiply complex numbers
extern "C" DIV_CMPLX();      // Divide complex numbers
extern "C" ABS_CMPLX();      // Absolute value of complex number
extern "C" POLAR_TO_CART();  // Polar coordinates to cartesian
extern "C" CART_TO_POLAR();  // Cartesian coordinates to polar
//**************************************************************
//                       Prototypes
//**************************************************************
long Factorial(int);
double SolverForA(double, double);
void Order(double, double);
double Max(double, double);
double Min(double, double);
void InitMU(void);
double Pythagoras(double, double);
double FracPart(double num);
int IntPart(double num);
void QuadFormReal(double, double, double, double *, double *);
void QuadForm(double, double, double, double *, double *);
void AddComplex(double, double, double, double,\
				double *, double *);
void SubComplex(double, double, double, double,\
				double *, double *);
void MulComplex(double, double, double, double,\
				double *, double *);
void DivComplex(double, double, double, double,\
				double *, double *);
double AbsComplex(double, double);
void PolarToCart(double, double, double *, double *);
void CartToPolar(double, double, double *, double *);
//**************************************************************
//            C++ interface functions to Un32_7 module
//**************************************************************
long Factorial(int value)
// Calculate factorial of argument
// On entry:
//         parameter value holds x
// On exit:
//         returns x!
{
long fac; 
_asm
{
    FILD    value
	CALL    FACTORIAL
    FISTP   fac
}
return fac;
}


double SolveForA(double angle, double sideb)
// Solve a right triangle in terms of one side and the adjacent
// angle using the formula
//                     a = b * Tan A
// On entry:
//         parameter angle and sideb hold triangle data
// On exit:
//        returns side a

{
double sidea; 
_asm
{
    FLD     sideb
    FLD     angle
	CALL    SOLVE_FOR_A
    FSTP    sidea
}
return sidea;
}


void Order(double x, double y)
// Place the larger value in ST(0) and the smaller one in
// ST(1)
// On entry:
//         parameters x and y contain values
// On exit:
//        ST(0) holds larger value
//        ST(1) holds smaller value

{
_asm
{
    FLD     x
    FLD     y
	CALL    ORDER
}
return;
}

void InitMU(void)
{
// Re-initialize and clear the math unit
_asm
{
    FINIT
}
}


double Max(double x, double y)
// Return the larger value
// On entry:
//         parameters x and y contain values
// On exit:
//         returns larger
{
double value;
_asm
{
    FLD     x
    FLD     y
	CALL    MAX
    FSTP    value
}
return value;
}

double Min(double x, double y)
// Return the smallest value
// On entry:
//         parameters x and y contain values
// On exit:
//         returns smaller
{
double value;
_asm
{
    FLD     x
    FLD     y
	CALL    MIN
    FSTP    value
}
return value;
}

int Modulus(int x, int y)
// Return the modulus x/y
// On entry:
//         parameters x and y contain values
// On exit:
//         x modulo y
{
int modulo;
_asm
{
    FILD    y
    FILD    x
	CALL    MODULUS
    FISTP   modulo
}
return modulo;
}

double Pythagoras(double a, double b)
// Return the square root of the sum of 2 squares
// On entry:
//         parameters a and b contain values
// On exit:
//         returns square root of a^2 + b^2
{
double c;
_asm
{
    FLD     a
    FLD     b
	CALL    PYTHAGORAS
    FSTP    c
}
return c;
}

double FracPart(double num)
// Returns the fractional part of a float
// On entry:
//         parameter num holds operand
// On exit:
//         returns fractional part of num
{
double frac;
_asm
{
	FLD    num
	CALL   INT_FRAC
	FXCH
	FSTP   frac
}
return frac;
}

int IntPart(double num)
// Returns the integer part of a float
// On entry:
//         parameter num holds operand
// On exit:
//         returns integer part of num
{
int part;
_asm
{
	FLD    num
	CALL   INT_FRAC
	FISTP  part
}
return part;
}


void QuadFormReal(double a, double b, double c, double *x1, double *x2)
// Solution of a quadratic equation in the form:
//
//                     ax**2 + bx + c = 0
//
// by means of the formula
//
//                   -b + SQR [(b**2 - 4ac)]
//            x(1) = -------------------------
//                            2a
//
//                   -b - [SQR (b**2 - 4ac)]
//            x(2) = -------------------------
//                            2a
//
// where x(1) and x(2) are two possible roots and a, b, and c are
// constants
//
// On entry:
//          parameters a, b, and c contain equation coefficients
// On exit:
//          parameters x1 and x2 contain roots
{
double r1, r2;
_asm
{
	FLD    a
    FLD    b
	FLD    c
	CALL   QUAD_FORM_REAL
	FSTP   r1
	FSTP   r2
}
*x1 = r1;
*x2 = r2;
return;
}

int QuadFormula(double a, double b, double c, double *x1, double *x2)
// Solution of a quadratic equation in the form:
//
//                     ax**2 + bx + c = 0
//
// for its real or imaginary roots, using the formula
//
//                   -b + SQR [(b**2 - 4ac)]
//            x(1) = -------------------------
//                            2a
//
//                   -b - [SQR (b**2 - 4ac)]
//            x(2) = -------------------------
//                            2a
//
// where x(1) and x(2) are two possible roots and a, b, and c are
// constants
// Solution sets:
//          x(1) and x(2) are the real roots
//          x(r) +/- x(i) is the complex root
//
// On entry:
//                 ST(0) = c
//                 ST(1) = b
//                 ST(2) = a
// On exit:
//        returns 0 if roots are real numbers, then
//                 parameter x1 holds first root
//                 parameter x2 holds second root
//        returns 1 if roots are complex, then
//        solution is:
//                 x1= -b/2a = real part = x(r)
//                 x2 = +/- i * SQR(b**2 - 4ac)/2a = imaginary
//                      part = +/- x(i)
{
double r1, r2;
int rootType = 0;    // 0 if real, 1 if imaginary roots 
_asm
{
	FLD    a
    FLD    b
	FLD    c
	CALL   QUAD_FORMULA
	FSTP   r1
	FSTP   r2
	JNC    ROOT_EXIT
	MOV    rootType, 1
ROOT_EXIT:		
}
*x1 = r1;
*x2 = r2;
return rootType;
}

void AddComplex(double ra, double rbi, double rc, double rdi,\
				double *real, double *imaginary)
// Addition of two complex numbers (a + bi and c + di) by means of
// the formula:
//
//       (a + bi) + (c + di) = (a + c) + i(b + d)
//
// On entry:
//         parameters a, bi, c, and di holds input data
// On exit:
//         parameter real holds real part
//         parameter imaginary holds imaginary part
{
double rPart, iPart;
_asm
{
	FLD    ra
    FLD    rbi
	FLD    rc
	FLD    rdi
	CALL   ADD_CMPLX
	FSTP   rPart
	FSTP   iPart
}
*real = rPart;
*imaginary = iPart;
return;
}

void SubComplex(double ra, double rbi, double rc, double rdi,\
				double *real, double *imaginary)
// Subtraction of two complex numbers (a + bi and c + di) by means of
// the formula:
//
//       (a + bi) - (c + di) = (a - c) + i(b - d)
//
// On entry:
//         parameters a, bi, c, and di holds input data
// On exit:
//         parameter real holds real part
//         parameter imaginary holds imaginary part
{
double rPart, iPart;
_asm
{
	FLD    ra
    FLD    rbi
	FLD    rc
	FLD    rdi
	CALL   SUB_CMPLX
	FSTP   rPart
	FSTP   iPart
}
*real = rPart;
*imaginary = iPart;
return;
}

void MulComplex(double ra, double rbi, double rc, double rdi,\
				double *real, double *imaginary)
// Multiplication of two complex numbers (a + bi and c + di) by means of
// the formula:
//
//       (a + bi)(c + di) = (ac - bd) + i(ad + bc)
//
// On entry:
//         parameters a, bi, c, and di holds input data
// On exit:
//         parameter real holds real part
//         parameter imaginary holds imaginary part
{
double rPart, iPart;
_asm
{
	FLD    ra
    FLD    rbi
	FLD    rc
	FLD    rdi
	CALL   MUL_CMPLX
	FSTP   rPart
	FSTP   iPart
}
*real = rPart;
*imaginary = iPart;
return;
}

void DivComplex(double ra, double rbi, double rc, double rdi,\
				double *real, double *imaginary)
// Division of two complex numbers (a + bi and c + di) by means of
// the formula:
//       (a + bi)            (ac + bd) + i(bc - ad)
//       --------     =      ----------------------
//       (c + di)                   c^2 + d^2
//
// On entry:
//         parameters a, bi, c, and di holds input data
// On exit:
//         parameter real holds real part
//         parameter imaginary holds imaginary part
{
double rPart, iPart;
_asm
{
	FLD    ra
    FLD    rbi
	FLD    rc
	FLD    rdi
	CALL   DIV_CMPLX
	FSTP   rPart
	FSTP   iPart
}
*real = rPart;
*imaginary = iPart;
return;
}

double AbsComplex(double ra, double rbi)
// Find the absolute value of a complex number using the formula:
//
//           |a + bi| = SQR (a^2 + b^2)
//
// On entry:
//         parameters ra and rbi hold input data
//         ST(1) = bi
// On exit:
//         returns |a + bi|
{
double absPart;
_asm
{
	FLD    ra
    FLD    rbi
	CALL   ABS_CMPLX
	FSTP   absPart
}
return absPart;
}

void PolarToCart(double angle, double r, double *x, double *y)
// Function for converting a coordinate pair in polar form to a
// coordinate pair in Cartesina (rectangular) form by means of the
// formulas:
//
//            x = r cos A
//            y = r sin A
//
// Where r is the directed distance from the pole, A is the
// directed angle from the polar axis, expressed in degrees,
// and x, y are the Cartesian coordinate pair
// On entry:
//          parameters r and angle hold input values
//          ST(1) = r
// On exit:
//          parameter x is x Cartesian coordinate
//          parameter y is y Cartesian coordinate
{
double xCoord, yCoord;
_asm
{
	FLD    r
    FLD    angle
	CALL   POLAR_TO_CART
	FSTP   xCoord
	FSTP   yCoord
}
*x = xCoord;
*y = yCoord;
return;
}

void CartToPolar(double x, double y, double *angle, double *r)
// Function for converting a coordinate pair in Cartesian form to
// a coordinate pair in Polar form, by means of the formulas:
//
//                      y
//            tan A = ----
//                      x
//
//            r = SQR (x^2 + y^2)
//
// Where x and y are the Cartesian coordinates, r is the directed
// distance from the pole (radius) and A is the directed angle
// from the polar axis, expressed in degrees
// On entry:
//          parameters x and y hold the Cartesian coordinates
// On exit:
//          parameter angle is the polar angle
//          parameter r is the polar radius
{
double rValue, aValue;
_asm
{
	FLD    y       
    FLD    x
	CALL   CART_TO_POLAR
	FSTP   aValue
	FSTP   rValue
}
*angle = aValue;
*r = rValue;
return;
}

