; Module name: KEYBRD.ASM                     Library:   SOLUTION.LIB
; Created:
; Update:
;
;********************************************************************
;********************************************************************
;                             KEYBRD.ASM
;********************************************************************
;********************************************************************
;
; Library module for keyboard services
;
;*******************|
;      PUBLICS      |
;*******************|
PUBLIC	KBR_STATUS
PUBLIC	KBR_FLUSH
PUBLIC	KBR_WAIT
PUBLIC	DOS_KBR_INPUT
PUBLIC	KBR_INPUT
;
;********************************************************************
;                              code
;********************************************************************
;
P_CODE	SEGMENT PUBLIC
	ASSUME	CS:P_CODE
;
;*******************************************************************
KBR_STATUS	PROC	FAR
; On entry:
;         nothing
; On exit:
;         carry set if keystroke in buffer
;
	MOV	AH,1		; BIOS service request numbers
	INT	16H
	JNZ	PRESSED
; No key waiting
	CLC
	RET
PRESSED:
	STC
	RET
KBR_STATUS	ENDP
;********************************************************************
;
KBR_FLUSH	PROC	FAR
; Flush old characters from keyboard buffer
; On entry:
;         nothing
; On exit:
;         carry clear
FLUSH_1:
	MOV	AH,1		; BIOS service request number
	INT	16H
	JZ	NOCHAR		; OK to exit, buffer clear
; Flush old character
	MOV	AH,0		; BIOS service request number
	INT	16H
	JMP	FLUSH_1
NOCHAR:
	CLC			; No error detection
	RET
KBR_FLUSH	ENDP
;********************************************************************
;
KBR_WAIT	PROC	FAR
; Wait for one keyboard character
; On entry:
;         nothing
; On exit:
;         AL = keyboard character
;         AH = keyboard scan code
;
	MOV	AH,0		; Service request number
	INT	16H
	RET
KBR_WAIT	ENDP
;********************************************************************
;
DOS_KBR_INPUT	PROC	FAR
;
; Read a line from the keyboard and place the input in a buffer
; designated by the caller
;
; Buffer structure:
;        Offset 0 = Maximum number of characters to input
;                   (provided by the caller)
;        Offset 1 = Character count actually input, excluding
;                   the CR (this area is filled by DOS)
; On entry:
;          DX --> buffer start
; On exit:
;          Caller's buffer contains keyboard input and CR
;
	MOV	AH,10
	INT	21H
	RET
DOS_KBR_INPUT	ENDP
;********************************************************************
;
KBR_INPUT	PROC	FAR
;
; General keyboard input routine with the following parameters:
; Control codes and exit codes:
;               0DH <Enter> .... Exit with carry clear
;               011BH <Esc> .... Exit with carry set
;               08H <Backspace>. Erase character and backspace
;                                cursor
; On entry:
;          DI --> storage buffer
;          SI --> Input format area:
;                 Offset 0 = screen start row
;                 Offset 1 = screen start column
;                 Offset 2 = total characters allowed
; On exit:
;          Text is stored in buffer by DI
;                 DI --> start of character buffer
;                 CX = total characters input
;                 carry clear
;
	PUSH	DX		; Save entry DX
	MOV	DH,[SI]		; Get display row
	INC	SI		; Bump to offset 2
	MOV	DL,[SI]		; Get display column
	CALL	SET_CUR		; Local procedure
	INC	SI		; Bump to offset 3
	MOV	CL,[SI]		; Set-up counter
	MOV	CH,CL		; Copy start value in CH
; Save entry value of buffer pointer in SI
	MOV	SI,DI
;
GET_KEY:
	CALL	KBR_WAIT	; Get character
;******************|
; input processing |
;******************|
	CMP	AL,0DH		; <Enter> key
	JNE	NOT_0DH
	JMP	KBR_EXIT_0	; Take exit
NOT_0DH:
	CMP	AX,011BH	; <Esc> key
	JNE	NOT_ESC
	JMP	KBR_EXIT_1	; Take exit
NOT_ESC:
	CMP	AL,08H		; <Backspace> key
	JNE	NOT_BAK
;*****************|
;    backspace    |
;*****************|
; Test for cursor at start of buffer
	CMP	CH,CL		; Test present count with start
				; value
	JE	GET_KEY		; Ignore backspace at start position
; Execute backspace
	DEC	DI		; Buffer pointer
	DEC	DL		; Display column counter
	CALL	SET_CUR
	INC	CL		; Adjust maximum characters count
	MOV	AL,' '          ; Blank space
	CALL	SHOW_CHR	; Display a blank
	MOV	[DI],AL		; and put blank in buffer
	JMP	GET_KEY		; Continue
;******************|
; test for errors  |
;******************|
NOT_BAK:
; Test for invalid input (less than 20H or more than 79H)
	CMP	AL,20H
	JL	GET_KEY		; Illegal, too small
	CMP	AL,79H
	JG	GET_KEY		; Illegal, too large
; Test for buffer full
	CMP	CL,0		; CL is counter
	JNZ	DISPLAY_IT
	JMP	GET_KEY		; Buffer is full
;******************|
;  display char.   |
;******************|
DISPLAY_IT:
	CALL	SHOW_CHR
; Store it in buffer
	MOV	[DI],AL
;******************|
;  bump pointers   |
;******************|
	INC	DI		; Bump buffer pointer
	INC	DL		; Bump display column
	CALL	SET_CUR
	DEC	CL		; Character counter
	JMP	GET_KEY		; Continue
;
KBR_EXIT_0:
	PUSH	SI		; Save start of buffer
	SUB	DI,SI		; Get total input
	MOV	CX,DI		; into counter
	POP	DI		; Restore buffer start pointer
	CLC			; <Enter> key exit, no carry
	POP	DX		; Restore entry DX
	RET
KBR_EXIT_1:
	PUSH	SI		; Save start of buffer
	SUB	DI,SI		; Get total input
	MOV	CX,DI		; into counter
	POP	DI		; Restore buffer start pointer
	STC			; <Esc> key exit, carry set
	POP	DX		; Restore entry DX
	CLC			; No error detection
	RET
KBR_INPUT	ENDP
;********************************************************************
;                           local procedures
;********************************************************************
SET_CUR	PROC	NEAR
; Set system cursor to a row and column coordinate
; On entry:
;         DH = screen row (range 0 to 24)
;         DL = screen column (range 0 to 79)
; On exit:
;         carry clear if no error
;
	PUSH	AX		; Save entry registers
	PUSH	BX
	PUSH	DX
	MOV	AH,2		; Service request number
	MOV	BH,0		; Assume display page No. 0
	INT	10H
	POP	DX		; Restore registers
	POP	BX
	POP	AX
	RET
SET_CUR	ENDP
;*******************************************************************
;
SHOW_CHR	PROC	NEAR
; Display the character in AL at the current cursor position
; On entry:
;         AL = character
; On exit:
;         carry clear
;
	PUSH	AX		; Save caller's registers
	PUSH	BX
	PUSH	CX
	MOV	AH,10		; Service request number
	MOV	BH,0		; Display page zero
	MOV	CX,1		; One character
	INT	10H
	POP	CX		; Restore caller's registers
	POP	BX
	POP	AX
	RET
SHOW_CHR	ENDP
;
P_CODE	ENDS
	END
;