;*****************************************************************
;*****************************************************************
;                           TEST_GRF.ASM
;*****************************************************************
;*****************************************************************
; Description:
;            Test and demonstration program for the graphics
;            procedures in the UNIT_8 module of the MATHUNIT
;            library
;
;*****************************************************************
EXTRN   DEG_TO_RAD:     FAR
EXTRN   TANGENT:        FAR
;
EXTRN   CLS_18:         FAR
EXTRN   BRESENHAM:      FAR
EXTRN   LINE_BY_SLOPE:  FAR
EXTRN   CIRCLE:         FAR
EXTRN   ELLIPSE:        FAR
EXTRN   PARABOLA:       FAR
EXTRN   HYPERBOLA:      FAR
EXTRN   SET_CURSOR:     FAR
EXTRN   SET_VIDEO_MODE: FAR
EXTRN   SET_WRITE_MODE: FAR
EXTRN   ES_TO_VIDEO:    FAR
EXTRN   VGA18_TEXT:     FAR
EXTRN   DO_4_QUADS:     FAR
EXTRN   SHOW_CHR:       FAR
;
INCLUDELIB MATHUNIT
INCLUDELIB SOLUTION
;****************************************************************
;                          stack segment
;****************************************************************
STACK  SEGMENT stack

               DB      0400H DUP ('?'); Default stack is 1K
STACK_TOP      EQU     THIS BYTE
;
STACK  ENDS
;
;****************************************************************
;                          data segment
;****************************************************************
DATA    SEGMENT
;
TEXT_MESS_1     DB      28      ; Display row
                DB      10      ; Display column
                DB      10000111B       ; Black XORed
        DB      'FPU graphics in VGA mode 18',0FFH
        DB      'Demo program for "Numerical Programming"'
        DB      ' by Sanchez and Canton',00H

;
DATA    ENDS
;
;****************************************************************
;                          code segment
;****************************************************************
;
CODE    SEGMENT
        ASSUME  CS:CODE

START:
;*******************|
;  initialization   |
;*******************|
; Establish data and extra segment addressability
        MOV     AX,DATA         ; Address of DATA to AX
        MOV     DS,AX           ; and to DS
        ASSUME  DS:DATA         ; Assume from here on
;***********************************|
;           program code            |
;***********************************|
        CALL    ES_TO_VIDEO
; ES --> start of video buffer in graphics mode
; Initialize 8087 hardware component
        FINIT
        MOV     AL,18           ; VGA 640-by-480 pixels
        CALL    SET_VIDEO_MODE  ; in 16 colors
        MOV     AL,2            ; Write mode
        CALL    SET_WRITE_MODE
        MOV     AL,00000111B    ; Grey color bit pattern
        CALL    CLS_18          ; Clear screen to gray
;
;***************************|
;        black lines        |
;***************************|
; Leave 60 pixels clear on the horizontal axis
;          AL = color of line
;          BX = start point x coordinate
;          CX = start point y coordinate
;          SI = end point x coordinate
;          DI = end point y coordinate
        MOV     BX,60
        MOV     CX,240
        MOV     SI,580
        MOV     DI,240
        MOV     AL,00000000B    ; Black
        CALL    BRESENHAM
;
; Leave 30 pixels clear on vertical axis
        MOV     BX,320
        MOV     CX,30
        MOV     SI,320
        MOV     DI,450
        MOV     AL,00000000B    ; Black
        CALL    BRESENHAM
;
;***************************|
;        red circles        |
;***************************|
        MOV     BX,150
        CALL    CIRCLE
;           AL = color code in IRGB format (for mode 18)
;           BX = x coordinate of origin
;           CX = y coordinate of origin
        MOV     BX,320
        MOV     CX,240
        MOV     AL,00000100B    ; Red
        CALL    DO_4_QUADS
;***************************|
;      blue ellipses        |
;***************************|
        MOV     BX,150
        MOV     CX,100
        CALL    ELLIPSE
;
        MOV     BX,320
        MOV     CX,240
        MOV     AL,00000001B    ; Blue
        CALL    DO_4_QUADS
;
        MOV     BX,100
        MOV     CX,150
        CALL    ELLIPSE
        MOV     BX,320
        MOV     CX,240
        MOV     AL,00000001B    ; Blue
        CALL    DO_4_QUADS
;***************************|
;      green hyperbolas     |
;***************************|
        MOV     BX,100                  ; Focus
        MOV     CX,60                   ; Vertex
        MOV     DX,150                  ; x start
        CALL    HYPERBOLA
;
        MOV     BX,320
        MOV     CX,240
        MOV     AL,00000010B
        CALL    DO_4_QUADS
;***************************|
;      brown parabolas      |
;***************************|
        MOV     BX,20                   ; Focus
        MOV     CX,140                  ; x start
        CALL    PARABOLA
;
        MOV     BX,320
        MOV     CX,240
        MOV     AL,00000110B
        CALL    DO_4_QUADS
;***************************|
;     lines by slope        |
;***************************|
        MOV     AX,320          ; x of start point
        MOV     BX,240          ; y of start point
        MOV     CX,500          ; x of end point
        MOV     DX,45           ; Slope (in degrees)
        CALL    LINE_BY_SLOPE
;
        MOV     BX,320
        MOV     CX,240
        MOV     AL,00000000B
        CALL    DO_4_QUADS
;***************************|
;  text in graphics mode    |
;***************************|
; Set write mode 0 for text display
        MOV     AL,0
        CALL    SET_WRITE_MODE
; Display text message
        LEA     DI,TEXT_MESS_1
        CALL    VGA18_TEXT
;***************************|
;     display characters    |
;***************************|
        MOV     DH,0    ; Row 0
        MOV     DL,40   ; Screen center
        CALL    SET_CURSOR
        MOV     AL,'y'          ; y axis designation
        MOV     BL,10000111B    ; Black color bits xored
        CALL    SHOW_CHR
;
        MOV     DH,14   ; Screen center
        MOV     DL,75
        CALL    SET_CURSOR
        MOV     AL,'x'          ; x axis designation
        MOV     BL,10000111B    ; Black color bits xored
        CALL    SHOW_CHR
;***************************|
;      wait for keystroke   |
;***************************|
        MOV     AH,0
        INT     16H
;*******************|
;   exit to DOS     |
;*******************|
DOS_EXIT:
        MOV     AL,7            ; VGA text mode
        CALL    SET_VIDEO_MODE  ; in 16 colors
        MOV     AH,4CH          ; DOS service request code
        MOV     AL,0            ; No error code returned
        INT     21H             ; TO DOS


CODE    ENDS
        END     START

AL,