;*****************************************************************
;*****************************************************************
;                           TEST_OUT.ASM
;*****************************************************************
;*****************************************************************
;
; Description:
;           Test and demonstration program for the NUM_AT_ST0
;           procedure in the UNIT_2 module of the MATHUNIT library
;
;*****************************************************************
EXTRN   NUM_AT_ST0:     FAR
EXTRN   FPU_OUTPUT:     FAR
;
EXTRN   KBR_WAIT:       FAR
;
INCLUDELIB MATHUNIT
INCLUDELIB SOLUTION
;****************************************************************
;                          stack segment
;****************************************************************
STACK  SEGMENT stack

               DB      0400H DUP ('?'); Default stack is 1K
STACK_TOP      EQU     THIS BYTE
;
STACK  ENDS
;
;****************************************************************
;                          data segment
;****************************************************************
DATA    SEGMENT
; Error message area
ERROR_MSG       DB      '                              '
                DB      0AH,0DH,'$'
;
; Number types to be tested
NORMAL_POS      DT      1234.56
NORMAL_NEG      DT      -1234.56
;
ZERO_POS        DT      00000000000000000000H
ZERO_NEG        DT      80000000000000000000H
DENORMAL_POS    DT      000000000000000000FFH
DENORMAL_NEG    DT      800000000000000000FFH
NAN_POS         DT      7FFF80000000000000FFH
NAN_NEG         DT     0FFFF80000000000000FFH
INFINITY_POS    DT      7FFF8000000000000000H
INFINITY_NEG    DT     0FFFF8000000000000000H
;                       |  |               |
;                       |  |               |
; sign and exponent ------^|       ^----------- significand
;
SAMPLE          DD      42FEC000H
DUMMY           DD      0
DENORMAL_BIG    DT      00007FFFFFFFFFFFFFFFH
;                       |  |               |
; sign and exponent ------^|       ^----------- significand

;
DATA    ENDS
;
;****************************************************************
;                          code segment
;****************************************************************
;
CODE    SEGMENT
        ASSUME  CS:CODE
;
START:
;*******************|
;  initialization   |
;*******************|
; Establish data and extra segment addressability
        MOV     AX,DATA         ; Address of DATA to AX
        MOV     DS,AX           ; and to DS
        ASSUME  DS:DATA         ; Assume from here on
; Initialize stack. Note that this instruction is necessary for
; compatibility with non-Microsoft assemblers
        LEA     SP,SS:STACK_TOP ; Stack pointer to top of stack
;***************************|
;   testing NUM_AT_ST0      |
;***************************|
; Code assumes the presence of a FPU
        FINIT                   ; Initialize FPU
        CALL    SHOW_ST0
;
        FLD     NORMAL_POS
        CALL    SHOW_ST0
        FSTP    ST(0)
;
        FLD     NORMAL_NEG
        CALL    SHOW_ST0
        FSTP    ST(0)
;
        FLD     ZERO_POS
        CALL    SHOW_ST0
        FSTP    ST(0)
;
        FLD     ZERO_NEG
        CALL    SHOW_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR DENORMAL_POS
        CALL    SHOW_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR DENORMAL_NEG
        CALL    SHOW_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR NAN_POS
        CALL    SHOW_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR NAN_NEG
        CALL    SHOW_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR INFINITY_POS
        CALL    SHOW_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR INFINITY_NEG
        CALL    SHOW_ST0
        FSTP    ST(0)
;
;***************************|
;   testing FPU_OUTPUT      |
;***************************|
;
        CALL    KBR_WAIT        ; Procedure in SOLUTION.LIB
;
        FINIT                   ; Initialize FPU
        CALL    OUTP_ST0
;
        FLD     NORMAL_POS
        CALL    OUTP_ST0
        FSTP    ST(0)
;
        FLD     NORMAL_NEG
        CALL    OUTP_ST0
        FSTP    ST(0)
;
        FLD     ZERO_POS
        CALL    OUTP_ST0
        FSTP    ST(0)
;
        FLD     ZERO_NEG
        CALL    OUTP_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR DENORMAL_POS
        CALL    OUTP_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR DENORMAL_NEG
        CALL    OUTP_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR NAN_POS
        CALL    OUTP_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR NAN_NEG
        CALL    OUTP_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR INFINITY_POS
        CALL    OUTP_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR INFINITY_NEG
        CALL    OUTP_ST0
        FSTP    ST(0)
;
        JMP     DOS_EXIT
;
; Testing a denormal in single format
TEST_1:
        FLD     TBYTE PTR DENORMAL_POS
        CALL    SHOW_ST0
        FSTP    DUMMY           ; Store as single
        FWAIT
        FLD     DWORD PTR DUMMY ; Reload
        CALL    SHOW_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR DENORMAL_BIG
        CALL    SHOW_ST0
        FSTP    DUMMY           ; Store as single
        FWAIT
        FLD     DWORD PTR DUMMY ; Reload
        CALL    SHOW_ST0
        FSTP    ST(0)
;
        FLD     TBYTE PTR NAN_POS
        CALL    SHOW_ST0
        FSTP    DUMMY           ; Store as single
        FWAIT
        FLD     DWORD PTR DUMMY ; Reload
        CALL    SHOW_ST0
        FSTP    ST(0)
;*******************|
;   exit to DOS     |
;*******************|
DOS_EXIT:
        MOV     AH,4CH          ; DOS service request code
        MOV     AL,0            ; No error code returned
        INT     21H             ; TO DOS
;
;*****************************************************************
;                      auxiliary  procedures
;*****************************************************************
SHOW_ST0        PROC    NEAR
; Test and display type for number at ST(0)
        LEA     DI,ERROR_MSG    ; Pointer to caller's buffer
        CALL    NUM_AT_ST0      ; Library procedure to identify
                                ; number type at ST(0)
; Display message
        LEA     DX,ERROR_MSG    ; Pointer to DX register
        MOV     AH,09           ; Code for MS DOS service
        INT     21H
        RET
SHOW_ST0        ENDP
;
OUTP_ST0        PROC    NEAR
; Test and display type for number at ST(0)
        LEA     DX,ERROR_MSG    ; Pointer to caller's buffer
        PUSH    DX              ; Save pointer
        XCHG    DX,SI           ; Set up new pointer
; Clear 30 bytes in caller's buffer
        MOV     CX,30           ; Byte counter
DO_30BYTES:
        MOV     BYTE PTR [SI],' '
        INC     SI
        LOOP    DO_30BYTES
        POP     DX
;
        INT     03H
;
        CALL    FPU_OUTPUT      ; Library procedure to identify
                                ; number type at ST(0)
; Display message
        LEA     DX,ERROR_MSG    ; Pointer to DX register
        MOV     AH,09           ; Code for MS DOS service
        INT     21H
        RET
OUTP_ST0        ENDP
;
CODE    ENDS
        END     START           ; Reference to label at which
                                ; execution begins
DE    ENDS
        END     START           ; Reference to label at which
                                ; execution begins
