;*****************************************************************
;*****************************************************************
;                           UNIT_3.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date:  January 13, 1994
; Updated:        January 25, 1994
;                 February 5, 1994
;
; Libary name: MATHUNIT.LIB
;
;*****************************************************************
;                           publics
;*****************************************************************
; MATHEMATICAL ROUTINES IN THIS MODULE
PUBLIC SIGN_MUL_BCD12   ; .. Signed BCD multiplication
PUBLIC SIGN_DIV_BCD12   ; .. Signed BCD division
PUBLIC GET_TAG          ; .. Return FPU register tag
;
;*****************************************************************
;                           externals
;*****************************************************************
; PROCEDURES IN OTHER MODULES OF MATHUNIT.LIB
EXTRN   BCD_INTO_REG:   FAR
EXTRN   BIN_TO_BCD:     FAR
;
;****************************************************************
;                          code segment
;****************************************************************
;
CODE    SEGMENT BYTE
        ASSUME  CS:CODE
;
;***************************|
;   code segment data for   |
; procedures in this module |
;***************************|
; Scratch pad area for routine work variables
ENTRY_SI        DW      0       ; Storage for entry pointer SI
ENTRY_DI        DW      0       ; Storage for entry pointer DI
ENTRY_BX        DW      0       ; Storage for entry pointer BX
; Biased exponents
BIAS_X          DW      0       ; Biased exponent for x
BIAS_Y          DW      0       ; Biased exponent for y
BIAS_Z          DW      0       ; Biased exponent for sum
; Work areas for significand operations
SIG_L   DB      36 DUP (00H)    ; Larger significand area
SIG_S   DB      36 DUP (00H)    ; Smaller significand area
SIG_R   DB      37 DUP (00H)    ; Significand of result
                                ; Extra byte holds possible carry
; Others
LARGER  DB      0               ; 'x' or 'y', whichever is larger
CASE_12 DB      0               ; Switch for CASE 1 or CASE 2
; Partial products buffer for multiplication operations
PART_PROD       DB      37 DUP (00H)
                                ; Buffer for partial products
; Buffers for BCD division
DIVISOR         DB      18 DUP (00H)    ; Buffer for divisor
                DB      0FFH            ; Buffer terminator
DIVIDEND        DB      19 DUP (00H)    ; Buffer for dividend
                DB      0FFH            ; Buffer terminator
NEW_DIVIDEND    DB      19 DUP (00H)    ; Buffer for trial dividend
                DB      0FFH            ; Buffer terminator
TRIAL_PROD      DB      19 DUP (00H)    ; Buffer for trial product
                DB      0FFH            ; Buffer terminator
QUOTIENT        DB      37 DUP (00H)    ; Buffer for quotient
                DB      0FFH            ; Buffer terminator
;*******************************************************************
;                     BCD arithmetic routines
;*******************************************************************
;************************************|
; signed floating point BCD multiply |
;************************************|
SIGN_MUL_BCD12  PROC    FAR
; Multiplication of two signed BCD numbers stored in BCD12 floating
; point format
; Operation:
;           z = x * y
;           where x, y, and z are signed, floating point numbers
; On entry:
;          DS:SI --> Muliplicand (element x in z = x * y)
;          DS:DI --> Multiplier (element y in z = x * y)
;          DS:BX --> 12-byte storage area for product in BCD12 format
; Note: the code assumes that the BCD12 numbers are in normalized
;       form, that is, that there are no leading zeros in the
;       significand
; On exit:
;       DS:BX --> Product (element z in z = x * y)
;       Entry pointers are preserved
;*******************************************************************
; BCD signed multiplication algorithm
; A. If one of the factors is zero, the product is zero
;    If neither factor is zero then:
; B. If the factors have equal signs the product is positive
;    If the factors have unequal sings the product is negative
; C. The significand of the product is the product of the
;    significands of the factors
; D. The exponent of the product is the sum of the exponents of
;    the factors
;*******************************************************************
; Store entry variables and clear buffers
        CALL    STORE_AND_CLEAR         ; Local procedure
;**********************|
;    test x or y = 0   |
;**********************|
; If x or y = 0 then the product is zero
        CALL    TEST_4_ZERO     ; Test routine
        JNC     NOT_ZERO_X      ; Go if no carry
; At this point x is zero
MOVE_00:
        MOV     CX,12           ; Count for digits to move
MOVE_12_00:
        MOV     BYTE PTR[BX],00 ; Store zero by BX pointer
        INC     BX              ; Bump pointer
        LOOP    MOVE_12_00
;**********************************|
;    exit from multiplication      |
;**********************************|
SIGN_MUL_EXIT:
        MOV     SI,CS:ENTRY_SI  ;Restore entry pointers
        MOV     DI,CS:ENTRY_DI
        MOV     BX,CS:ENTRY_BX
        RET                     ; End of processing
;**********************|
;    test for y = 0    |
;**********************|
NOT_ZERO_X:
        PUSH    SI              ; Save x pointer
        MOV     SI,DI           ; Pointer to SI
        CALL    TEST_4_ZERO     ; Test routine
        POP     SI              ; Restore x pointer
        JNC     NOT_ZERO_XY     ; Continue if no carry
        JMP     MOVE_00         ; Exit routine if y = 0
;**********************|
;  neither x nor y = 0 |
;     add exponents    |
;**********************|
NOT_ZERO_XY:
        CALL    EXP_2_BIAS      ; Conversion procedure
; DX now holds binary exponent for x in bias 10,000 form
        MOV     CS:BIAS_X,DX    ; Store in scratch pad area
; Obtain biased exponent for y
        XCHG    SI,DI           ; Save x and set y
        CALL    EXP_2_BIAS      ; Conversion procedure
        MOV     CS:BIAS_Y,DX    ; Store in scratch pad area
        XCHG    SI,DI           ; Restore x and y pointers
;**********************|
;    add exponents     |
;**********************|
        MOV     AX,CS:BIAS_X    ; Exponent of x to AX
        ADD     AX,DX           ; Add exponents of x and y
        SUB     AX,10000        ; Subtract bias from sum
                                ; to re-scale exponent
        INC     AX              ; Normalize siginifcand by
                                ; incrementing the exponent
        MOV     CS:BIAS_Z,AX    ; Store sum in scratch pad area
;**********************|
;    unpack and move   |
;      significands    |
;**********************|
; Move 18 packed BCD digits from BCD12 encoding to SIG_L buffer
        MOV     SI,CS:ENTRY_SI  ; Set pointer to x
        ADD     SI,3            ; Mantissa digits in BCD12 encoding
        LEA     DI,CS:SIG_L     ; Offset of CS buffer
        MOV     CX,9            ; 9 packed digits to move
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
; Move significand of y to SIG_S buffer
        MOV     SI,CS:ENTRY_DI  ; Set pointer to y
        ADD     SI,3            ; Mantissa digits in BCD12 encoding
        LEA     DI,CS:SIG_S     ; Offset of CS buffer
        MOV     CX,9            ; 9 packed digits to move
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
;**********************|
; multiply significands|
;**********************|
;*******************************************************************
; Structures for multidigit unpacked BCD multiplication
; 1. PART_PROD is a 37-digit buffer to hold the product or the
;    multiplicand times each digit in the multiplier. The operation
;    can be visualized as follows:
;       COLUMN NUMBERS
;       3 3 3 3 3 3 3 3 2 2 ..... 1 1 1 1 1 1 1 1 1
;       7 6 5 4 3 2 1 0 9 8 ..... 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1
;                                 x x x x x x x x x x x x x x x x x x
;                                 y y y y y y y y y y y y y y y y y y
; ROWS --------------------------------------------------------------
; 1     0 0 0 0 0 0 0 0 0 0 ....C z z z z z z z z z z z z z z z z z z
; 2     0 0 0 0 0 0 0 0 0 ... C z z z z z z z z z z z z z z z z z z 0
; 3     0 0 0 0 0 0 0 0 ... C z z z z z z z z z z z z z z z z z z 0 0
; 4     0 0 0 0 0 0 0 ... C z z z z z z z z z z z z z z z z z z 0 0 0
; 17    0 C z z z z ... z z z z z z 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
; 18    C z z z z ... z z z z z z 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
;       -------------------------------------------------------------
;       C p p p p ... p p p p p p p p p p p p p p p p p p p p p p p p
;
; In the actual routine the PART_PROD buffer contains a single row
; of digits. This product row is added to the previous contents of
; the SIG_R buffer before the next partial product is calculated
;*******************************************************************
; At this point the significands of x and y are in the buffers SIG_L
; and SIG_S
        LEA     DI,CS:SIG_S     ; DI --> to multiplier
        ADD     DI,17           ; Point SI to last multiplier digit
        MOV     BP,1            ; BP is counter for 18 rows of digits
;**********************|
; clear product buffer |
;**********************|
ONE_ROW_OP:
        LEA     SI,CS:SIG_L     ; SI --> to multiplicand
        ADD     SI,17           ; Point SI to end of multiplicand
        LEA     BX,CS:PART_PROD ; BX --> buffer for partial products
        MOV     CX,37           ; Bytes to clear
        PUSH    BX              ; Save buffer start address
CLEAR_PP:
        MOV     BYTE PTR CS:[BX],0H     ; Store a zero digit
        INC     BX              ; Bump buffer pointer
        LOOP    CLEAR_PP
; Offset of rightmost digit in PART_PROD is 37 minus the row
; count. Row count is in BP. The offset of the last digit in the
; first partial product is BX + (37-1) = BX + 36
        POP     BX              ; BX --> start of products buffer
        MOV     AX,37           ; Offset of last buffer digit
        SUB     AX,BP           ; Minus current row
        ADD     BX,AX           ; Add offset to pointer
;**********************|
;    multiply row      |
;**********************|
MULT_DIGIT:
        MOV     DL,CS:[DI]      ; Multiplier digit
        MOV     DH,0            ; Register for carry
        MOV     CX,18           ; Counter for 18 digits
; At this point:
; SI --> rightmost digit in the multiplicand
; DL = multiplier digit
; BX --> last digit offset into partial products buffer
MUL_18_DIGITS:
        MOV     AH,0            ; Clear high digit for product
        MOV     AL,BYTE PTR CS:[SI]     ; Multiplicand digit to AL
        MUL     DL              ; AX = AL * DL
        ADD     AL,DH           ; Add in carry from last product
        AAM                     ; Adjust product to unpacked BCD
        MOV     BYTE PTR CS:[BX],AL     ; Store digit in buffer
        MOV     DH,AH           ; Carry into next digit
        DEC     BX              ; Bump to next storage position
        DEC     SI              ; and to next multiplicand digit
        LOOP    MUL_18_DIGITS
; 18 digits multiplied. Store possible carry
        MOV     BYTE PTR CS:[BX],DH     ; Digit to buffer
; At this point PART_PROD holds product of multiplicand times
; current multiplier digit
;**********************|
;      add row         |
;**********************|
; PART_PROD is now added to the accumulated sum stored in SIG_R
; Both buffers have 37 digits
        PUSH    DI              ; Save pointer to multiplier
        LEA     SI,CS:PART_PROD ; Pointer to first digit row
        LEA     DI,CS:SIG_R     ; Pointer to destination of sum
        ADD     SI,36           ; Point SI to last digit in buffer
        ADD     DI,36           ; also DI
        MOV     CX,37           ; 37 digits to add
        CLC                     ; Clear carry for first addition
ADD_ROW:
        MOV     AL,BYTE PTR CS:[SI]     ; One digit to AL
        ADC     AL,BYTE PTR CS:[DI]     ; Add other digit
        AAA                     ; Convert sum to unpacked BCD
        MOV     BYTE PTR CS:[DI],AL     ; Store digit in same buffer
        DEC     SI              ; Bump pointers to next digit
        DEC     DI
        LOOP    ADD_ROW         ; Loop for 37 digits
; Since the buffers contain 37 digits there can be no carry out of
; the last digit
        POP     DI              ; Restore multiplier pointer
;**********************|
;   multiply next row  |
;**********************|
        DEC     DI              ; Index to next multiplier
        INC     BP              ; Increment row counter
        CMP     BP,19           ; Test for all rows multiplied
        JE      DONE_18         ; Go if at end of operation
        JMP     ONE_ROW_OP
; At this point the SIG_R buffer contains the product of the
; significands
;*************************************|
;   Construct BCD12 number and exit   |
;*************************************|
DONE_18:
        CALL    RESTORE_BCD     ; Local procedure used by multiply
                                ; and divide rouitnes
        JMP     SIGN_MUL_EXIT
SIGN_MUL_BCD12  ENDP
;*******************************************************************
;************************************|
; signed floating point BCD division |
;************************************|
SIGN_DIV_BCD12  PROC    FAR
; Division of two signed BCD numbers stored in BCD12 floating point
; format
; Operation:
;           z = x / y
;           where x, y, and z are signed, floating point numbers
; On entry:
;          DS:SI --> Dividend (element x in z = x / y)
;          DS:DI --> Divisor (element y in z = x / y)
;          DS:BX --> 12-byte storage area for quotient
; Note: the code assumes that the BCD12 numbers are in normalized
;       form, that is, that there are no leading zeros in the
;       significand
; On exit:
;       DS:BX --> Quotient (element z in z = x / y)
;       carry set if divisor equals zero
;       carry clear if divisor not equal zero
;       Entry pointers are preserved
;*******************************************************************
; BCD signed division algorithm
; A. If the dividend is zero the quotient is zero
;    If the divisor is zero the operation is undefined
;    If neither element is zero then:
; B. If the factors have equal signs the quotient is positive
;    If the factors have unequal sings the quotient is negative
; C. The significand of the quotient is the quotient resulting from
;    dividing the significand of the dividend by the siginificand of
;    the divisor
; D. The exponent of the quotient is the difference between the
;    exponent of the dividend minus the exponent of the divisor
;*******************************************************************
;**********************|
;    clear division    |
;        buffers       |
;**********************|
        PUSH    BX              ; Save pointer passed by caller
        LEA     BX,CS:DIVIDEND          ; First of 5 buffers
        MOV     CX,19           ; bytes to clear
        CALL    CLEAR_CS_BUF    ; Local procedure to clear buffer
        LEA     BX,CS:DIVISOR           ; Next buffer
        MOV     CX,18           ; bytes to clear
        CALL    CLEAR_CS_BUF    ; Local procedure to clear buffer
        LEA     BX,CS:TRIAL_PROD        ; Next buffer
        MOV     CX,19           ; bytes to clear
        CALL    CLEAR_CS_BUF    ; Local procedure to clear buffer
        LEA     BX,CS:NEW_DIVIDEND      ; Next buffer
        MOV     CX,19           ; bytes to clear
        CALL    CLEAR_CS_BUF    ; Local procedure to clear buffer
        LEA     BX,CS:QUOTIENT          ; Last buffer
        MOV     CX,37           ; bytes to clear
        CALL    CLEAR_CS_BUF    ; Local procedure to clear buffer
        POP     BX              ; Restore caller's pointer
;**********************|
; save entry variables |
;**********************|
        CALL    STORE_AND_CLEAR         ; Local procedure
;**********************|
; test for dividend = 0|
;**********************|
; If the dividiend = 0 then the quotient = 0
        CALL    TEST_4_ZERO     ; Test routine
        JNC     NOT_ZERO_DID    ; Go if no carry
; At this point x is zero
MOVE_0_QUOT:
        MOV     CX,12           ; Count for digits to move
MOVE_12_QUOT:
        MOV     BYTE PTR[BX],00 ; Store zero by BX pointer
        INC     BX              ; Bump pointer
        LOOP    MOVE_12_QUOT
;**********************************|
;   normal exit from division      |
;**********************************|
SIGN_DIV_EXIT:
        MOV     SI,CS:ENTRY_SI  ;Restore entry pointers
        MOV     DI,CS:ENTRY_DI
        MOV     BX,CS:ENTRY_BX
        CLC                     ; Divisor is not zero
        RET                     ; End of processing
;**********************|
; test for divisor = 0 |
;**********************|
; Division is undefined if the divisor = 0
; This is signaled to the caller with the carry flag set
NOT_ZERO_DID:
        PUSH    SI              ; Save x pointer
        MOV     SI,DI           ; Pointer to SI
        CALL    TEST_4_ZERO     ; Test routine
        POP     SI              ; Restore x pointer
        JNC     NOT_ZERO_ELMS   ; Continue if no carry
;**********************************|
;    error exit from division      |
;**********************************|
; Divisor = zero
        MOV     SI,CS:ENTRY_SI  ;Restore entry pointers
        MOV     DI,CS:ENTRY_DI
        MOV     BX,CS:ENTRY_BX
        STC                     ; Zero divisor. Operation undefined
        RET                     ; End of processing
;**********************|
;  neither x nor y = 0 |
;  subtract exponents  |
;**********************|
NOT_ZERO_ELMS:
        CALL    EXP_2_BIAS      ; Conversion procedure
; DX now holds binary exponent for x in bias 10,000 form
        MOV     CS:BIAS_X,DX    ; Store in scratch pad area
; Obtain biased exponent for y
        XCHG    SI,DI           ; Save x and set y
        CALL    EXP_2_BIAS      ; Conversion procedure
        MOV     CS:BIAS_Y,DX    ; Store in scratch pad area
        XCHG    SI,DI           ; Restore x and y pointers
;**********************|
;  subtract exponents  |
;**********************|
        MOV     AX,CS:BIAS_X    ; Exponent of x to AX
        SUB     AX,DX           ; Subtract exponent of x - y
        ADD     AX,10000        ; Add bias to difference
                                ; to re-scale exponent
        MOV     CS:BIAS_Z,AX    ; Store sum in scratch pad area
;*******************************************************************
; Buffers used in the division of significands:
;        DIVIDEND     (--> DI)        19 bytes
;        NEW_DIVIDEND                 19 bytes
;        DIVISOR      (--> SI)        18 bytes
;        TRIAL_PROD                   19 bytes
;        QUOTIENT                     37 bytes
;*******************************************************************
;**********************|
;    unpack and move   |
;      significands    |
;**********************|
; Move 18 packed BCD digits from BCD12 encoding to DIVIDEND buffer
        MOV     SI,CS:ENTRY_SI  ; Set pointer to x
        ADD     SI,3            ; Mantissa digits in BCD12 encoding
        LEA     DI,CS:DIVIDEND  ; Offset of CS buffer
        INC     DI              ; Right-justify dividend in buffer
        MOV     CX,9            ; 9 packed digits to move
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
; Move significand of y to DIVISOR buffer
        MOV     SI,CS:ENTRY_DI  ; Set pointer to y
        ADD     SI,3            ; Mantissa digits in BCD12 encoding
        LEA     DI,CS:DIVISOR   ; Offset of CS buffer
        MOV     CX,9            ; 9 packed digits to move
        CALL    UNPACK_MOVE     ; Unpack and move digits routine
;**********************|
; init quotiend digit  |
;    counter (BP)      |
;**********************|
        MOV     BP,0            ; BP register is counter for digits
                                ; in the quotient buffer
; Go to entry point for initial undersize dividend
        JMP     EXPU
;**********************|
; obtain trial divisor |
;**********************|
TEST_DIVIDE:
        CALL    TRIAL_QUOT      ; Local routine to find a trial
                                ; quotient digit
        MOV     DL,AL           ; Trial quotient digit to DL
TEST0:
        LEA     SI,CS:DIVISOR   ; Set pointer to divisor
        LEA     BX,CS:TRIAL_PROD        ; Pointer to trial product
        ADD     SI,17           ; To last digit in buffer
        ADD     BX,18           ; also this pointer
        MOV     DH,0            ; Clear carry register
        MOV     CX,18           ; Set counter for 18 digits
;**********************|
;  multiply dividend   |
; times trial quotient |
;**********************|
TRI0:
        MOV     AH,0            ; Clear high byte
        MOV     AL,CS:[SI]      ; Get high order divisor digit
        MUL     DL              ; Divisor times trial quotient
        ADD     AL,DH           ; Add possible carry to product
        AAM                     ; Convert product to BCD
        MOV     DH,AH           ; Save carry in DH
        MOV     CS:[BX],AL      ; Store digit as trial product
        DEC     BX              ; Bump pointers
        DEC     SI
        LOOP    TRI0            ; Continue for 18 digits
; At this point there can be a carry into the 19th digit
; Place possible carry or zero in buffer
        MOV     CS:[BX],DH      ; Store extra digit
;**********************|
;   subtract product   |
;    from dividend     |
;**********************|
RESTA:
        PUSH    DX              ; Save registers and pointers
        PUSH    SI
        PUSH    DI
        PUSH    BX
        LEA     BX,CS:NEW_DIVIDEND      ; Set context pointers
        LEA     SI,CS:TRIAL_PROD
        LEA     DI,CS:DIVIDEND
        ADD     BX,18           ; Index pointers to low order byte
        ADD     SI,18
        ADD     DI,18
        MOV     CX,19           ; Counter for 19 digits
        CLC                     ; Clear carry flag
; Perform subtraction operation
REST0:
        MOV     AL,CS:[DI]      ; Dividend digit to AL
        MOV     DL,CS:[SI]      ; Trial product digit to DL
        SBB     AL,DL           ; Dividend minus trial product
        AAS                     ; Convert to BCD
        MOV     CS:[BX],AL      ; Store result in buffer
        DEC     BX              ; Bump buffer pointers
        DEC     SI
        DEC     DI
        LOOP    REST0           ; Continue for 19 bytes
        POP     BX              ; Restore entry pointers
        POP     DI
        POP     SI
        POP     DX
; If there is a carry into the buffer's 19th digit then the trial
; quotient is too large
        JNC     PUTQ            ; Go if no carry (quotient digit OK)
;**********************|
;   invalid quotient   |
;        digit         |
;**********************|
; At this point the trial quotient must be adjusted and retried
; If the trial digit goes to zero, then it is stored as a quotient
; digit and processing continues
        DEC     DL              ; Decrement quotient digit
        JZ      PUTQ            ; Store zero
        JMP     TEST0           ; Retry new trial quotiend digit
;**********************|
; valid quotient digit |
;**********************|
; Processing:
; 1. Store quotient digit in buffer (DIGIT_TO_QUOT procedure)
; 2. Make the trial dividend (in NEW_DIVIDEND) the dividend
PUTQ:
        CALL    DIGIT_TO_QUOT   ; Store digit in buffer
        JC      SIG_DONE        ; Go if buffer full returned
; Move NEW_DIVIDEND to DIVIDEND multiplying by 10
        LEA     DI,CS:DIVIDEND  ; Pointer to dividend buffer
        LEA     SI,CS:NEW_DIVIDEND      ; Pointer to new dividend
        MOV     BYTE PTR CS:[SI],0H     ; Clear high byte
        INC     SI
        MOV     CX,18           ; Bytes to move from CS:SI to CS:DI
        CALL    TRANSFER_CS_CX  ; Local procedure
; New dividend is now in NEW_DIVIDEND buffer
; Chech size of dividend. Place zeros in buffer is the dividend
; is smaller than the divisor
EXPU:
        CALL    COMPARE_DI_DR   ; Compare divisor and dividend
        JC      EXPU1           ; Go if divisor > dividend
        JMP     TEST_DIVIDE     ; Continue processing
;**********************|
;   quotient = 0       |
;**********************|
EXPU1:
        MOV     DL,0            ; Digit to store
        CALL    DIGIT_TO_QUOT   ; Local procedure
        JC      SIG_DONE        ; Exit routine if carry set
; Multiply dividend by 10 by offsetting into buffer
        LEA     DI,CS:DIVIDEND  ; DI points to dividend area
        MOV     SI,DI           ; and so does SI
        INC     SI              ; Skip first digit (0) in source
                                ; effectively multiplying by 10
        MOV     CX,18           ; Bytes to move from CS:SI to CS:DI
        CALL    TRANSFER_CS_CX  ; Local procedure
;       MOV     BYTE PTR CS:[DI],0      ; Add zero at end of buffer
        JMP     EXPU
;**********************|
;  move significand    |
;**********************|
; Significand digits from QUOTIENT buffer to SIG_R buffer
SIG_DONE:
        LEA     SI,CS:QUOTIENT  ; Source buffer
        LEA     DI,CS:SIG_R     ; Destination buffer
        INC     DI              ; Skip first digit
        MOV     CX,36           ; Digits to move
        CALL    TRANSFER_CS_CX  ; Local procedure
        CALL    RESTORE_BCD     ; Local procedure used by multiply
                                ; and divide rouitnes
        JMP     SIGN_DIV_EXIT
SIGN_DIV_BCD12  ENDP
;
;*******************************************************************
;                     auxiliary procedures
;*******************************************************************
RESTORE_BCD     PROC    NEAR
; Reconstruct BCD number after multiplication and division operation
; Note that the sign rules for significand and exponents are the
; same
; On entry:
;         CS:BIAS_Z holds biased exponent of result
;         SIG_R buffer holds significand
; On exit:
;         Buffer at CS:ENTRY_BX contains results in BCD12 format
;
        CALL    ROUND_SIGF      ; Round off significand
        MOV     DI,CS:ENTRY_BX  ; Set pointer to result buffer
                                ; for BCD12 output
        MOV     BYTE PTR [DI],00H       ; Set default positive sign
        ADD     DI,3            ; Point to significand area
;**********************|
;   store significand  |
;**********************|
; Significand of result is unpacked BCD form in the first 18 digits
; of the SIG_R buffer
; DI --> significand field of the BCD12 result buffer
        MOV     CX,9            ; Counter for 18 packed digits
        LEA     SI,CS:SIG_R     ; Pointer to unpacked digits
        INC     SI              ; Skip leading zero
; Skip possible leftmost zero in buffer
        CMP     BYTE PTR CS:[SI],00H    ; Test for zero
        JNE     PACK_DIGITS     ; Go if no leading zero
        INC     SI              ; Skip zero
        DEC     CS:BIAS_Z       ; Adjust exponent
; In multiplication there can be only one leading zero in the buffer
; because the input significands are in normalized form
PACK_DIGITS:
        MOV     AH,BYTE PTR CS:[SI]     ; Get one digit
        INC     SI              ; Bump pointer
        MOV     AL,BYTE PTR CS:[SI]     ; Get next digit
        INC     SI              ; Bump pointer to next digit
        PUSH    CX              ; Save counter
        MOV     CL,4            ; Count for 4 digits
        SHL     AH,CL           ; Shift high digit to high nibble
        POP     CX              ; Restore counter
        OR      AL,AH           ; Merge high nibble in AH with low
                                ; nibble in AH (pack digits)
        MOV     BYTE PTR [DI],AL        ; Store packed digits
        INC     DI
        LOOP    PACK_DIGITS
; All digits moved to the destination BCD12 buffer
        MOV     BX,CS:ENTRY_BX  ; Set pointer to result buffer
                                ; sign area
;**********************|
;   sign of number     |
;**********************|
; Rule: if the factors have equal signs the product is positive
;       If the factors have unequal sings the product is negative
; The sign of the number is the high nibble of the first BCD12 digit
        MOV     SI,CS:ENTRY_SI  ; Set pointers to entry values
        MOV     DI,CS:ENTRY_DI
; SI --> x and DI --> y
        MOV     AL,BYTE PTR [SI]        ; x's first digit to AL
        MOV     AH,BYTE PTR [DI]        ; y's first digit to AH
        AND     AL,11110000B    ; Clear x's low nibble
        AND     AH,11110000B    ; and y's low nibble
        CMP     AL,AH           ; Test for equal sign
        JE      EXPN_SIGN       ; Numbers have equal sign
                                ; This encoding already in buffer
; At this point the factors have unequal sign. Product is negative
        MOV     BYTE PTR [BX],10H       ; Negative sign for number
;**********************|
; get sign of exponent |
;**********************|
EXPN_SIGN:
; Exponent of result is stored in BIAS_Z in bias 10000 form
; Values < 10000 indicate a negative exponent
        MOV     DX,CS:BIAS_Z    ; Exponent of result to DX
        CMP     DX,10000        ; Compare with bias
        JAE     POS_VALUE       ; Go if positive exponent
; Positive exponent is already encoded in result buffer
; Encode negative exponent (0001B) into BCD12 number
; DI --> sign field
        MOV     AL,BYTE PTR [BX]        ; Get sign digit into AL
        MOV     AH,00000001B    ; Code for negative exponent
        OR      AL,AH           ; OR exponent sign into byte
        MOV     BYTE PTR [BX],AL        ; Store in buffer
; Calculate absolute value of negative exponent
        MOV     AX,10000        ; Bias to AX
        SUB     AX,DX           ; Bias minus exponent
        XCHG    AX,DX           ; Move exponent to DX
        JMP     CALCULATE_VALUE ; Go to conversion routine
;**********************|
;  calculate value of  |
;  positive exponent   |
;**********************|
; DX = exponent in bias 10000 form
POS_VALUE:
        MOV     AX,10000        ; Bias to AX
        SUB     DX,AX           ; Exponent minus bias
CALCULATE_VALUE:
; DX = absolute value of exponent (in binary)
        MOV     AX,DX           ; For conversion routine
        PUSH    BX              ; Save pointer for result
        CALL    FAR PTR BIN_TO_BCD       ; Conversion routine
        POP     BX              ; Restore pointer
; DH and DL holds 4 packed BCD digits
        INC     BX              ; Bump pointer to exponent filed
        MOV     BYTE PTR [BX],DH        ; Store two digits
        INC     BX
        MOV     BYTE PTR [BX],DL        ; Store two other digits
        RET
RESTORE_BCD     ENDP
;*******************************************************************
;
TRIAL_QUOT      PROC    NEAR
; Obtain largers possible trial quotient for division
; On entry:
;          nothing
; On exit:
;          AL = trial quotient
;
        LEA     BX,CS:DIVIDEND  ; Pointer to dividend buffer
        MOV     AL,CS:[BX]      ; First dividend digit to AL
        INC     BX              ; Pointer to next dividend digit
        MOV     CH,CS:[BX]      ; Second dividend digit to CH
        MOV     DL,CS:DIVISOR   ; High order divisor digit to DL
; Convert dividend digits (in AL-CH) to binary by multiplying the
; first digit by 10 and adding the second digit
        MOV     CL,10           ; Multiplier
        MUL     CL              ; First digit (in AL) times 10
        ADD     AL,CH           ; Add in second digit (in CH)
; AL now holds the binary dividend
; Trial quotient is obtained by dividing by the high order divisor
; digit (in DL)
        MOV     AH,0            ; Clear high nibble
        DIV     DL              ; Dividend / divisor = quotient
; The quotient must be a single digit. Test for a trial quotient
; larger than 9 and adjust
        CMP     AL,9            ; Test for limit value
        JA      ADJUST_DIGIT    ; Go is larger than 9
; At this point there is a valid trial quotient digit in AL
        RET
ADJUST_DIGIT:
        MOV     AL,9            ; Use largest single digit
        RET
TRIAL_QUOT      ENDP
;*******************************************************************
;
TRANSFER_CS_CX  PROC    NEAR
; Transfer CX byte from a code segment buffer (by SI) to another
; one (by DI)
; On entry:
;         CS:SI --> source buffer for move
;         CS:DI --> destination buffer for move
;         CX = number of bytes to move
; On exit:
;         CS:DI --> moved string
;         all registers preserved
;
        PUSH    SI              ; Save entry registers
        PUSH    DI
        PUSH    CX
MOVE_TO_CS:
        MOV     AL,CS:[SI]      ; Get byte from source buffer
        MOV     CS:[DI],AL      ; and place it in the destination
        INC     SI              ; Bump buffer pointers
        INC     DI
        LOOP    MOVE_TO_CS
        POP     CX              ; Restore caller's registers
        POP     DI
        POP     SI
        RET
TRANSFER_CS_CX  ENDP
;*******************************************************************
;
DIGIT_TO_QUOT   PROC    NEAR
; Place digit (in DL) in QUOTIENT buffer (if not full)
; On entry:
;         DL = quotient digit to store
; On exit:
;         carry set if QUOTIENT buffer full
;
        PUSH    BX              ; Save entry registers
        PUSH    AX
        LEA     BX,CS:QUOTIENT  ; Pointer to destination buffer
        ADD     BX,BP           ; BP register holds buffer offset
; Test for end of QUOTIENT buffer
        CMP     BYTE PTR CS:[BX],0FFH   ; Marker for end of buffer
        JE      QUOT_BUF_END    ; Go if at end
; Place digit in buffer and index pointer
        MOV     CS:[BX],DL      ; Store digit
        INC     BP              ; Bump offset counter
;**********************|
;    normal exit       |
;**********************|
        POP     AX              ; Restore registers
        POP     BX
        CLC                     ; Signal buffer not full
        RET
;**********************|
;  buffer full exit    |
;**********************|
QUOT_BUF_END:
        POP     AX              ; Restore entry registers
        POP     BX
        STC                     ; Signal buffer full to caller
        RET
DIGIT_TO_QUOT   ENDP
;*******************************************************************
;
COMPARE_DI_DR   PROC    NEAR
; Compare dividend to divisor
; On entry:
;          nothing
; On exit:
;          carry set if divisor > dividend
;          carry clear if divisor < dividend
;
        LEA     DI,CS:DIVIDEND  ; DI --> dividend
        LEA     SI,CS:DIVISOR   ; SI --> divisor
; If first digit of divided is not zero then it must be larger than
; the divisor
        CMP     BYTE PTR CS:[DI],00     ; Test for zero digit
        JNE     LARGER_DIVIDEND         ; Dividend is larger
        INC     DI              ; Bump pointer to next digit
COMP_DI_DR:
        MOV     AL,CS:[SI]      ; Divisor digit
        CMP     AL,0FFH         ; Test for end of divisor
        JE      LARGER_DIVIDEND ; Dividend is larger exit
        MOV     AH,CS:[DI]      ; Get dividend digit to AH
; Compare dividend and divisor digits
        CMP     AH,AL           ; Operation is dividend - divisor
        JB      LARGER_DIVISOR  ; Go if divisor > dividend
        JA      LARGER_DIVIDEND ; or if divisor < dividend
; At this point divisor = dividend. Continue comparing digits
        INC     SI              ; Bump pointers
        INC     DI
        JMP     COMP_DI_DR
;**********************|
;  dividend > divisor  |
;**********************|
LARGER_DIVIDEND:
        CLC                     ; Signal to caller
        RET
;**********************|
;  divisor > dividend  |
; or divisor = dividend|
;**********************|
LARGER_DIVISOR:
        STC
        RET
COMPARE_DI_DR   ENDP
;*******************************************************************
;
CLEAR_CS_BUF    PROC    NEAR
; Clear CX number of bytes in code segment buffer
; On entry:
;         CS:BX --> buffer area
;            CX = number of bytes to clear
CLEAR_BYCS:
        MOV     BYTE PTR CS:[BX],00H    ; Clear one byte
        INC     BX                      ; Bump pointer
        LOOP    CLEAR_BYCS
        RET
CLEAR_CS_BUF    ENDP
;*******************************************************************
;
STORE_AND_CLEAR         PROC    NEAR
; Store entry pointers in scratchpad area and clear work buffers
;**********************|
; store entry pointers |
;**********************|
        MOV     CS:ENTRY_SI,SI          ; Store SI pointer
        MOV     CS:ENTRY_DI,DI          ; Store DI pointer
        MOV     CS:ENTRY_BX,BX          ; Store BX pointer
;**********************|
;  clear work buffers  |
;**********************|
        PUSH    SI              ; Save x value entro pointer
        LEA     SI,CS:SIG_L     ; Pointer to buffer
        MOV     CX,109          ; Clear three buffers
CLEAR_109:
        MOV     BYTE PTR CS:[SI],00H    ; Clear one byte in buffer
        INC     SI                      ; Bump pointer
        LOOP    CLEAR_109
        POP     SI              ; Restore x pointer
        RET
STORE_AND_CLEAR         ENDP
;*******************************************************************
;
ROUND_SIGF      PROC    NEAR
; Rounding of significand digits stored in SIG_R buffer
; Rationale:
; SIG_R holds 37 digits in unpacked BCD format. The BCD12 format
; encodes 18 siginificand digits
; Rounding proceeds with the 19th digit in SIG_R, as follows:
; 1. If the 19th digit is < 5 then no rounding is required
; 2. If the 19th digit is > 5 then the first 18 digits are rounded
; 3. If the 19th digit = 5 then the digits to the right are examined.
;    If the digits to the right are all zeros (5000..00) then the
;    18th digit is incremented only if it is odd. This rounding
;    method is known as "round to nearest even".
; 4. If the result of incrementing the 18th digit is 10, the digit
;    to the left is also incremented, until no carry is generated.
;
        LEA     SI,CS:SIG_R     ; Set pointer to buffer in CS
        ADD     SI,18           ; Index to the 19th digit
        MOV     AL,CS:[SI]      ; 19th digit to AL
        CMP     AL,05H          ; Test for 5 or greater
        JB      EXIT_ROUNDING   ; No rounding necessary
; 19th digit is 5 or greater
        JE      MIDWAY_CASE     ; Go to "round to even" routine
; At this point the 19th digit is > 5
;**********************|
;  bump digit and test |
;**********************|
ROUND_OFF:
        DEC     SI              ; Point to 18th digit
ROUND_EVEN:
        MOV     AL,CS:[SI]      ; Get BCD digit into AL
        ADD     AL,01H          ; Add 1 to digit
        AAA                     ; Adjust result
        JNC     EXIT_ROUNDING   ; Go if no carry generated
        JMP     ROUND_OFF
; Note that no testing for the end of buffer is required since the
; leftmost digit in SIG_R cannot be greater than 8
;**********************|
; test for midway case |
;**********************|
; SI --> to the 19th digit
MIDWAY_CASE:
        PUSH    SI              ; Save pointer to 19th digit
        MOV     CX,17           ; Set counter for 17 digits
TEST_FOR_0H:
        INC     SI              ; Bump pointer to next digit
        CMP     BYTE PTR CS:[SI],00H    ; Test for a zero digit
        JNE     NOT_MIDWAY      ; Exit if not midway case
        LOOP    TEST_FOR_0H     ; Continue testing
; At this point the undisplayed digits are in the pattern 5000..00
        POP     SI              ; Restore pointer to 19th digit
        DEC     SI              ; Point to 18th digit
        MOV     AL,CS:[SI]      ; Get 18th digit into AL
;**********************|
;  test for odd 18th   |
;        digit         |
;**********************|
; Binary encodings of odd numbers have the LOB set
        TEST    AL,00000001B    ; Test LOB to determine if odd
        JNZ     ROUND_EVEN      ; Go to rounding routine if odd
        JMP     EXIT_ROUNDING   ; Do not round if already even
;**********************|
;    not midway case   |
;**********************|
; The number is rounded if the undiplayed significand is not in the
; form 50000..00
NOT_MIDWAY:
        POP     SI              ; Restore pointer
        JMP     ROUND_OFF       ; Go to rounding routine
;**********************|
;   exit procedure     |
;**********************|
EXIT_ROUNDING:
        RET
ROUND_SIGF      ENDP
;*******************************************************************
;
UNPACK_MOVE     PROC    NEAR
; Move the number of packed digits in CX from DS:SI to CS:DI 
; and unpack
; On entry:
;         CX = number of packed BCD digits to be moved
;         DS:SI --> packed BCD string to move
;         CS:DI --> destination of unpacked string
; On exit:
;         CS:DI = unpacked string of BCD digits
;
        PUSH    SI
        PUSH    DI
; Move CX digits
MOVE_DIGS:
        MOV     AL,[SI]         ; Get one byte from source
; AL has two packed BCD digits
        MOV     AH,AL           ; Copy packed digits to AH
        AND     AL,00001111B    ; Clear high nibble to isolate low
                                ; BCD digit
        PUSH    CX              ; Save digits counter
        MOV     CL,4            ; Counter for 4 bits
        SHR     AH,CL           ; Shift right to isolate high BCD
                                ; digit
        POP     CX              ; Restore counter
; AH holds the unpacked high digit and AL holds the low digit
        MOV     CS:[DI],AH      ; High BCD digit to buffer
        INC     DI              ; Bump buffer pointer
        MOV     CS:[DI],AL      ; Low BCD digit to buffer
        INC     DI              ; Bump pointer again
        INC     SI              ; Bump source pointer
        LOOP    MOVE_DIGS
        POP     DI
        POP     SI
        RET
UNPACK_MOVE     ENDP
;*******************************************************************
;
TEST_4_ZERO     PROC    NEAR
; Test a number encoded in BCD12 format for a zero value
; Note: a number in BCD12 format is zero if all numbers in the
;       exponent and significand are zero
; On entry:
;         SI --> Number stored in BCD12 format
; On exit:
;         Carry set if number is zero
;         Carry clear if number is not zero
;
        PUSH    SI              ; Save entry pointer
        INC     SI              ; Point to first exponent digit
        MOV     CX,11           ; Counter for 11 bytes to test
TEST_11_ZERO:
        CMP     BYTE PTR[SI],0  ; Is the byte zero
        JNE     NOT_11_ZERO     ; Go if not zero
        INC     SI              ; Bump pointer
        LOOP    TEST_11_ZERO    ; Test all digits
; At this point the number is zero
        POP     SI              ; Restore entry pointer
        STC                     ; Set zero return signal
        RET
NOT_11_ZERO:
        POP     SI              ; Restore entry pointer
        CLC                     ; Set non-zero return signal
        RET
TEST_4_ZERO     ENDP
;*******************************************************************
;
EXP_2_BIAS      PROC    NEAR
; Convert exponent of BCD12 number into binary and add bias
; On entry:
;         SI --> Number encoded in BCD12 format
; On exit:
;         DX = binary value of exponent
;
        PUSH    BX              ; Save pointers
        PUSH    SI
        PUSH    DI
        INC     SI              ; Pointer to exponent field
        INC     SI              ; Second exponent digit
; Move BCD digits to AL-BH-BL-CH-CL for conversion routine
        MOV     AL,0            ; Highest digit is zero
        MOV     DL,BYTE PTR [SI]        ; DL holds 2 packed digits
        MOV     DH,DL           ; Copy digits to DH
        AND     DL,00001111B    ; Erase high nibble in DL
        AND     DH,11110000B    ; and low in DH
        MOV     CX,4            ; For a count of 4
        SHR     DH,CL           ; Shift digit to low position
        DEC     SI              ; Bump pointer to previous digit
        MOV     BL,BYTE PTR [SI]        ; BL holds 2 packed digits
        MOV     BH,BL           ; Copy digits to BH
        AND     BL,00001111B    ; Erase high nibble in BL
        AND     BH,11110000B    ; and low in BH
        MOV     CX,4            ; For a count of 4
        SHR     BH,CL           ; Shift digit to low position
        MOV     CX,DX           ; Move digits to CX
        CALL    FAR PTR BCD_INTO_REG    ; BCD to binary conversion
        POP     DI              ; Restore entry pointers
        POP     SI
        POP     BX
; At this point DX hold binary value for exponent
; SI --> sign byte
; Test for negative exponent. Add bias if positive
        TEST    BYTE PTR[SI],01H        ; Test for negative exponent
        JNZ     EXP_IS_NEG
; At this point exponent is positive
        ADD     DX,10000        ; Add bias to exponent value
        RET                     ; End of processing
; Subtract exponent from bias
EXP_IS_NEG:
        MOV     AX,10000        ; Bias
        SUB     AX,DX           ; Value minus bias
        MOV     DX,AX           ; Result to DX
        RET                     ; End of processing
EXP_2_BIAS      ENDP
;*****************************************************************
;
;GET_TAG PROCEDURE
; Procedure to return tag code corresponding to the requested FPU
; register
; A 7-character string with the tag text is passed to the caller
; On entry:
;         DS:DI -> caller's text buffer with a minimum of 7 
;                  free bytes for message storage
;         AL = register stack desired (modulus 8)
;
; On exit:
;         caller's buffer (by DS:DI) contains tag string
;         BX = tag code for requested FPU register
;
; Code segment storage for FPU environment variables
ENVIRO_FPU      DW      0       ; FPU control word
STATUS_FPU      DW      0       ; FPU status word
TAG_WORD        DW      0       ; FPU tag word
; Storage for exception pointers
                DW      2 DUP(00H)      ; Instruction pointer
                DW      2 DUP(00H)      ; Data pointer
;
; Temporary storage for requested register number
STACK_REG       DB      0
                DW      0
;
; Error message table. Each entry is 7 bytes
TAG_MESS_TBL    DB      'Valid  '
                DB      'Zero   '
                DB      'Special'
                DB      'EMPTY  '
;
GET_TAG         PROC    FAR
; Read tag for stack top register
; On entry:
;         AL = Register number for which tag is requested
;              (range is 0 to 7)
; On exit:
;         BX = tag for register requested
;
; Save caller's requested stack register number
        MOV     CS:STACK_REG,AL         ; For later use
        PUSH    AX              ; Save caller's context
        PUSH    CX
        PUSH    DX
        PUSH    SI
        PUSH    DI
;**********************|
;   store environment  |
;**********************|
        FSTENV  CS:ENVIRO_FPU   ; Environment to memory
        FWAIT                   ; Wait for end of memory access
        MOV     BX,CS:TAG_WORD     ; Tag word into BX
        MOV     AX,CS:STATUS_FPU        ; Status word to AX
;**********************|
;   isolate TOS field  |
;**********************|
        AND     AH,00111000B    ; Mask off all other bits
        MOV     CL,3            ; Count of 3 bits to shift
        SHR     AH,CL           ; Shift TOS bits right 3 times
; AH now holds the number of the physical register presently
; mapped as the stack top (ST(0))
; The tag word is mapped as follows:
;   xx xx xx xx | xx xx xx xx 
;   -- -- -- --   -- -- -- --
;    |  |  |  |    |  |  |  |
;    7  6  5  4    3  2  1  0  <-- physical register
;
; To find the physical register that corresponds with the FPU
; stack register requested by the caller it is necessary to
; add the caller's register number to the current stack top
; and subtract 8 is the sum exceeds register 7. This calculates
; the modulus 8 address of the requested stack register
        MOV     AL,CS:STACK_REG ; Caller's requested stack
                                ; register number
        ADD     AL,AH           ; AL holds sum
        CMP     AL,7            ; Test for limit
        JNA     SUM_IN_MOD_8    ; Go if less
        SUB     AL,8            ; Convert to modulus 8
SUM_IN_MOD_8:
; Unit mask is 00000000 00000011B (0003H)
        MOV     DX,0003H        ; Mask to DX
        MOV     CL,AL           ; Register number to CL
        ADD     CL,CL           ; Double number to get shift count
        SHL     DX,CL           ; Shift mask bits left
        AND     BX,DX           ; Mask off all other tag bits
        SHR     BX,CL           ; Shift unmasked tag bits right
;***************************|
;  move message to caller's |
;     buffer and exit       |
;***************************|
; At this point BX holds the tag code
        MOV     AX,BX           ; Tag code to AX
; The value in AX is multiplied by 7 to obtain the offset
; of the corresponding tag code text message
; Message is then moved to the caller's buffer by DS:DI
        LEA     SI,CS:TAG_MESS_TBL      ; Offset of table
        MOV     CL,7            ; Length of each message
        MUL     CL              ; AX = offset of correct message
        ADD     SI,AX           ; Add to table offset
; At this point:
;       CS:SI --> 7-byte number type message
;       DS:DI --> caller's buffer with 7 bytes minimum space
        MOV     CX,7            ; Counter for 12 bytes
TRANSFER_7:
        MOV     AL,CS:[SI]      ; Get message character
        MOV     DS:[DI],AL      ; Place in caller's buffer
        INC     SI              ; Bump buffer pointers
        INC     DI
        LOOP    TRANSFER_7
; Restore caller's context
        POP     DI              ; Restore caller's environment
        POP     SI
        POP     DX
        POP     CX
        POP     AX
        RET
GET_TAG         ENDP

CODE    ENDS
        END


       