;*****************************************************************
;*****************************************************************
;                           UNIT_6.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date:  January 13, 1994
; Updated:          March 12, 1994
;
; Libary name: MATHUNIT.LIB
;
;*****************************************************************
; MATHEMATICAL ROUTINES IN THIS MODULE
PUBLIC  DEG_TO_RAD      ; .. degress to radian conversion
PUBLIC  RAD_TO_DEG      ; .. radians to degrees conversion
PUBLIC  CIRCLE_SCALE    ; .. scale angle to unit circle
PUBLIC  OCTANT_SCALE    ; .. scale angle to first octant
PUBLIC  TANGENT         ; .. calculate tangent function
PUBLIC  SINE            ; .. idem for sine
PUBLIC  COSINE          ; .. idem for cosine
PUBLIC  COTANGENT       ; .. idem for cotangent
PUBLIC  SECANT          ; .. idem for secant
PUBLIC  COSECANT        ; .. idem for cosecant
PUBLIC  ARC_SINE        ; .. calculate arcsine function
PUBLIC  ARC_COSINE      ; .. idem for arccosine
PUBLIC  ARC_TANGENT     ; .. idem for arctangent
PUBLIC  TANGENT_387     ; .. tangent in 387 or later FPU
PUBLIC  SINE_387        ; .. idem for sine
PUBLIC  COSINE_387      ; .. idem for cosine
PUBLIC  SINH            ; .. hyperbolic sine
PUBLIC  COSH            ; .. hyperbolic cosine
PUBLIC  TANH            ; .. hyperbolic tangent
PUBLIC  ARC_SINH        ; .. hyperbolic arcsine
PUBLIC  ARC_COSH        ; .. hyperbolic arcosin
PUBLIC  ARC_TANH        ; .. hyperbolic arctangent
;
;*****************************************************************
;                           externals
;*****************************************************************
; In other modules of the MATHUNIT library
EXTRN   LOG_E:          FAR
EXTRN   X_TO_Y_BYLOG:   FAR
;
;****************************************************************
;                          code segment
;****************************************************************
;
CODE    SEGMENT BYTE
        ASSUME  CS:CODE
;
;*****************************************************************
;             code segment data for trigonometry
;*****************************************************************
SIGN            DB      0       ; Storage for sign of input
STATUS_WW       DW      0       ; Storage for status word
REMAINDER       DB      1       ; Storage for zero remainder
                                ; of r MOD(pi/4)
;*****************************************************************
;                       80387 trigonometry
;*****************************************************************
;****************************************************************
;                      degrees-to-radian
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | YES |
;                       | 80286/80287     | YES |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;****************************|
;    code segment constants  |
;****************************|
; Constants are defined in extended precision format
ONE_80          DT      180.0   ; 180
;
DEG_TO_RAD      PROC    FAR
; Convert entry value in ST(0) to radian measure
;
; On entry:
;         ST(0) = value in degrees
; On exit:
;         ST(0) = value converted to radians
;
; Formula:
;         r = d * (pi/180)
;
; where r = angle in radian, d = angle in degrees
;

;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      d      |  --------  |  -------- |
        FLD     CS:ONE_80
                        ;     180     |     d      |
        FLDPI           ;    pi       |    180     |     d     |
        FDIV    ST,ST(1);  pi/180     |    180     |     d     |
        FSTP    ST(1)   ;  pi/180     |     d      |
        FMUL    ST,ST(1);     r       |     d      |
        FSTP    ST(1)   ;     r       |  --------- |
; ST(0) now holds entry value in radian
        RET
DEG_TO_RAD      ENDP


;****************************************************************
;                      radians-to-degrees
;****************************************************************
RAD_TO_DEG      PROC    FAR
; Convert entry value in ST(0) to degrees
;
; On entry:
;         ST(0) = value in radian
; On exit:
;         ST(0) = value converted to degrees
;
; Formula:
;         d = r * (180/pi)
;
; where r = angle in radian, d = angle in degrees
;
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      d      | ---------- | --------- |
        FLDPI           ;      r      |    pi      | --------- |
        FLD     CS:ONE_80
                        ;     180     |     pi     |     r     |
        FDIV    ST,ST(1);  180/pi     |     pi     |     r     |
        FSTP    ST(1)   ;  180/pi     |     r      | --------- |
        FMUL    ST,ST(1);     d       |     r      | --------- |
        FSTP    ST(1)   ;     d       |  --------  |
; ST(0) now holds entry value in degrees
        RET
RAD_TO_DEG      ENDP


CIRCLE_SCALE      PROC        FAR
; Reduce angle in ST(0) (r) to a value less than 2pi radian
; by calculating  n = r (MOD 2pi)
; The code assumes that a word-size variable, named STATUS_WW
; has been previously defined in the code segment.
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      |  --------- | --------- |
        FLDPI           ;     pi      |     r      | --------- |
        FADD    ST,ST(0);     2pi     |     r      | --------- |
        FXCH            ;      r      |     2pi    | --------- |
UNIT_CIRC:
        FPREM           ; REM(r/2pi)  |     2pi    | --------- |
        FSTSW   CS:STATUS_WW    ; Store status bits in memory
        FWAIT                   ; Delay until end of store
        MOV     AH,BYTE PTR CS:STATUS_WW+1
                        ; Move status bits into AH
        AND     AH,00000100B    ; Mask off all except C2
        JNZ     UNIT_CIRC       ; Repeat if C2 not clear
;                       ;    ST(0)    |    ST(1)   |    ST(2)
        FSTP    ST(1)   ;      n      | ---------- | --------- |
; ST(0) now holds an angle in the unit circle (n)
      RET
CIRCLE_SCALE    ENDP


OCTANT_SCALE    PROC    FAR
; Scale an input angle (in radian measure) to the first octant
; and report the octant of the original angle
; On entry:
;         ST(0) = positive angle in the unit circle
; On exit:
;         ST(0) = input scaled to pi/4 radian (first octant)
;         ST(1) = pi/4
;         AH = condition codes C3, C1, and C0 in their original
;              bit positions
;
; The code assumes that a word-size variable named STATUS_WW
; has been previously defined in the code segment.
;
;***************************|
;   reduce to unit circle   |
;***************************|
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      | ---------- | --------- |
        FLDPI           ;     pi      |     r      | --------- |
        FADD    ST,ST(0);     2pi     |     r      | --------- |
        FXCH            ;      r      |     2pi    | --------- |
SCALE_TO_UNIT:
        FPREM           ; r = MOD(2pi)|     2pi    | --------- |
        FSTSW   CS:STATUS_WW    ; Store status bits in memory
        FWAIT                   ; Delay until end of store
        MOV     AH,BYTE PTR CS:STATUS_WW+1
                        ; Move status bits into AH
        AND     AH,00000100B    ; Mask off all except C2
        JNZ     SCALE_TO_UNIT   ; Repeat if C2 not clear
;                       ;    ST(0)    |    ST(1)   |    ST(2)
        FSTP    ST(1)   ;  r (scaled  | ---------- | --------- |
;                       |   to 2pi)   |            |           |
; ST(0) now holds an angle in the unit circle
;***************************|
;      reduce to pi/4       |
;***************************|
        FCLEX
        FLD1                    ;     1    |     r    | ------- |
        FADD    ST(0),ST        ;     2    |     r    | ------- |
        FCHS                    ;    -2    |     r    | ------- |
        FLDPI                   ;    pi    |    -2    |     r   |
        FSCALE                  ;   pi/4   |    -2    |     r   |
        FSTP    ST(1)           ;   pi/4   |     r    | ------- |
        FXCH                    ;     r    |    pi/4  | ------- |
        FPREM                   ; MOD(pi/4)|   pi/4   | ------- |
        FSTSW   CS:STATUS_WW    ; Code segment storage for
                                ; status word
; At this point ST(0) contains an angle in the first octant
; The FPREM instruction will produce a valid remainder in one
; execution if the input is < 2pi radian
        MOV     AH,BYTE PTR CS:STATUS_WW+1
                        ; Move status bits into AH
        AND     AH,01000011B    ; Mask off unused bits
;                   |    ||_______ C0
;                   |    |________ C1
;                   |_____________ C3
; FPU stack on exit
;                               |r (scaled |   pi/4   | ------- |
;                               | to pi/4) |          |
;
        RET
OCTANT_SCALE    ENDP


;***************************************************************
;        8087/80287 procedure for tangent function
;***************************************************************
TANGENT         PROC    FAR
; Compute tangent of an angle in radians
; (no restrictions on range or sign of argument)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = tangent function
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | YES |
;                       | 80286/80287     | YES |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
;***************************|
;      test for negative    |
;        argument           |
;***************************|
        MOV     CS:SIGN,0       ; Assume positive
        FTST                    ; Test ST(0)
        FSTSW   CS:STATUS_WW    ; Store status word
        FWAIT
        MOV     AH,BYTE PTR CS:STATUS_WW + 1
        TEST    AH,00000001B    ; Is bit C0 set?
        JZ      POSITIVE_ANGLE  ; Go if positive
; At this point input angle is negative
        MOV     CS:SIGN,1       ; Store sign of input angle
        FABS                    ; Change sign
;***************************|
;   scale to first octant   |
;***************************|
; At this point:
;                               |    r    | -------- | -------- |
POSITIVE_ANGLE:
        CALL    FAR PTR OCTANT_SCALE    ; Scaling procedure
;  ST(0) = input scaled to pi/4 radian (first octant)
;  ST(1) = pi/4
;     AH = Condition code bits C3 to C0
;                               |    r    |   pi/4   | -------- |
;***************************|
;  test for r REM(pi/4) = 0 |
;***************************|
; If r MOD(pi/4) = 0 (no remainder) then angle lies exactly at
; one of the octant boundaries
        FTST                    ; Test ST(0)
        FSTSW   CS:STATUS_WW    ; Store status word
        FWAIT
        MOV     CS:REMAINDER,1  ; Initialize remainder control
                                ; Assume remainder not 0
        AND     BYTE PTR CS:STATUS_WW+1,01000001B
        CMP     BYTE PTR CS:STATUS_WW+1,01000000B
        JNE     ANGLE_OK        ; Go if remainder not zero
        MOV     CS:REMAINDER,0  ; Set remainder control
                                ; to no remainder (r = 0)
; Tangent function is calculated according to the octant position
; of the original angle, as follows:
; OCTANT       C3    C1      FORMULA         IF r = 0      CASE
; 0 or 4       0     0    FPTAN(r)               0           A
; 1 or 5       0     1    1/FPTAN(PI/4 - r)      1           B
; 2 or 6       1     0    -1/FPTAN(r)        infinity        C
; 3 or 7       1     1    -FPTAN(PI/4 - r)      -1           D
;
; First determine if bit C1 is set
ANGLE_OK:
        TEST    AH,00000010B    ; Test bit C1
        JZ      CASE_A_OR_C     ; Go if C1 not set
        CMP     CS:REMAINDER,0  ; Test for remainder = 0
        JE      TAN_IS_1        ; If remainder = 0, tangent = 1
;***************************|
; calculate FPTAN(pi/4 - r) |
;***************************|
;                               |    r    |   pi/4   | -------- |
        FSUBP   ST(1),ST        ;    r    | pi/4 - r | -------- |
        FPTAN                   ;  tan r  | pi/r - r | -------- |
        JMP     DO_CASES
;***************************|
;    if r REM(pi/4) = 0     |
;       tangent = 1         |
;***************************|
TAN_IS_1:
        FSTP    ST              ;  pi/4   | -------- | -------- |
        FSTP    ST              ; ------- | -------- |
        FLD1                    ;   1     | -------- |
        FLD1                    ;   1     |     1    | ---------|
; These values will assure that tangent = 1
        JMP     DO_CASES
;***************************|
;   calculate FPTAN(r)      |
;***************************|
CASE_A_OR_C:
        FSTP    ST(1)           ;   r     | -------- |
        CMP     CS:REMAINDER,0  ; Test for remainder = 0
        JE      TAN_IS_0        ; Go if remainder = 0
        FPTAN                   ; Function in cases A and C
        JMP     DO_CASES
;***************************|
;    if r REM(pi/4) = 0     |
;       tangent = 0         |
;***************************|
TAN_IS_0:
        FSTP    ST              ; -------- | -------- |
        FLDZ                    ;    0     | -------- |
        FLD1                    ;    1     |     0    | --------|
;******************************|
;  execute case logic:         |
;  change sign in cases C or D |
;  reciprocal in case B or C   |
;******************************|
DO_CASES:
; Test bit C3, if set, change sign
        TEST    AH,01000000B    ; Is bit C3 set?
        JZ      NOT_C_OR_D      ; C3 is not set
; Processing for cases C and D
        FCHS                    ; Change sign
NOT_C_OR_D:
        AND     AH,01000010B    ; Eliminate all except C3 and C1
        CMP     AH,01000010B    ; Test for case D
        JE      CASE_A_OR_D     ; Go if both bits are clear
        CMP     AH,0            ; Test for case A
        JE      CASE_A_OR_D
;***************************|
;  processing cases B, C    |
;***************************|
        FXCH                    ; Exchange ST(0) and ST(1)
;***************************|
;   calculate tangent r     |
;***************************|
CASE_A_OR_D:
        FDIVP   ST(1),ST
;***************************|
;   tan(-r) = -tan(r)       |
;***************************|
; Test for sign change in input angle
        CMP     CS:SIGN,0
        JE      LEAVE_POS
        FCHS                    ; Change sign of tangent
LEAVE_POS:
        RET
;
TANGENT         ENDP


;***************************************************************
;           8087/80287 procedures for sine function
;***************************************************************
; See code segment variables in the TANGENT procedure
;
SINE            PROC    FAR
; Compute sine of an angle in radians
; (no restrictions on range or sign of argument)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = sine function
;
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | YES |
;                       | 80286/80287     | YES |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;***************************|
;      test for negative    |
;        argument           |
;***************************|
        MOV     CS:SIGN,0       ; Assume positive
        FTST                    ; Test ST(0)
        FSTSW   CS:STATUS_WW    ; Store status word
        FWAIT
        MOV     AH,BYTE PTR CS:STATUS_WW + 1
        TEST    AH,00000001B    ; Is bit C0 set?
        JZ      POSITIVE_SINE   ; Go if positive
; At this point input angle is negative
        MOV     CS:SIGN,1       ; Store sign of input angle
        FABS                    ; Change sign
;***************************|
;   scale to first octant   |
;***************************|
; At this point:
;                               |    r    | -------- | -------- |
POSITIVE_SINE:
;***************************|
; entry point for cosine    |
;***************************|
        CALL    FAR PTR OCTANT_SCALE    ; Scaling procedure
;  ST(0) = input scaled to pi/4 radian (first octant)
;  ST(1) = pi/4
;     AH = Condition code bits C3 to C0
;                               |    r    |   pi/4   | -------- |
;***************************|
;  test for r REM(pi/4) = 0 |
;***************************|
; If r MOD(pi/4) = 0 (no remainder) then angle lies exactly at
; one of the octant boundaries
;***************************|
;  entry point for cosine   |
;         calculation       |
;***************************|
COSINE_ENTRY:
        FTST                    ; Test ST(0)
        FSTSW   CS:STATUS_WW    ; Store status word
        FWAIT
        MOV     CS:REMAINDER,1  ; Initialize remainder control
                                ; Assume remainder not 0
        AND     BYTE PTR CS:STATUS_WW+1,01000001B
        CMP     BYTE PTR CS:STATUS_WW+1,01000001B
        JNE     SINE_OK         ; Go if remainder not zero
        MOV     CS:REMAINDER,0  ; Set remainder control
                                ; to no remainder (r = 0)
; Sine function is calculated according to the octant position
; of the original angle, as follows:
; OCTANT   C3   C1   C0      FORMULA         IF r = 0      CASE
;   0      0    0    0      SIN(r)               0           A
;   1      0    1    0      COS(PI/4 - r)    SQRT(2)/2       B
;   2      1    0    0      COS(r)               1           C
;   3      1    1    0      SIN(PI/4 - r)    SQRT(2)/2       D
;
;   4      0    0    1     -SIN(r)              -0
;   5      0    1    1     -COS(PI/4 - r)   -SQRT(2)/2
;   6      1    0    1     -COS(r)              -1
;   7      1    1    1     -SIN(PI/4 - r)   -SQRT(2)/2
; Notice that octants 4 to 7 are the same as octants 0 to 3, but
; negative.
SINE_OK:
        TEST    AH,00000010B    ; Test bit C1
        JZ      SINE_A_OR_C     ; Go if C1 not set
        CMP     CS:REMAINDER,0  ; Test for remainder = 0
        JE      SINE_IS_1       ; If remainder = 0, tangent = 1
;***************************|
; calculate FPTAN(pi/4 - r) |
;***************************|
;                               |    r    |   pi/4   | -------- |
        FSUBP   ST(1),ST        ;    r    | pi/4 - r | -------- |
        FPTAN                   ;  tan r  | pi/r - r | -------- |
        JMP     DO_SINES
;***************************|
;    if r REM(pi/4) = 0     |
;       tangent = 1         |
;***************************|
SINE_IS_1:
        FSTP    ST              ;  pi/4   | -------- | -------- |
        FSTP    ST              ; ------- | -------- |
        FLD1                    ;   1     | -------- |
        FLD1                    ;   1     |     1    | ---------|
; These values will assure that sine = 1
        JMP     DO_SINES
;***************************|
;   calculate FPTAN(r)      |
;***************************|
SINE_A_OR_C:
        FSTP    ST(1)           ;   r     | -------- |
        CMP     CS:REMAINDER,0  ; Test for remainder = 0
        JE      SINE_IS_0       ; Go if remainder = 0
        FPTAN                   ;   1     |  tan r   | ---------|
        JMP     DO_SINES
;***************************|
;    if r REM(pi/4) = 0     |
;          sine = 0         |
;***************************|
SINE_IS_0:
        FSTP    ST              ; ------- | -------- |
        FLDZ                    ;   0     | -------- |
        FLD1                    ;   1     |     0    | ---------|
;******************************|
;  execute case logic:         |
;  if case B or C get cosine   |
;  if case A or D get sine     |
;******************************|
DO_SINES:
        MOV     AL,AH           ; Save condition code bits in AL
        AND     AH,01000010B    ; Eliminate all except C3 and C1
        CMP     AH,01000010B    ; Test for case D
        JE      SINE_A_OR_D     ; Go if both bits are clear
        CMP     AH,0            ; Test for case A
        JE      SINE_A_OR_D
;***************************|
;  processing cases A, D    |
;***************************|
        FXCH                    ; Exchange ST(0) and ST(1)
;***************************|
;   calculate sine of r     |
;***************************|
; At this point:                |  ST(0)   |   ST(1)  |  ST(2)  |
;                               |    x     |    y     | ------- |
SINE_A_OR_D:
        FMUL    ST(0),ST(0)     ;   x2     |    y     | ------- |
        FLD     ST(1)           ;   y      |   x2     |    y    |
        FMUL    ST(0),ST(0)     ;   y2     |   x2     |    y    |
        FADDP   ST(1),ST(0)     ; x2+y2    |    y     | ------- |
        FSQRT                   ;SQR(x2+y2)|    y     | ------- |
        FDIVP   ST(1),ST        ;   y /    | -------- |
;                               |SQR(x2+y2)|
;***************************|
; test for octants 4 to 7   |
;      and change sign      |
;***************************|
; Bit C0 is set if octant is in the range 4 to 7
        TEST    AL,00000001B    ; Test bit C0
        JZ      NOT_HIGH_OCTANT ; Go if not set
        NOT     CS:SIGN         ; Reverse sign control
;***************************|
;   sin(-r) = -sin(r)       |
;***************************|
; Test for sign change in input angle
NOT_HIGH_OCTANT:
        CMP     CS:SIGN,0
        JE      POS_SINE
        FCHS                    ; Change sign of tangent
POS_SINE:
        RET
;
SINE        ENDP


;***************************************************************
;            8087/80287 procedure for cosine function
;***************************************************************
; See code segment variables in the TANGENT procedure
;
COSINE          PROC    FAR
; Compute cosine of an angle in radians
; (no restrictions on range or sign)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = cosine function
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | YES |
;                       | 80286/80287     | YES |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
;***************************|
;  preliminary operations   |
;  for cosine calculation   |
;***************************|
        FABS                    ; Make argument positive
        MOV     CS:SIGN,0       ; Assume positive
;***************************|
;  adjust octant for cosine |
;***************************|
        CALL    FAR PTR OCTANT_SCALE    ; Scale argument to first
                                ; octant
; Cosine calculations are identical as sine calculations for an
; angle rotated by 90 degrees (PI/2 radian). Therefore adjustment
; for cosine function is made by adding 2 to the octant
; At this point function request is cosine
        XOR     AH,01000000B    ; Reverse bit C3
        TEST    AH,01000000B    ; Is it set?
        JNZ     SKIP_C0         ; Go if no carry
        XOR     AH,00000001B    ; Reverse bit C0
SKIP_C0:
        JMP     COSINE_ENTRY    ; Cosine entry point in SINE
                       ; SINE procedure
; Execution returns via SINE procedure
;
COSINE          ENDP


;***************************************************************
;               procedure for cotangent function
;***************************************************************
; See code segment variables in the TANGENT procedure
;
COTANGENT       PROC    FAR
; Compute cotangent of an angle in radians
; (no restrictions on range or sign)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = cotangent function
;
; Processing is based on the identity
;                 cot (r) = 1 / tan(r)
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | YES |
;                       | 80286/80287     | YES |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
        CALL    FAR PTR TANGENT ; Get tangent function
        FLD1                    ;    1    |   tan r  | -------- |
        FDIV    ST,ST(1)        ;  1/tan  |   tan r  | -------- |
        FSTP    ST(1)           ;  cot    | -------- |
        RET
;
COTANGENT       ENDP

SECANT          PROC    FAR
; Compute secant of an angle in radians
; (no restrictions on range or sign)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = secant function
;
; Processing is based on the identity
;                 sec (r) = 1 / sin(r)
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | YES |
;                       | 80286/80287     | YES |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
        CALL    FAR PTR SINE    ; Get function
        FLD1                    ;    1    |   sin r  | -------- |
        FDIV    ST,ST(1)        ;  1/sin  |   sin r  | -------- |
        FSTP    ST(1)           ;  sec    | -------- |
        RET
;
SECANT          ENDP

COSECANT        PROC    FAR
; Compute secant of an angle in radians
; (no restrictions on range or sign)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = secant function
;
; Processing is based on the identity
;                cosc (r) = 1 / cos(r)
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | YES |
;                       | 80286/80287     | YES |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
        CALL    FAR PTR COSINE  ; Get function
        FLD1                    ;    1    |   cos r  | -------- |
        FDIV    ST,ST(1)        ;  1/cos  |   cos r  | -------- |
        FSTP    ST(1)           ;  cosc   | -------- |
        RET
;
COSECANT        ENDP
;
;****************************************************************
;     8087/80287 procedures for trigonometric arc functions
;****************************************************************
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | YES |
;                       | 80286/80287     | YES |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;*********************************|
;           arc sine              |
;*********************************|
ARC_SINE        PROC    FAR
; Find the angle (in radian) whose sine is in ST(0)
; On entry:
;         ST(0) = sine of an angle (in radian)
;                 Must be a normal number in the range 0 < ST < 1
;                 this corresponds to an angle in the first
;                 quadrant
; On exit:
;         ST(0) = arc sine of entry value in ST(0)
;                 expressed in radian
; Formula:
;         arcsin (r) = arctangent r / square root (1 - x^2)
;
; Legend: x2 = x**2
;         y2 =  y**2
;         SQR = square root
;
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      | ---------- | --------- |
        FLD     ST(0)   ;      r      |     r      | --------- |
        FMUL    ST,ST(1);     r2      |     r      | --------- |
        FLD1            ;      1      |    r2      |     r     |
        FSUB    ST,ST(1);    1-r2     |    r2      |     r     |
        FSQRT           ; SQR(1-r2)   |    r2      |     r     |
        FSTP    ST(1)   ; SQR(1-r2)   |     r      | --------- |
        FPATAN          ; arcsin (r)  | ---------- |
        RET
ARC_SINE        ENDP
;***************************************************************
;
;*********************************|
;           arc cosine            |
;*********************************|
ARC_COSINE      PROC    FAR
; Find the angle (in radian) whose cosine is in ST(0)
; On entry:
;         ST(0) = cosine of an angle (in radian)
;                 Must be a normal number in the range 0 < ST < 1
;                 this corresponds to an angle in the first
;                 quadrant (0 < angle < pi/2)
; On exit:
;         ST(0) = arc cosine of entry value in ST(0)
; Formula:
;         arccos (r) = arctangent  square root (1 - x^2) / r
;
; Legend: x2 = x**2
;         y2 =  y**2
;         SQR = square root
;
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      | ---------- | --------- |
        FLD     ST(0)   ;      r      |     r      | --------- |
        FMUL    ST,ST(1);     r2      |     r      | --------- |
        FLD1            ;      1      |    r2      |     r     |
        FSUB    ST,ST(1);    1-r2     |    r2      |     r     |
        FSQRT           ; SQR(1-r2)   |    r2      |     r     |
        FSTP    ST(1)   ; SQR(1-r2)   |     r      | --------- |
        FXCH            ;     r       | SQR(1-r2)  | --------- |
        FPATAN          ; arcsin (r)  | ---------- |
        RET
ARC_COSINE      ENDP
;****************************************************************
;
;*********************************|
;          arc tangent            |
;*********************************|
ARC_TANGENT     PROC    FAR
; Find the angle (in radian) whose tangent is in ST(0)
; On entry:
;         ST(0) = tangent of an angle (in radian)
;                 Must be a normal number
; On exit:
;         ST(0) = arc tangent of entry value in ST(0)
;
; Entry values and identities:
; 1. If ST(0) is positive normal proceed as follows:
;    1a. If ST(0) = 1 then arctan (1) = pi/4
;    1b. If ST(0) > 1 then ATAN (x) = pi/2 - ATAN (1/x)
;    1c. If ST(0) < 1 then in range for FPATAN instruction
; 2. If ST(0) is negative then arctan(-x) = - arctan(x)
; 3. If ST(0) = 0 then arctan (0) = 0
;
; The BH register is used to store the sign of the original
; value. If BH = 0 original input was positive. If BH = 1
; original input was negative
        MOV     BX,0            ; Clear BX. Default sign is +
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      |   empty    |   empty
        FXAM            ; Examine stack top register
        FSTSW   CS:STATUS_WW
        FWAIT
; Relevant condition codes:
; C3     C2    C1    C0
;  0      1     0     0   if  ST(0) = positive normal
;  0      1     1     0   if  ST(0) = negative normal
;  1      0     ?     0   if  ST(0) = zero
;
        MOV     AH,BYTE PTR CS:STATUS_WW+1
        AND     AH,01000111B    ; Mask off all unused bits
        CMP     AH,00000110B    ; Bit code for negative
        JNE     NON_NEGATIVE    ; Argument is not negative
;**********************|
;   negative argument  |
;**********************|
; Change sign of ST(0) and store original negative sign
        MOV     BH,1            ; Code for negative input
        FCHS            ; Change sign of ST(0)
        JMP     CASE_1  ; Continue as if input was positive
NON_NEGATIVE:
        AND     AH,01000011B    ; Mask off C3 bit
        CMP     AH,01000000B    ; Test for zero
        JNE     CASE_1          ; Argument is not zero
;**********************|
;     zero argument    |
;**********************|
; The arc tangent of zero is zero
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      | ---------- |
        FSTP    ST(0)   ; ----------- |
        FLDZ            ;     0.0     | ---------- |
        RET
;**********************|
;   positive normal    |
;      argument        |
;**********************|
CASE_1:
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;      r      | ---------- |
        FLD1            ;     1.0     |      r     | ---------- |
        FCOM            ; Compare angle with 1 to determine if
                        ; r < 1, r > 1, or r = 1
        FSTSW   CS:STATUS_WW
        FWAIT
; Condition codes after FCOM:
; C3     C2     C0
;  0      0      0   if  1 > r
;  0      0      1   if  1 < r
;  1      0      0   if  1 = r
;
        MOV     AH,BYTE PTR CS:STATUS_WW+1
        AND     AH,01000011B    ; Mask off all unused bits
        JZ      R_LESS_1        ; Go if r < 1
        CMP     AH,00000001B    ; Test for r > 1
        JE      R_GREATER_1     ; Go if r > 1
;**********************|
; case 1A              |
;           ST(0) = 1  |
;**********************|
; If ST(0) = 1 the arc tangent ST(0) = pi/4
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;     1.0     |      r     | --------- |
        FADD    ST(0),ST;     2.0     |      r     | --------- |
        FADD    ST(0),ST;     4.0     |      r     | --------- |
        FLDPI           ;     pi      |     4.0    |     r     |
        FDIV    ST,ST(1);    pi/4     |     4.0    |     r     |
        FSTP    ST(1)   ;    pi/4     |      r     | --------- |
        JMP     SIGN_RTN
;**********************|
; case 1B              |
;           ST(0) > 1  |
;**********************|
R_GREATER_1:
;                       ;   ST(0)    |    ST(1)   |    ST(2)
                        ;    1.0     |      r     | --------- |
        FXCH            ;     r      |     1.0    | --------- |
        FPATAN          ; arctan(r)  |     1.0    | --------- |
        FLD1            ;     1      |  arctan(r) |    1.0    |
        FADD    ST,ST(0);     2      |  arctan(r) |    1.0    |
        FLDPI           ;    pi      |     2.0    | arctan(r) |..
        FDIV    ST,ST(1);   pi/2     |     2.0    | arctan(r) |..
        FSTP    ST(1)   ;   pi/2     | arctan(r)  |    1.0    |
        FSUBRP  ST(1),ST;  angle     |     1.0    |   empty
        JMP     SIGN_RTN
;**********************|
; case 1C              |
;           ST(0) < 1  |
;**********************|
R_LESS_1:
;                       ;    ST(0)    |    ST(1)   |    ST(2)
                        ;     1.0     |      r     | --------- |
        FPATAN          ;     angle   |      r     | --------- |
;**********************|
;     sign routine     |
;**********************|
; Change sign of angle if original tangent was negative
SIGN_RTN:
        CMP     BH,1            ; BH = 1 if negative input
        JNE     EXIT_ARCTAN     ; Go if not negative
        FCHS
EXIT_ARCTAN:
        RET
ARC_TANGENT     ENDP

.386
TANGENT_387     PROC    FAR
; Compute tangent of an angle in radians in the 80387 or the math
; unit of the 486 and Pentium. (no restrictions on range or sign)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = tangent function
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | NO  |
;                       | 80286/80287     | NO  |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
        CALL    FAR PTR CIRCLE_SCALE
        FPTAN                   ;    1    | tan r    | -------- |
        FSTP    ST(0)           ; tan r   | -------- |
        RET
TANGENT_387     ENDP


SINE_387        PROC    FAR
; Compute sine of an angle in radians in the 80387 or the math
; unit of the 486 and Pentium. (no restrictions on range or sign)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = sine function
;
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | NO  |
;                       | 80286/80287     | NO  |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
        CALL    FAR PTR CIRCLE_SCALE
        FSIN                    ;  sin r  | -------- | -------- |
        RET
SINE_387        ENDP

COSINE_387      PROC    FAR
; Compute sine of an angle in radians in the 80387 or the math
; unit of the 486 and Pentium. (no restrictions on range or sign)
;
; On entry:
;       ST(0) = an angle in radian (can be positive or negative)
; On exit:
;       ST(0) = sine function
;
;****************************************************************
; CPU compatibility chart:
;                       | 8086/8088/8087  | NO  |
;                       | 80286/80287     | NO  |
;                       | 80386/80387     | YES |
;                       | 486 DX          | YES |
;                       | PENTIUM         | YES|
;****************************************************************
;
        CALL    FAR PTR CIRCLE_SCALE
        FCOS                    ;  cos r  | -------- | -------- |
        RET
COSINE_387      ENDP
.8086
;
;*****************************************************************
;               procedures for hyperbolic functions
;*****************************************************************
;***************************|
;      code segment data    |
;***************************|
; Definition of constant e
E               DT      4000ADF85458A2BB4A9BH
;
E_TO_Y          PROC    NEAR
; Procedure for obtaining e**y in support of hyperbolic function
; calculations
;
; On entry:
;          ST(0) = y
; On exit:
;          ST(0) = e**y
;
        FLD     CS:E            ;     e     |    y     | -------- |
        FXCH                    ;     y     |    e     | -------- |
        CALL    X_TO_Y_BYLOG    ;   e**y    | -------- |
        RET
E_TO_Y          ENDP
;
SINH    PROC    FAR
; Calculate hyperbolic sine using formula
;        sinh(x) = (e**x - e**-x)/2
; On entry:
;         ST(0) = x
; On exit:
;         ST(0) = sinh(x)
;
;                               ;    x     | -------- |
; First obtain e**-x
        FLD     ST(0)           ;    x     |     x    | -------- |
        FCHS                    ;   -x     |     x    | -------- |
        CALL    E_TO_Y          ;  e**-x   |     x    | -------- |
        FXCH                    ;    x     |   e**-x  | -------- |
        CALL    E_TO_Y          ;   e**x   |   e**-x  | -------- |
        FSUB    ST,ST(1)        ;  (e**x - |
                                ;   e**-x) |   e**-x  | -------- |
        FSTP    ST(1)           ;(e**x ..) | -------- | -------- |
        FLD1                    ;    1     |(e**x ..) | -------- |
        FADD    ST,ST(0)        ;    2     |(e**x ..) | -------- |
        FDIV    ST(1),ST        ;    2     | sinh (x) | -------- |
        FSTP    ST(0)           ; sinh(x)  | -------- |
        RET
SINH    ENDP
;
COSH    PROC    FAR
; Calculate hyperbolic cosine using formula
;        cosh(x) = (e**x + e**-x)/2
; On entry:
;         ST(0) = x
; On exit:
;         ST(0) = cosh(x)
;
;                               ;    x     | -------- |
; First obtain e**-x
        FLD     ST(0)           ;    x     |     x    | -------- |
        FCHS                    ;   -x     |     x    | -------- |
        CALL    E_TO_Y          ;  e**-x   |     x    | -------- |
        FXCH                    ;    x     |   e**-x  | -------- |
        CALL    E_TO_Y          ;   e**x   |   e**-x  | -------- |
        FADD    ST,ST(1)        ;  (e**x - |
                                ;   e**-x) |   e**-x  | -------- |
        FSTP    ST(1)           ;(e**x ..) | -------- | -------- |
        FLD1                    ;    1     |(e**x ..) | -------- |
        FADD    ST,ST(0)        ;    2     |(e**x ..) | -------- |
        FDIV    ST(1),ST        ;    2     | sinh (x) | -------- |
        FSTP    ST(0)           ; sinh(x)  | -------- |
        RET
COSH    ENDP
;
TANH    PROC    FAR
; Calculate hyperbolic tangent using formula
;        tanh(x) = sinh(x) / cosh(x)
; On entry:
;         ST(0) = x
; On exit:
;         ST(0) = tanh(x)
;
;                               ;    x     | -------- |
        FLD     ST(0)           ;    x     |     x    | -------- |
        CALL    FAR PTR SINH    ;  sinh(x) |     x    | -------- |
        FXCH                    ;    x     |  sinh(x) | -------- |
        CALL    FAR PTR COSH    ;  cosh(x) |  sinh(x) | -------- |
        FDIV    ST(1),ST        ;  cosh(x) |  tanh(x) | -------- |
        FSTP    ST(0)           ;  tanh(x) | -------- |
        RET
TANH    ENDP
;*****************************************************************
;            procedures for hyperbolic arcfunctions
;*****************************************************************
ARC_SINH        PROC    FAR
; Calculate hyperbolic arcsine using the formula:
;
;       arcsinh(x) = log(e)[x + SQR(x**2 + 1)]
;
; On entry:
;         ST(0) = x
; On exit:
;         ST(0) = arcsin(x)
; Calculate x**2
        FLD     ST(0)           ;    x     |     x    | -------- |
        FMUL    ST(0),ST        ;   x**2   |     x    | -------- |
        FLD1                    ;    1     |   x**2   |    x     |..
        FADDP   ST(1),ST        ;(x**2 + 1)|    x     | -------- |
        FSQRT                   ;SQR(x2+1) |    x     |
        FADDP   ST(1),ST        ; x + SQR( |
                                ; x**2 + 1)| -------- |
        CALL    LOG_E           ;arcsinh(x)| -------- |
        RET
ARC_SINH        ENDP
;
ARC_COSH        PROC    FAR
; Calculate hyperbolic arcosine using the formula:
;
;       arcosh(x) = log(e)[x + SQR(x**2 - 1)]
;
; On entry:
;         ST(0) = x
; On exit:
;         ST(0) = arcosh(x)
; Calculate x**2
        FLD     ST(0)           ;    x     |     x    | -------- |
        FMUL    ST(0),ST        ;   x**2   |     x    | -------- |
        FLD1                    ;    1     |   x**2   |    x     |..
        FSUBP   ST(1),ST        ;(x**2 + 1)|    x     | -------- |
        FSQRT                   ;SQR(x2+1) |    x     |
        FADDP   ST(1),ST        ; x + SQR( |
                                ; x**2 + 1)| -------- |
        CALL    LOG_E           ;arcsinh(x)| -------- |
        RET
ARC_COSH        ENDP
;
ARC_TANH        PROC    FAR
; Calculate hyperbolic arctangent using the formula:
;
;       arctanh(x) = log(e) SQR [(1 + x)/(1 - x)]
;
; On entry:
;         ST(0) = x
; On exit:
;         ST(0) = arctanh(x)
;
;***************************|
;      test for negative    |
;        argument           |
;***************************|
        MOV     CS:SIGN,0       ; Assume positive
        FTST                    ; Test ST(0)
        FSTSW   CS:STATUS_WW    ; Store status word
        FWAIT
        MOV     AH,BYTE PTR CS:STATUS_WW + 1
        TEST    AH,00000001B    ; Is bit C0 set?
        JZ      POSITIVE_VALUE  ; Go if positive
; At this point input angle is negative
        MOV     CS:SIGN,1       ; Store sign of input angle
        FABS                    ; Change sign
;***************************|
;  calculate arctanh(x)     |
;***************************|
; Calculate 1 + x 
POSITIVE_VALUE:
        FLD     ST(0)           ;    x     |     x    | -------- |
        FLD1                    ;    1     |     x    |     x    |
        FADDP   ST(1),ST        ;  1 + x   |     x    | -------- |
; Calculate 1 - x
        FXCH                    ;    x     |   1 + x  | -------- |
        FLD1                    ;    1     |     x    |   1 + x  |
        FXCH                    ;    x     |     1    |   1 + x  |
        FSUBP   ST(1),ST        ;  1 - x   |   1 + x  | -------- |
        FDIVRP  ST(1),ST        ; 1+x/1-x  | -------- |
        FSQRT                   ;SQR(1+x/  |
                                ;  1-x)    | -------- |
        CALL    LOG_E           ;arctanh(x)| -------- |
        FABS                    ; Change to positive
;***************************|
;      determine sign       |
;      of arcfunction       |
;***************************|
; Test for sign change in input angle
        CMP     CS:SIGN,0
        JE      POS_ATANH
        FCHS                    ; Change sign of tangent
POS_ATANH:
        RET
ARC_TANH        ENDP
;
CODE    ENDS
        END



                  ; Change sign of tangent
