;*****************************************************************
;*****************************************************************
;                           UNIT_8.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date:  March 31, 1994
; Updated:        April 12, 1994
;
; Libary name: MATHUNIT.LIB
;
;*****************************************************************
;                           publics
;*****************************************************************
; PROCEDURES IN THIS MODULE
PUBLIC  CLS_18          ; .. clear screen in VGA mode 18
PUBLIC  SET_VIDEO_MODE  ; .. set VGA video mode using BIOS
PUBLIC  SET_WRITE_MODE  ; .. set VGA write mode
PUBLIC  ES_TO_VIDEO     ; .. ES register to graphics video base
PUBLIC  PIX_ADD_18      ; .. pixel address in VGA mode 18
PUBLIC  PIX_SET_18      ; .. set pixel in VGA mode 18
PUBLIC  PIXEL_WRITE     ; .. pixel address and set super procedure
                        ;    for VGA mode 18
PUBLIC  EQUATION_PLOT   ; .. plot an equation defined by caller
PUBLIC  BRESENHAM       ; .. plot straight line
PUBLIC  LINE_BY_SLOPE   ; .. plot straight line using it's slope
PUBLIC  DISPLAY_LINE    ; .. display stored coordinates of line
PUBLIC  CIRCLE          ; .. plot circle in fourth quadrant
PUBLIC  ELLIPSE         ; .. idem for ellipse
PUBLIC  PARABOLA        ; .. idem for parabola
PUBLIC  HYPERBOLA       ; .. idem for hyperbola
PUBLIC  QUAD_I          ; .. curve display in quadrant I
PUBLIC  QUAD_II         ; .. idem for quadrant II
PUBLIC  QUAD_III        ; .. idem for quadrant III
PUBLIC  QUAD_IV         ; .. idem for quadrant IV
PUBLIC  DO_4_QUADS      ; .. super procedure to display quadrants
PUBLIC  VGA18_TEXT      ; .. text display in VGA mode 18
PUBLIC  SET_CURSOR      ; .. set system cursor for text operation
PUBLIC  SHOW_CHR        ; .. display individual character in
                        ;    graphics mode
;
;*****************************************************************
;                           externals
;*****************************************************************
; In other modules of the MATHUNIT library
EXTRN   DEG_TO_RAD:     FAR
EXTRN   TANGENT:        FAR
EXTRN   CALCULATE_Y:    FAR
;****************************************************************
;                          code segment
;****************************************************************
;
CODE    SEGMENT BYTE
        ASSUME  CS:CODE
;
;****************************|
;     code segment data for  |
;      graphics procedures   |
;****************************|
; Scratch-pad for temporary data
THIS_X  DW      0        ; Normalized coordinate of x
THIS_Y  DW      0        ; Normalized coordinate of y
LAST_Y  DW      0        ; Internal control for
                         ; noncontinuous y-axis points
; Buffers for storing 1K of x and y coordinates of the first
; quadrant and a counter for the number of points computed
Y_BUFFER        DB      2048 DUP (00H)
X_BUFFER        DB      2048 DUP (00H)
POINTS_CNT      DW      0        ; Number of entries in buffers

; Data for BRESEMHAM procedure
END_Y   DW      0        ; Storage for temporary y coordinate
END_X   DW      0        ; Storage for temporary x coordinate
INCR_FOR_Y      DW      0       ; y increment
INCR_FOR_X      DW      0       ; x increment
STRT_HSEGS      DW      0       ; Start of horizontal segment
STRT_VSEGS      DW      0       ; Start of vertical segment
STRT_TOTAL      DW      0       ; Total start values
DIAG_TOTAL      DW      0       ; Total diagonal values
LINE_COLOR      DB      0       ; IRGB color code for line
;
; Data for LINE_BY_SLOPE procedure
; Coordinates of points
X1        DW      320        ; x coordinate of origin
X2        DW      240        ; y coordinate of origin
Y1        DW      0
Y2        DW      0
SLOPE     DW      0          ; Slope in degrees
;
; Origin coordinates for curve display
ORIGIN_X        DW      0       ; x of origin
ORIGIN_Y        DW      0       ; y of origin
;
; Data for curve plotting procedures
RADIUS          DW      0       ; Radius in pixels
X_AXIS          DW      0       ; x semi-axis of ellipse
Y_AXIS          DW      0       ; y semi-axis of ellipse
FOCUS           DW      0       ; Focus of hyperbola
VERTEX          DW      0       ; Vertex
X_START         DW      0       ; Start x coordinate
B_PARAM         DW      0       ; b parameter in hyperbola
VERTEX2         DW      0       ; Square of VERTEX in hyperbola
; Scratchpad
X_ABS           DW      0       ; Absolute coordinate of x
Y_ABS           DW      0       ; Absolute coordinate of y
OPS_CNT         DW      0       ; Operational points counter
                                ; for circle display procedures
; Data for VGA18_TEXT procedure
START_COL       DB      0
;
; Data for EQUATION_PLOT procedure
X_LIMIT         DW      0       ; Limit for plot
EQ_ADD          DW      0       ; Offset of equation in DS
FPU_STATUS      DW      0       ; Storage for status word
;
; These equates determine the pixel coordinate limits of the plot
X_MAX_PIX       EQU     319     ; For origin at screen center
Y_MAX_PIX       EQU     239     ; idem
;
;*****************************************************************
;                    auxiliary procedures
;*****************************************************************
CLS_18  PROC    FAR
; Clear the entire video screen while in VGA mode number 18
; On entry:
;          AL = IRGB color code
; Make sure ES --> video buffer
        CALL    FAR PTR ES_TO_VIDEO     ; Support procedure
        PUSH    AX      ; Save caller's machine
        PUSH    BX
        PUSH    CX
        PUSH    DX
; Set parameters for the TILE_FILL_18 procedure
        MOV     CX,0            ; Start at top left tile
        MOV     DH,80           ; Tiles per screen row
        MOV     DL,60           ; Number of screen rows
        CALL    TILE_FILL_18
        POP     DX              ; Restore caller's context
        POP     CX
        POP     BX
        POP     AX
        RET
CLS_18  ENDP
;*****************************************************************
;
TILE_FILL_18    PROC     NEAR
; Fill a screen area on a tile (8 by 8 pixels) boundary while in
; mode number 18
; On entry:
;         CH = x coordinate of tile (range 0 to 79)
;         CL = y coordinate of tile (range 0 to 59)
;         DH = Total horizontal tiles to fill
;         DL = Total vertical tiles to file
;         AL = I R G B color to use in fill area
;
; Make sure ES --> video buffer
        CALL    FAR PTR ES_TO_VIDEO     ; Support procedure
; Save entry registers
        PUSH    CX
        PUSH    DX
        PUSH    AX
;
        MOV     AH,0FFH     ; Bit pattern for all bits set
        PUSH    AX
        CALL    TILE_ADD_18     ; Device driver in VGA1 module
; Start address is now set
; Next setup column counter in CX and row counter in DX
        MOV     CL,DH       ; Horizontal fill
        MOV     CH,0        ; CX = horizontal counter
        PUSH    CX          ; Save temporarily
; Row counter = DL * 8
        MOV     AX,DX
        MOV     AH,0        ; Clear high byte
        MOV     CL,8        ; Multiplier
        MUL     CL          ; AX = DL * 8
; Set up outer counter in DX
        MOV     DX,AX
        POP     CX          ; Inner loop counter
        POP     AX          ; AL = color
                            ; AH = 11111111B (all bits set)
OUTER:
        PUSH    CX          ; Save inner loop counter
        PUSH    BX          ; Save start address
INNER:
        CALL    FAR PTR PIX_SET_18
        INC     BX          ; Index to next 8-pixel block
        LOOP    INNER
; Index to next column
        POP     BX          ; Restore start of this row
        ADD     BX,80       ; Index to next row start
        POP     CX          ; Restore inner loop counter
        DEC     DX          ; Decrement total rows
        JNZ     OUTER       ; Continue if not end of count
; Restore input registers
        POP     AX
        POP     DX
        POP     CX
        RET
TILE_FILL_18    ENDP
;*****************************************************************
;
TILE_ADD_18     PROC    NEAR
; Procedure to calculate the coarse-grain address at an 8-by-8
; pixel level in VGA mode number 18
;
; On entry:
;         CH = horizontal tile number (range 0 to 79) = x coordinate
;         CL = vertical tile number (range 0 to 59) = y coordinate
;
; Compute coarse-grain address (in BX) as follows:
;     BX = (CL * 640) + CH
;
; On exit:
;          BX = tile offset into video buffer
;          CX is destroyed
;
        PUSH    AX          ; Save accumulator
        PUSH    DX          ; For word multiply
        PUSH    CX          ; To save CH for addition
        MOV     AX,CX       ; Copy CX in AX
; AL = CL
        MOV     AH,0        ; Clear high byte
        MOV     CX,640            ; Multiplier
        MUL     CX          ; AX * CX results in AX
; The multiplier (640) is the product of 80 tiles columns
; times 8 vertical pixels in each tile row
        POP     CX          ; Restore CH
        POP     DX          ; and DX
        MOV     CL,CH       ; Prepare to add in CH
        MOV     CH,0 
        ADD     AX,CX       ; Add
        MOV     BX,AX       ; Move sum to BX
        POP     AX          ; Restore accumulator
        RET
TILE_ADD_18     ENDP
;
;****************************************************************
;     procedures to set the video mode and the write mode
;                       in a VGA system
;****************************************************************
SET_VIDEO_MODE   PROC     FAR
; Set video display mode using BIOS service number 0 of INT 10H
; On entry:
;         AL = number of vide mode to set
; On exit:
;         carry clear
;
        PUSH     AX             ; Save AX
        MOV      AH,0           ; BIOS service request number
        INT      10H
        POP      AX             ; Restore caller's register
        RET
SET_VIDEO_MODE   ENDP
;*****************************************************************

SET_WRITE_MODE  PROC     FAR
; Set the Graphics Controller's Graphic Mode Register to the
; desired write mode
; On entry:
;         AL = write mode requested
; Note: Write mode bits are 0 and 1, all others are preserved
;
; Also set default bit mask
;
        MOV     AH,AL           ; Mode to AH
        MOV     DX,3CEH     ; Graphic Controller Address
                            ; Register
        MOV     AL,5        ; Offset of the Mode Register
        OUT     DX,AL       ; Select this register
        INC     DX          ; Point to data register
; Read contents of mode register
; Write modes bits are 000000xxB
        IN      AL,DX           ; AL has current bit map
        AND     AL,11111100B    ; Clear write mode bits
        OR      AL,AH           ; OR with write mode bits
                                ; passed by caller
; AL now holds bit pattern for mode register
        OUT     DX,AL           ; Set register
; Set Bit Mask Register to default setting
        MOV     DX,3CEH     ; Graphic controller latch
        MOV     AL,8
        OUT     DX,AL       ; Select data register 8
        INC     DX          ; To 3CFH
        MOV     AL,0FFH     ; Default mask
        OUT     DX,AL       ; Load bit mask
        RET
SET_WRITE_MODE  ENDP
;*****************************************************************

ES_TO_VIDEO      PROC    FAR
; Set the ES register to the base address of the video buffer
; in VGA graphics mode
        PUSH    AX              ; Save accumulator
        MOV     AX,0A000H       ; Video segment base for
                                ; graphics modes
        MOV     ES,AX           ; To ES
        POP     AX              ; Restore accumulator
        RET
ES_TO_VIDEO       ENDP


;****************************************************************
;          procedure to calculate a pixel address
;                    in VGA mode 18
;****************************************************************
PIX_ADD_18   PROC     FAR
; Address computation from x and y pixel coordinates 
; On entry: 
;               CX = x coordinate of pixel (range 0 to 639) 
;               DX = y coordinate of pixel (range 0 to 479) 
; On exit:
;               BX = byte offset into video buffer 
;               AH = bit mask for the write operation using
;                    VGA write modes 0 or 2
;               AL is preserved
; Save all entry registers 
        PUSH    CX 
        PUSH    DX 
; Compute address
        PUSH    AX          ; Save accumulator 
        PUSH    CX          ; Save x coordinate 
        MOV     AX,DX       ; y coordinate to AX 
        MOV     CX,80       ; Multiplier (80 bytes per row)
        MUL     CX          ; AX = y times 80 
        MOV     BX,AX       ; Free AX and hold in BX 
        POP     AX          ; x coordinate from stack 
; Prepare for division 
        MOV     CL,8        ; Divisor
        DIV     CL          ; AX / CL = quotient in AL and 
                            ; remainder in AH 
; Add in quotient 
        MOV     CL,AH       ; Save remainder in CL 
        MOV     AH,0        ; Clear high byte 
        ADD     BX,AX       ; Offset into buffer to BX 
        POP     AX          ; Restore AX
; Compute bit mask from remainder 
        MOV     AH,10000000B    ; Unit mask for 0 remainder 
        SHR     AH,CL       ; Shift right CL times 
; Restore all entry registers 
        POP     DX 
        POP     CX 
        RET
PIX_ADD_18      ENDP


;****************************************************************
;          procedure to set an individual screen pixel
;                       in VGA mode 18
;****************************************************************
PIX_SET_18     PROC     FAR
; VGA mode number 18 device driver for writing an individual
; pixel or a pixel pattern to the graphics screen
;
; On entry:
;            ES:BX = byte offset into the video buffer
;               AL = pixel color in IRGB format
;               AH = bit pattern to set
;
; This routine assumes that write mode 2 has been set
;
; Note: programs using this procedure will usually precede its
;       call by one to ES_TO_APA (to set the video segment base)
;       and another one to PIXEL_ADD_18 (to obtain the byte
;       offset and pixel mask).
;       This procedure does not reset the default write mode nor
;       the contents of the Bit Mask register
;
        PUSH    DX          ; Save outer loop counter
        PUSH    AX          ; Color byte
        PUSH    AX          ; Twice
; Set Bit Mask Register according to mask in AH
        MOV     DX,3CEH     ; Graphic controller latch
        MOV     AL,8
        OUT     DX,AL       ; Select data register 8
        JMP     SHORT $+2
        INC     DX          ; To 3CFH
        POP     AX          ; AX once from stack
        MOV     AL,AH       ; Bit pattern
        OUT     DX,AL       ; Load bit mask
        JMP     SHORT $+2
; Write color code
        MOV     AL,ES:[BX]  ; Dummy read to load latch
                            ; registers
        POP     AX          ; Restore color code
        MOV     ES:[BX],AL  ; Write the pixel with the
                            ; color code in AL
        POP     DX          ; Restore outer loop counter
        RET
PIX_SET_18      ENDP
;
;****************************************************************
;          super procedure to calculate and set pixel
;****************************************************************
PIXEL_WRITE     PROC    FAR
; Super procedure to call the procedures that calculate the pixel
; address and set the pixel in VGA mode 18
; On entry:
;         CX = x pixel coordinate
;         DX = y pixel coordinate
;         AL = pixel color in IRGB format
; On exit:
;         pixel is set
;
        PUSH    AX              ; Save context
        PUSH    BX
        CALL    FAR PTR PIX_ADD_18 ; Calculate pixel offset and mask
        CALL    FAR PTR PIX_SET_18 ; Set pixel
        POP     BX              ; Restore context
        POP     AX
        RET
PIXEL_WRITE     ENDP

;****************************************************************
;   support procedure for correcting non-adjacent pixel plots
;****************************************************************
ADJACENT        PROC    NEAR
; Test for adjacent y coordinates
; On entry:
;          CS:SI --> buffer holding x coordinates of curve
;          CS:DI --> buffer holding y coordinates of curve
; Note:
; Adjacency correction is required if the previous y coordinate
; is not adjacent (one less) to the present y coordinate. Code
; assumes that the data variables are located in the code segment
;
        MOV     DX,CS:THIS_Y
        MOV     CX,CS:THIS_X
; Test for first iteration
        CMP     CS:LAST_Y,0        ; LAST_Y = 0 only on the first
                                   ; iteration
        JE      STORE_PIX_XYS   ; No adjacency test required
TEST_4_ADJACENT:
; Condition No. 1.
;               This y < last y minus 1
        MOV     BX,CS:LAST_Y
        DEC     BX              ; Last y minus 1
        CMP     DX,BX           ; Compare to this y
        JB      FILL_IN_PIXEL
; Condition No. 2.
;               This y > last y plus 1
        MOV     BX,CS:LAST_Y
        INC     BX              ; Last y plus 1
        CMP     DX,BX           ; Compare to this y
        JA      FILL_IN_PIXEL
        JMP     STORE_PIX_XYS
; y points are not adjacent. Correct non-adjacency
; BX = last y coordinate minus 1
FILL_IN_PIXEL:
        MOV     CS:[SI],BX      ; Store y coordinate adjacent
                                ; to previous point
        MOV     CS:[DI],CX      ; Store this x coordinate
        ADD     SI,2            ; Bump pointers
        ADD     DI,2
        INC     CS:POINTS_CNT   ; Bump points counter
        MOV     CS:LAST_Y,BX    ; Update to this point
        JMP     TEST_4_ADJACENT
; Store coordinates
STORE_PIX_XYS:
        MOV     CS:[SI],DX      ; Store normalized y coordinate
        MOV     CS:[DI],CX      ; Store normalized x coordinate
; Bump both buffer pointers
        ADD     SI,2
        ADD     DI,2
        INC     CS:POINTS_CNT   ; Bump points counter
        MOV     CS:LAST_Y,DX    ; Update LAST_Y variable
        RET
ADJACENT        ENDP

;*****************************************************************
;    procedure for plotting an equation defined by the caller
;*****************************************************************
EQUATION_PLOT   PROC    FAR
; Procedure to graphics plot an equation by calculating the
; values of the dependent variable (y) from x = 0 to x = ST(0)
; On entry:
;         DS:SI --> ASCII equation for curve
;         ST(0) = integer representing the high limit for the
;                 plot
; On exit:
;         Plotted coordinates are stored in the arrays
;         CS:X_BUFFER and CS:Y_BUFFER
;
; During computations:
;          x coordinate ... CS:THIS_X
;          y coordinate ... CS:THIS_Y
;
;*********************|
;     preparations    |
;*********************|
; Store variables
        FISTP   CS:X_LIMIT      ; Store limit of plot
        MOV     CS:EQ_ADD,SI    ; Store address of equation
; Reset counters and controls
        MOV     CS:LAST_Y,0           ; and LAST_Y
        MOV     CS:THIS_X,0           ; Start value for x
; Buffer pointers:
;       SI --> Y values buffer
;       DI --> X values buffer
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CS:POINTS_CNT,0        ; Reset counter
;*********************|
; calculate y values  |
;*********************|
PLOT_POINT:
        PUSH    SI              ; Save pointers
        PUSH    DI
        FILD    CS:THIS_X      ; x coordinate to ST(0)
        MOV     SI,CS:EQ_ADD   ; Address of equation in DS
        CALL    CALCULATE_Y    ; Procedure in UNIT_9 module
; ST(0) now holds y coordinate
; Test ST(0) for NAN or infinity
        FTST                    ; Set condition codes
        FSTSW   CS:FPU_STATUS   ; Store status
        FWAIT
        MOV     AH,BYTE PTR CS:FPU_STATUS + 1
        AND     AH,01000101B    ; Clear all except C3, C2, and C0
        JZ      NOT_NAN         ; Go if positive and not 0
; At this point ST(0) holds NAN, infinity, zero, or negative
; Replace ST(0) with 0 to avoid crashing display routine
        FSTP    ST(0)           ; Get rid of NAN or infinity
        FLDZ                    ; Load 0
NOT_NAN:
        FISTP   CS:THIS_Y       ; Store y coordinate
        POP     DI              ; Restore local pointer
        POP     SI
        CALL    ADJACENT        ; Local routine
        INC     CS:THIS_X       ; Next x coordinate
; Test scceen limits of plot
        MOV     AX,CS:THIS_X    ; Present x coordinate
        CMP     AX,X_MAX_PIX    ; Test for end of screen
        JB      NO_LIMIT_X      ; Go if less
; At this point the limits of the x coordinate has been reached
        RET
NO_LIMIT_X:
        MOV     AX,CS:THIS_Y    ; Present y coordinate
        CMP     AX,Y_MAX_PIX    ; Test for end of screen
        JB      NO_LIMIT_Y      ; Go if less
; At this point the limits of the y coordinate has been reached
        RET
; Test for end of user-defined range for the x coordinate
NO_LIMIT_Y:
        MOV     AX,CS:THIS_X   ; Present x coordinate
        CMP     AX,CS:X_LIMIT  ; Is this the last coordinate
        JNE     PLOT_POINT     ; Continue if not zero
        RET
EQUATION_PLOT   ENDP

;****************************************************************
;         procedure for plotting a stright line using
;                  Bresenham's algorithm
;****************************************************************
;
BRESENHAM       PROC    FAR
; Routine to draw a straight line using Bresenham's algorithm
; On entry:
;          AL = color of line
;          BX = start point x coordinate
;          CX = start point y coordinate
;          SI = end point x coordinate
;          DI = end point y coordinate
;**********************|
;  entry operations    |
;**********************|
; Move register data to variables
        MOV     CS:LINE_COLOR,AL        ; Store line color
        MOV     CS:ORIGIN_X,BX          ; Store start x
        MOV     CS:ORIGIN_Y,CX          ; Store start y
        MOV     CS:END_X,SI             ; Store end x
        MOV     CS:END_Y,DI             ; Store end y
; Set unit increments for pixel-to-pixel operation
        MOV     CX,1
        MOV     DX,1
; Determine negative or positive slope from difference between
; the y and x coordinates of the start and end points of the line
; This difference is also the length of the line
        MOV     DI,CS:END_Y
        SUB     DI,CS:ORIGIN_Y  ; Length
        JGE     POS_VERTICAL    ; Vertical length is positive
        NEG     DX              ; DX = -1 in 2's complement form
        NEG     DI              ; Make distance positive
POS_VERTICAL:
        MOV     CS:INCR_FOR_Y,DX; Increments on the y axis will
                                ; be positive or negative
; Calculate horizontal distance
        MOV     SI,CS:END_X
        SUB     SI,CS:ORIGIN_X
        JGE     POS_HORZ        ; Horizontal length is positive
        NEG     CX              ; CX = -1 in 2's complement form
        NEG     SI              ; Distance has to be positive
POS_HORZ:
        MOV     CS:INCR_FOR_X,CX ; Increments on the x axis can 
                                 ; also be positive or negative
; Compare the horizontal and vertical lengths of the line to
; determine if the straight segments will be horizontal (if
; this length is greater) or vertical (otherwise)
        CMP     SI,DI            ; SI = horizontal length
                                 ; DI = vertical length
        JGE     HORZ_SEGMENTS
; Vertical length is greater, straight segments are vertical
        MOV     CX,0             ; No horizontal segments
        XCHG    SI,DI            ; Invert lengths
        JMP     SET_CONTROLS
HORZ_SEGMENTS:
        MOV     DX,0             ; No vertical segments
SET_CONTROLS:
        MOV     CS:STRT_HSEGS,CX ; Will be 1 or 0
        MOV     CS:STRT_VSEGS,DX ; Also 1 or 0
; Calculate adjustment factor
        MOV     AX,DI            ; Smaller direction component
        ADD     AX,AX            ; Double the shorter distance
        MOV     CS:STRT_TOTAL,AX ; Straight component total
        SUB     AX,SI            ; Subtract larger direction
                                 ; component
        MOV     BX,AX            ; General component counter
        SUB     AX,SI            ; Calculate
        MOV     CS:DIAG_TOTAL,AX ; Diagonal component total
                                 ; pixels
; Prepare to draw line
        MOV     CX,CS:ORIGIN_X
        MOV     DX,CS:ORIGIN_Y
; SI = the length of the line along the longer axis
        INC     SI
        MOV     AL,CS:LINE_COLOR ; Color code for line
;*********************|
;   draw line points  |
;*********************|
LINE_POINTS:
        DEC     SI               ; Counter for total pixels
        JNZ     PIX_DRAW
        JMP     END_OF_LINE      ; Line is finished
;*********************|
;   display pixel     |
;*********************|
PIX_DRAW:
        CALL    FAR PTR PIXEL_WRITE     ; Routine to set pixel in
                                     ; mode 18
        CMP     BX,0            ; If BX < 0 then straight segment
                                ; diagonal segment otherwise
        JGE        DIAGONAL
; Draw straight line segments
        ADD     CX,CS:STRT_HSEGS ; Increment CX if horizontal
        ADD     DX,CS:STRT_VSEGS ; Increment DX if vertical
        ADD     BX,CS:STRT_TOTAL ; Counter plus adjustment
        JMP     LINE_POINTS
; Draw diagonal segment
DIAGONAL:
        ADD     CX,CS:INCR_FOR_X       ; X direction
        ADD     DX,CS:INCR_FOR_Y       ; Y direction
        ADD     BX,CS:DIAG_TOTAL       ; Adjust counter
        JMP     LINE_POINTS
END_OF_LINE:
        RET
BRESENHAM       ENDP


;****************************************************************
;           plotting a straight line by its slope
;****************************************************************
LINE_BY_SLOPE   PROC    FAR
; Procedure to plot and store the pixel coordinates of a straight
; line of slope s, located in the fourth quadrant
; The slope must be in the range 0 < m > 90 degrees
; On entry:
;          AX = x coordinate of start point
;          BX = y coordinate of start point
;          CX = x coordinate of end point
;          DX = slope
; Working storage
;          CS:X1    = x coordinate of origin
;          CS:Y1    = y coordinate of origin
;          CS:X2    = x coordinate of end point
;          CS:SLOPE = Slope in degrees
;
; Formula:
;       y = x Tan m
;
;*********************|
;     preparations    |
;*********************|
; Move register data to variables
        MOV     CS:X1,AX        ; Store start x
        MOV     CS:Y1,BX        ; Store start y
        MOV     CS:X2,CX        ; Store end x
        MOV     CS:SLOPE,DX     ; Store slope
; Set registers and variables to coordinates
        LEA     SI,CS:Y_BUFFER  ; y buffer pointer
        LEA     DI,CS:X_BUFFER  ; x buffer pointer
        MOV     CS:LAST_Y,0     ; First iteration
        MOV     CS:POINTS_CNT,0 ; Reset points counter
; Calculate the normalized x coordinate for the rightmost point
;                              |  ST(0)   |   ST(1)  |   ST(2)  |
        FILD    CS:X1          ;   x1     |
        FILD    CS:X2          ;   x2     |     x1   |
        FSUB    ST,ST(1)       ; x2 - x1  |     x1   |
        FSTP    ST(1)          ; x2 - x1  |  empty   |
; Store in variable for the normalized x coordinate of
; rightmost point
        FIST    THIS_X
        FSTP    ST(0)         ; empty    |
; Obtain and store tangent of slope
;                             |  ST(0)   |   ST(1)  |   ST(2)  |
        FILD    CS:SLOPE      ; m(deg)   |
        CALL    DEG_TO_RAD    ; m(rads)  |
        CALL    TANGENT       ; tan m    |
Y_BY_SLOPE:
;                             |  tan m   |
        FILD    CS:THIS_X     ;     x    |  tan s  |
        FMUL    ST,ST(1)      ;     y    |  tan s  |
; Store in variable for normalized y coordinate of this point
        FISTP   CS:THIS_Y     ;  tan s   |
;*********************|
;  test for adjacent  |
;      y values       |
;*********************|
        CALL    ADJACENT      ; Adjacency test procedure
;*********************|
; test for last pixel |
;*********************|
        CMP     CS:THIS_X,0     ; x = 0 must be calculated
        JE      EXIT_SLOPE
        DEC     CS:THIS_X
        JMP     Y_BY_SLOPE
; Adjust FPU stack
EXIT_SLOPE:
        FSTP    ST(0)
        RET
LINE_BY_SLOPE   ENDP


;************************************************************
;        procedure to display a straight line from its
;                   stored coordinates
;************************************************************
DISPLAY_LINE    PROC    FAR
; Display coordinates stored in CS:X_BUFFER and CS:Y_BUFFER in
; VGA mode 18
; Total number of coordinates is stored in CS:POINTS_CNT
; Setup pointers and counter
        LEA     SI,CS:Y_BUFFER       ; y coordinates
        LEA     DI,CS:X_BUFFER       ; x coordinates
        MOV     CX,CS:POINTS_CNT
        MOV     CS:OPS_CNT,CX        ; Operational counter
DISP_1:
        MOV     CX,CS:X1             ; x coordinate of origin
        MOV     DX,CS:Y1             ; y coordinate of origin
; Add stored values to origin
        ADD     CX,WORD PTR CS:[DI]
        SUB     DX,WORD PTR CS:[SI]
; CS:CX = x coordinate
; CS:DX = y coordinate of point
        PUSH    AX                 ; Save color code
        CALL    FAR PTR PIXEL_WRITE        ; Local procedure
        POP     AX                 ; Restore color code
        ADD     SI,2               ; Bump coordinates pointers
        ADD     DI,2
        DEC     CS:OPS_CNT         ; Operation points counter
        JNZ     DISP_1
        RET
DISPLAY_LINE    ENDP

;****************************************************************
;            procedure for plotting a circular arc
;****************************************************************

CIRCLE  PROC    FAR
; Routine to plot and store the pixel coordinates of a circular
; arc in the fourth quadrant
;
; On entry:
;                           REGISTER        STORED IN
;       radius of circle       BX           CS:RADIUS
;
; During computations:
;          x coordinate ... CS:THIS_X
;          y coordinate ... CS:THIS_Y
;*********************|
;     preparations    |
;*********************|
; Store variables
        MOV     CS:RADIUS,BX    ; Entry parameter for radius
; Reset counters and controls
        MOV     CS:THIS_X,0           ; Start values for x

        MOV     CS:LAST_Y,0           ; and LAST_Y
; Buffer pointers:
;       SI --> Y values buffer
;       DI --> X values buffer
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CS:POINTS_CNT,0        ; Reset counter
;*********************|
; calculate y values  |
;*********************|
CIRCLE_Y:
;                              |  ST(0)   |   ST(1)  |
        FILD    CS:THIS_X      ;    x      |
        FMUL    ST,ST(0)       ;    x^2    |
        FILD    CS:RADIUS      ;    r      |    x^2   
        FMUL    ST,ST(0)       ;    r^2    |    x^2   
        FSUB    ST,ST(1)       ; r^2 - x^2 |    x^2
        FSQRT                  ;Rt(r^2-x^2)|    x^2
        FISTP   CS:THIS_Y      ;    x^2    |
        FSTP    ST(0)          ;  EMPTY    |
; Test adjacency condition
        CALL    ADJACENT       ; Library procedure
        INC     CS:THIS_X      ; x increments in a circle's
                               ; fourth quadrant
        CMP     CS:THIS_Y,0    ; Test for end of execution
        JNE     CIRCLE_Y
; At this point all coordinates have been plotted
        RET
CIRCLE  ENDP


;****************************************************************
;            procedure for plotting an ellipse
;****************************************************************

ELLIPSE         PROC    FAR
; Routine to plot and store the pixel coordinates of an elliptical
; curve in the fourth quadrant
;
; On entry:
;                           REGISTER        STORED IN
;       x semiaxis (M)         BX           CS:X_AXIS
;       y semiaxis (m)         CX           CS:Y_AXIS
;
; During computations:
;          x coordinate ... CS:THIS_X
;          y coordinate ... CS:THIS_Y
;         CS:m = minor axis (X_AXIS or Y_AXIS variables)
;         CS:M = major axis (X_AXIS or Y_AXIS variables)
;
;*********************|
;     preparations    |
;*********************|
; Move data to variables
        MOV     CS:X_AXIS,BX            ; Store x semiaxis
        MOV     CS:Y_AXIS,CX            ; Store y semiaxis
; Reset counters and controls
        MOV     CS:THIS_X,0        ; Start value for x
        MOV     CS:LAST_Y,0        ; and for LAST_Y
; Buffer pointers:
;       SI --> Y values buffer
;       DI --> X values buffer
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CS:POINTS_CNT,0    ; Reset counter
ELLIPSE_Y:
; Calculate primitive coordinate of y
; First solve x^2 / M^2
;                              |  ST(0)     |   ST(1)  |
        FILD    CS:X_AXIS      ;    M       |
        FMUL    ST,ST(0)       ;    M^2     |
        FILD    CS:THIS_X      ;    x       |     M^2   |
        FMUL    ST,ST(0)       ;    x^2     |     M^2   | 
        FDIV    ST,ST(1)       ; x^2/M^2    |     M^2   |
; Solve 1 - (x^2 / M^2)
        FLD1                   ;    1       | x^2/M^2   | ? |
        FSUB    ST,ST(1)       ;1-(x^2/M^2) | x^2/M^2   | ? |
; Solve  m^2 * [1-(x^2/M^2)]
        FILD    CS:Y_AXIS      ;    m       |1-(x^2/M^2)| ? | ? |
        FMUL    ST,ST(0)       ;    m^2     |1-(x^2/M^2)| ? | ? |
        FMUL    ST,ST(1)       ;m^2 * [1-(x^2/M^2)]|  ? | ? | ? |
; Find square root
        FSQRT                  ;    y       | ? | ? | ? |
        FISTP   CS:THIS_Y      ; Store y in memory
; Adjust stack
        FSTP    ST(0)          ; ? | ? |
        FSTP    ST(0)          ; ? |
        FSTP    ST(0)          ; Stack is empty
; Insure pixel adjacency condition
        CALL    ADJACENT       ; Library procedure
        INC     CS:THIS_X      ; x increments in a clockwise plot
                               ; of the first quadrant
        CMP     CS:THIS_Y,0    ; Test for end of processing
        JNE     ELLIPSE_Y
        RET
ELLIPSE ENDP


;****************************************************************
;            procedure for plotting a parabola
;****************************************************************
PARABOLA        PROC    FAR
; Routine to plot and store the pixel coordinates of a parabolic
; curve in the fourth quadrant
;
; On entry:
;                           REGISTER        STORED IN
;    focus of parabola         BX           CS:FOCUS 
;    start x coordinate        CX           CS:X_START
;
; During computations:
;          x coordinate .......... CS:THIS_X
;          y coordinate .......... CS:THIS_Y
; Formula:
;       y = SQR. ROOT (4ax)
;       Y_ABS = SQR. ROOT (4 * FOCUS * X_ABS)
;
;*********************|
;     preparations    |
;*********************|
; Move register data to variables
        MOV     CS:FOCUS,BX     ; Store focus
        MOV     CS:X_START,CX   ; Store start value for x
; Reset counters and controls
        MOV     AX,CS:X_START        ; Start value for X
        MOV     CS:THIS_X,AX
        MOV     CS:LAST_Y,0          ; Reset LAST_Y
; Buffer pointers:
;       SI --> Y values buffer
;       DI --> X values buffer
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CS:POINTS_CNT,0      ; Reset counter
PARA_Y:
; Calculate primitive coordinate of y
;       y = SQR. ROOT (4ax)
;       THIS_Y = SQR. ROOT (4 * FOCUS * THIS_X)
;                              |   ST(0)    |   ST(1)   |   |   |
        FILD    CS:THIS_X      ;     x      |
        FILD    CS:FOCUS       ;     a      |      x    |
        FMUL    ST,ST(1)       ;    ax      | ? |
        FLD1                   ;    1       |     ax    | ? |
        FADD    ST,ST(0)       ;    2       |     ax    | ? |
        FADD    ST,ST(0)       ;    4       |     ax    | ? |
        FMUL    ST,ST(1)       ;   4ax      | ? | ? |
        FSQRT                  ;    y       | ? | ? |
        FISTP   CS:THIS_Y      ; Store y in memory
; Adjust stack
        FSTP    ST(0)          ; ? |
        FSTP    ST(0)          ; Stack is empty
; Insure pixel adjacency conditions
        CALL    ADJACENT       ; Library procedure
        DEC     CS:THIS_X
        CMP     CS:THIS_Y,0    ; Test for end of processing
        JNE     PARA_Y
        RET
PARABOLA        ENDP


;****************************************************************
;            procedure for plotting a hyperbola
;****************************************************************
HYPERBOLA       PROC    FAR
; Routine to plot and store the pixel coordinates of a hyperbolic
; curve in the fourth quadrant
; On entry:
;                           REGISTER        STORED IN
;    focus of hyperbola        BX           CS:FOCUS 
;    vertex of hyperbola       CX           CS:VERTEX 
;    start x coordinate        DX           CS:X_START
;
; During computations:
;          X coordinate .......... CS:THIS_X
;          Y coordinate .......... CS:THIS_Y
; Scratch-pad variables:
;          Numerator radix ..... CS:B_PARAM
;          Vertex squared ...... CS:VERTEX2
;
;*********************|
;     preparations    |
;*********************|
; Move register data to variables
        MOV     CS:FOCUS,BX     ; Store focus
        MOV     CS:VERTEX,CX    ; Store vertex
        MOV     CS:X_START,DX   ; Store start value for x
; Reset counters and controls
        MOV     AX,CS:X_START        ; Start value for X
        MOV     CS:THIS_X,AX
        MOV     CS:LAST_Y,0          ; Reset LAST_Y
; Buffer pointers:
;       SI --> Y values buffer
;       DI --> X values buffer
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CS:POINTS_CNT,0      ; Reset counter
; Compute numerator radical from VERTEX and FOCUS
; Solve: B_PARAM = SQR (FOCUS**2 - VERTEX**2)
;                             |   ST(0)    |   ST(1)   |   |   |
        FILD    CS:VERTEX     ;     a      |
        FMUL    ST,ST(0)      ;    a**2    |
        FILD    CS:FOCUS      ;     c      |   a**2    |
        FMUL    ST,ST(0)      ;    c**2    |   a**2    |
        FSUB    ST,ST(1)      ; c**2 - a**2|   a**2    |
        FSQRT                 ;     b      |   a**2    |
; Store b
        FISTP   CS:B_PARAM    ;   a**2     |
; Store VERTEX^2 for calculations
        FISTP   CS:VERTEX2    ; Stack is empty
HYPER_Y:
; Calculate primitive coordinate of y
;  y = b / a * SQR (x**2 - a**2)
;  or:
;  Y_ABS = B_PARAM / VERTEX * SQR (X_ABS**2 - VERTEX2)
;                             |   ST(0)    |   ST(1)   |   |   |
        FILD    CS:VERTEX2    ;   a**2     |
        FILD    CS:THIS_X     ;    x       |    a**2   |
        FMUL    ST,ST(0)      ;   x**2     |    a**2   |
        FSUB    ST,ST(1)      ; x**2-a**2  | ? |
        FSQRT                 ;SR(x**2-    |
                              ;  a**2)
        FILD    CS:B_PARAM    ;     b      | # | ? |
        FILD    CS:VERTEX     ;     a      |     b     | # | ? |
        FDIV    ST(1),ST      ;     b      |    b/a    | # | ? |
        FSTP    ST(0)         ;    b/a     | # | ? |
        FMUL    ST,ST(1)      ;    y       | ? | ? |
        FISTP   CS:THIS_Y     ; Store y in memory
; Adjust stack
        FSTP    ST(0)         ; ? |
        FSTP    ST(0)         ; Stack is empty
; Insure pixel adjacency condition
        CALL    ADJACENT      ; Library procedure
        DEC     CS:THIS_X
        CMP     CS:LAST_Y,0   ; Test for end of processing
        JNE     HYPER_Y
        RET
HYPERBOLA       ENDP


;****************************************************************
;           quadrant display procedures for conic curves
;****************************************************************
QUAD_I          PROC    FAR
; Display figure with stored coordinates at X_BUFFER and 
; Y_BUFFER with total number of coordinates at POINTS_CNT
; On entry:
;           AL = color code in IRGB format (for mode 18)
;           BX = x coordinate of origin
;           CX = y coordinate of origin
;
; Move entry parameters to variables
        MOV     CS:ORIGIN_X,BX          ; x start
        MOV     CS:ORIGIN_Y,CX          ; y start
        PUSH    AX                      ; Save entry parameters
        PUSH    BX
        PUSH    CX
;
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CX,CS:POINTS_CNT
        MOV     CS:OPS_CNT,CX
QUAD_1:
        PUSH    AX                 ; Save color code
        MOV     CX,CS:ORIGIN_X     ; X offset of this circle
        MOV     AX,CS:[DI]         ; Computed absolute coordinate
        ADD     CX,AX              ; Add x in quadrant I
        MOV     DX,CS:ORIGIN_Y     ; Y offset of this circle
        MOV     AX,CS:[SI]         ; Computed y abs. coordinate
        SUB     DX,AX              ; Subtract y in quadrant I
        POP     AX
; CX = x coordinate of point
; DX = y coordinate of point
        CALL    FAR PTR PIXEL_WRITE        ; Local procedure
        ADD     SI,2
        ADD     DI,2               ; Bump counters forward
        DEC     CS:OPS_CNT         ; Operation points counter
        JNZ     QUAD_1
        POP     CX              ; Restore entry values
        POP     BX
        POP     AX
        RET
QUAD_I          ENDP
;****************************************************************
;
QUAD_IV         PROC    FAR
; Move entry parameters to variables
        MOV     CS:ORIGIN_X,BX          ; x start
        MOV     CS:ORIGIN_Y,CX          ; y start
        PUSH    AX                      ; Save entry parameters
        PUSH    BX
        PUSH    CX
;
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CX,CS:POINTS_CNT
        MOV     CS:OPS_CNT,CX
; Set pointers to end of buffer
        ADD     CX,CX           ; Two bytes per point
; Adjust pointer to last entry in table
        DEC     CX
        DEC     CX
        ADD     SI,CX
        ADD     DI,CX
;
QUAD_4:
        PUSH    AX              ; Save color code
        MOV     CX,CS:ORIGIN_X  ; X offset of this circle
        MOV     AX,CS:[DI]      ; Computed absolute coordinate
        ADD     CX,AX           ; Add x in quadrant I
        MOV     DX,CS:ORIGIN_Y  ; Y offset of this circle
        MOV     AX,CS:[SI]      ; Computed y absolute
        ADD     DX,AX           ; Subtract y in quadrant I
        POP     AX
;
        CALL    FAR PTR PIXEL_WRITE     ; Local procedure
        SUB     SI,2
        SUB     DI,2            ; Bump counters forward
        DEC     CS:OPS_CNT      ; Operation points counter
        JNZ     QUAD_4
        POP     CX              ; Restore entry values
        POP     BX
        POP     AX
        RET
QUAD_IV         ENDP
;****************************************************************
;
QUAD_III        PROC    FAR
; Move entry parameters to variables
        MOV     CS:ORIGIN_X,BX          ; x start
        MOV     CS:ORIGIN_Y,CX          ; y start
        PUSH    AX                      ; Save entry parameters
        PUSH    BX
        PUSH    CX
;
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CX,CS:POINTS_CNT
        MOV     CS:OPS_CNT,CX
;
QUAD_3:
        PUSH    AX                 ; Save color code
        MOV     CX,CS:ORIGIN_X     ; X offset of this circle
        MOV     AX,CS:[DI]         ; Computed absolute
        SUB     CX,AX              ; Subtract x in quadrant III
        MOV     DX,CS:ORIGIN_Y     ; Y offset of this circle
        MOV     AX,CS:[SI]         ; Computed y absolute
        ADD     DX,AX              ; Add y in quadrant III
        POP     AX
        CALL    FAR PTR PIXEL_WRITE        ; Local procedure
        ADD     SI,2
        ADD     DI,2               ; Bump counters forward
        DEC     CS:OPS_CNT         ; Operation points counter
        JNZ     QUAD_3
        POP     CX                 ; Restore entry values
        POP     BX
        POP     AX
        RET
QUAD_III        ENDP
;****************************************************************
;
QUAD_II         PROC    FAR
; Move entry parameters to variables
        MOV     CS:ORIGIN_X,BX          ; x start
        MOV     CS:ORIGIN_Y,CX          ; y start
        PUSH    AX                      ; Save entry parameters
        PUSH    BX
        PUSH    CX
;
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CX,CS:POINTS_CNT
        MOV     CS:OPS_CNT,CX
; Set pointers to end of buffer
        ADD     CX,CX           ; Two bytes per point
; Adjust pointer to last entry in table
        DEC     CX
        DEC     CX
        ADD     SI,CX
        ADD     DI,CX
;
QUAD_2:
        PUSH    AX              ; Save color code
        MOV     CX,CS:ORIGIN_X     ; X offset of this circle
        MOV     AX,CS:[DI]         ; Computed absolute
        SUB     CX,AX           ; Subtract x in quadrant II
        MOV     DX,CS:ORIGIN_Y     ; Y offset of this circle
        MOV     AX,CS:[SI]         ; Computed y absolute
        SUB     DX,AX           ; Subtract y in quadrant II
        POP     AX
        CALL    FAR PTR PIXEL_WRITE        ; Local procedure
        SUB     SI,2
        SUB     DI,2            ; Bump counters forward
        DEC     CS:OPS_CNT         ; Operation points counter
        JNZ     QUAD_2
        POP     CX              ; Restore entry values
        POP     BX
        POP     AX
        RET
QUAD_II         ENDP
;****************************************************************
;
DO_4_QUADS      PROC    FAR
        CALL    FAR PTR QUAD_I
        CALL    FAR PTR QUAD_IV
        CALL    FAR PTR QUAD_III
        CALL    FAR PTR QUAD_II
        RET
DO_4_QUADS      ENDP

;****************************************************************
;   procedure to display text in a VGA alpha or graphics mode
;****************************************************************
VGA18_TEXT      PROC    FAR
; Display a message pointed at by DS:DI on the graphic screen
; Message format as follows:
; OFFSET:       CONTENTS:
;       0 ----> Screen row
;       1 ----> Screen column
;       2 ----> Color code
; DISPLAY CODES:
;       00H --> End of message
;       FFH --> End of screen line
; On entry:
;          DS:DI --> Graphic display block
;
; Note: code assumes that VGA write mode 0 is active
;
        MOV     DH,[DI]         ; Get row into DH
        INC     DI              ; Bump pointer
        MOV     DL,[DI]         ; And column into DL
        MOV     CS:START_COL,DL ; Store start column
; Set system cursor using BIOS service number 2 of INT 10H
        CALL    FAR PTR SET_CURSOR      ; Support procedure
        INC     DI              ; Bump pointer to attribute
        MOV     BL,[DI]         ; Get color code into BL
CHAR_WRITE:
        INC     DI              ; Bump to message start
        MOV     AL,[DI]         ; Get character
        CMP     AL,0FFH         ; End of line ?
        JE      BUMP_ROW        ; Next row
        CMP     AL,0            ; Test for terminator
        JZ      END_TEXT        ; Exit routine
;***********************|
;    display character  |
;***********************|
; Display character in AL and using the color code in BL
        MOV     AH,9            ; BIOS service request number
        MOV     BH,0            ; Page
        MOV     CX,1            ; No repeat
        INT     10H
;**********************|
;     bump cursor      |
;**********************|
        INC     DL
        CALL    FAR PTR SET_CURSOR ; Support procedure
        JMP     CHAR_WRITE
END_TEXT:
        RET
BUMP_ROW:
        INC     DH              ; Row control register
        MOV     DL,CS:START_COL ; Column control to start column
        CALL    FAR PTR SET_CURSOR      ; Support procedure
        JMP     CHAR_WRITE
VGA18_TEXT      ENDP

SET_CURSOR      PROC    FAR
; Set system cursor to a row and column coordinate
; On entry:
;         DH = screen row (range 0 to 24)
;         DL = screen column (range 0 to 79)
; Note: routine does not test for valid parameters
; On exit:
;         cursor set
;
        PUSH    AX              ; Save entry registers
        PUSH    BX
        PUSH    DX
        MOV     AH,2            ; Service request number
        MOV     BH,0            ; Assume display page No. 0
        INT     10H
        POP     DX              ; Restore registers
        POP     BX
        POP     AX
        RET
SET_CURSOR      ENDP

SHOW_CHR        PROC    FAR
; Display character in AL and using the color code in BL
        MOV     AH,9            ; BIOS service request number
        MOV     BH,0            ; Page
        MOV     CX,1            ; No repeat
        INT     10H
        RET
SHOW_CHR        ENDP


CODE    ENDS
        END

 MOV     CX,1            ; No repeat
        INT     10H
        RET
SHOW_CHR        ENDP


