;*****************************************************************
;*****************************************************************
;                           UN32_9.ASM
;*****************************************************************
;*****************************************************************
; Library of math routines from the book SOFTWARE SOLUTIONS by
; Julio Sanchez and Maria P. Canton
; Creation date:  February 9, 2001
; Last update:    
;
; Libary name: MATH32.LIB
; This module: UN32_9.ASM
; Operations : Probability and statistical calculations
;*****************************************************************
;
;*****************************************************************
;                           publics
;*****************************************************************
; MATHEMATICAL ROUTINES IN THIS MODULE
; Probability
PUBLIC  _PERMUTATIONS        ; .. Number of permutations
PUBLIC  _COMBINATIONS        ; .. Number of combinations
PUBLIC  _BINOMIAL_PROB       ; .. Binomial probability
PUBLIC  _NORMAL_CURVE_Y      ; .. f(x) for normal dist. curve
PUBLIC  _LOADE               ; .. Load constant e
;****************************************************************
;               definitions for 32-bit flat model
;****************************************************************
        .486
        .MODEL flat
;*****************************************************************
;                           externals
;*****************************************************************
; In other modules of the MATH32 library
EXTRN   _FACTORIAL:        PROC   ; ----  Un32_7 module  
EXTRN   _X_TO_Y_BYFAC:     PROC   ; ----  Un32_5 module
EXTRN   _X_TO_Y_MIXED:     PROC   ; ----  Un32-5 module
        .DATA
;***************************|
;   Data for this module    |
;***************************|
NELS        DD         0     ; Storage for n
RELS        DD         0     ; Storage for r
TEMPV       DD         0     ; Temporary storage
; For normal distributions curve
X           DD         0     ; x coordinate of graph
U           DD         0     ; Mean
SD          DD         0     ; Standard deviation
; Definition of constant e
E           DT         4000ADF85458A2BB4A9BH
;****************************************************************
;                           code
;****************************************************************
        .CODE  
;******************************************
;         P R O B A B I L I T Y 
;******************************************

_PERMUTATIONS    PROC
; Ordered permutation without repetitions of n elements, in set P,
; taken r at a time, according to the formula:
;
;              n!
;     nPr = ---------
;            (n - r)!
;
; where:
;          nPr = number of permutations of r elements in a set of
;                n elements
;          n = number of elements in set P
;          r = number of elements in permuation (r at a time)
;
; Note: code assumes that n and r are integers
; On entry:
;          ST(0) = n (number of elements)
;          ST(1) = r (taken r at a time)
; On exit:
;          ST(0) = number of possible permutations
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   r   | ----- | ----- |
; Calculate (n - r)!
        FIST       NELS         ; Save n
        FSUB       ST,ST(1)     ;  n-r  |   r   | ----- |
        CALL       _FACTORIAL   ;(n-r)! |   r   | ----- |
        FILD       NELS         ;   n   | (n-r)!| ----- |
        CALL       _FACTORIAL   ;  n!   | (n-r)!| ----- |		
        FDIV       ST,ST(1)     ; n!/
		                        ;(n-r)! | (n-r)!| ----- |
        FSTP       ST(1)
		FRNDINT                 ; Round ST to integer        	        
        CLD
        RET
_PERMUTATIONS      ENDP
;
;
_COMBINATIONS      PROC
; Combination of r elements, in set P, that contains 
; n elements, according to the formula:
;
;               n!
;     nCr =  ----------
;            r!(n - r)!
;
; where:
;          nCr = number of combinations of r elements in a set of
;                n elements
;          n = number of elements in set P
;          r = number of combinations of r elements
;
; Note: code assumes that n and r are integers
; On entry:
;          ST(0) = n (number of elements)
;          ST(1) = r (taken r at a time)
; On exit:
;          ST(0) = number of possible conbinations
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   r   | ----- | ----- |
; Calculate r!(n - r)!
        FISTP      NELS         ; Save n
                                ;   r   | ----- |
		FIST       RELS         ; Save r  
;
; Calculate r!
        CALL       _FACTORIAL   ;  r!   |
		FILD       RELS         ;  r    |   r!  | ----- |
		FILD       NELS         ;  n    |   r   |   r!  | ----- |
; Calculate (n-r)!		       
        FSUB       ST,ST(1)     ;  n-r  |   r   |   r!  |
        FSTP       ST(1)        ;  n-r  |   r!  | ----- |
        CALL       _FACTORIAL   ;(n-r)! |   r!  | ----- |
; Calculate r!(n-r)! = [d]
        FMUL                    ;  [d]  | ----- |
        FILD       NELS         ;   n   |  [d]  | ----- |
        CALL       _FACTORIAL   ;  n!   |  [d]  | ----- |		
        FDIV       ST,ST(1)     ; n!/[d]|  [d]  | ----- |
        FSTP       ST(1)
		FRNDINT                 ; Round ST to integer        	        
        CLD
        RET
_COMBINATIONS     ENDP


_BINOMIAL_PROB    PROC
; In an event that can occur in two equally likely ways,
; calculate the probability of obtaining r repetitions out
; of n trials, according to the formula:
;
;              nCr
;     P =  ----------
;              2^n
;
; where:
;          nCr = number of combinations of r elements in a set of
;                n elements
;          n = number of trials
;          r = number of repetitions
;          P = binomial probability of event r
;
; Note: code assumes that n and r are integers
; On entry:
;          ST(0) = n (number of elements)
;          ST(1) = r (taken r at a time)
; On exit:
;          ST(0) = binomial probability
;                               | ST(0) | ST(1) | ST(2) | ST(3) |
;                               |   n   |   r   | ----- | ----- |
        FIST       NELS         ; Save n
; Calculate nCr
        CALL       _COMBINATIONS 
;                               |  nCr  | ----- |
        FISTP      TEMPV        ; Store nCr
; Calculate 2^n
        FLD1                    ;   1   | ----- |
		FLD1                    ;   1   |   1   | ----- |
		FADD                    ;   2   | ----- |
		FILD       NELS         ;   n   |   2   | ----- |
        CALL       _X_TO_Y_BYFAC
		                        ;  2^n  | ----- |
        FILD       TEMPV        ; nCr   |  2^n  |
        FDIV       ST,ST(1)		;nCr/2^n|  2^n  |
		FSTP       ST(1)        ;nCr/2^n| ----- |
        CLD
        RET
_BINOMIAL_PROB     ENDP


_NORMAL_CURVE_Y   PROC
; Calculates y = f(x) for the normal distribution curve based on
; the formula:
;
;             e^-((x-u)^2/2s^2)
;     y = ---------------------
;                  s SQR(2Pi)  
;
; where:
;       x is the cartesian coordinate in terms of the
;         mean. 
;       y is the value of the dependent variable
;       u is the mean
;       s is the standard deviation
; On entry:
;       ST(0) = x
;       ST(1) = u
;       ST(2) = s
;       e is defined as a memory variable
; On exit:
;       ST(0) = Y
;*************************|
;    save parameters      |
;*************************|
;                           ;  ST0  |  ST1  |  ST2  |
;                           |   x   |   u   |   s   |
        FSTP     X          ;   u   |   s   | ----- |
        FSTP     U          ;   s   | ----- |
		FSTP     SD         ; ----- |
      ; Calculate 2s^2
        FLD1                ;   1   | ----- |
		FADD     ST,ST(0)   ;   2   | ----- |
		FLD		 SD         ;   s   |   2   | ----- |
        FMUL     ST,ST(0)   ;  s^2  |   2   | ----- |
        FMUL                ; 2s^2  | ----- |
     ; Calculate (x-u)^2
        FLD      U          ;   u   | 2s^2  | ----- |    
	    FLD      X          ;   x   |   u   | 2s^2  |
        FSUB     ST,ST(1)   ;  x-u  |   u   | 2s^2  |
		FSTP     ST(1)      ;  x-u  | 2s^2  | ----- |
		FMUL     ST,ST(0)   ;(x-u)^2| 2s^2  | ----- |
        FDIV     ST,ST(1)   ;(x-u)^2
		                    ; /2s^2 | 2s^2  | ----- |
        ; (x-u)^2/2s^2 = w
		FSTP     ST(1)      ;   w   | ----- |
	; calculate e^w
	    FLD      E          ;  e    |   w   | ----- |
        FXCH                ;  w    |   e   | ----- |
		CALL     _X_TO_Y_MIXED ; In Un32_5
		                    ;  e^w  | ----- |
    ; Calculate e^-w = 1/e^w
	    FLD1                ;   1   | e^w   | ----- |
		FDIV     ST,ST(1)   ; 1/e^w | e^w   | ----- |
		FSTP     ST(1)      ; e^-w  | ----- |
    ; Calculate denominator: s * SQR 2Pi 
	    FLDPI               ;  Pi   |  e^-w | ----- |
		FADD     ST,ST(0)   ;  2Pi  |  e^-w | ----- |
	    FSQRT               ;SR(2Pi)|  e^-w | ----- |
		FLD      SD         ;   s   |SR(2Pi)|  e^-w |
        FMUL                ;S *
		                    ;SR(2Pi)|  e^-w | ----- |
	 ; s * SR(2Pi) = v						
		FXCH                ;  e^-w |   v   | ----- |
		FDIV     ST,ST(1)   ;  f(x) |   v   | ----- |
		FSTP     ST(1)      ;  f(x) | ----- |
        CLD
		RET
_NORMAL_CURVE_Y  ENDP


_LOADE       PROC
; Load the value of e onto ST(1)

		FLD       E
        RET
_LOADE       ENDP


        end


    